/*

nhPhyml :  a program that builds phylogenetic trees from DNA alignments using the non-homogeneous, non-stationary model of evolution conceived by Galtier and Gouy (1998) and the algorithmic structure of phyml (Guindon and Gascuel, 2003).

Copyright (C) Bastien Boussau. jul. 2005 onward

All parts of  the source except where indicated  are distributed under
the GNU public licence.  See http://www.opensource.org for details.

*/

#include "utilities.h"
#include "model.h"
#include "options.h"
#include "free.h"
#include "ml.h"
#include "simu.h"
#include "optimiz.h"
int main(int argc, char **argv)
{
  seq **data;
  allseq *alldata;
  option *input;
  char *s, *sEq;
  FILE *fp_tree_out,*fp_out,*fp_lk_out,*fp_tree_out_Eq, *fp_seq_out, *fp_lnf_out, *fp_dist_out, *fp_comp_out;
  arbre *tree;
  int n_otu, n_data_sets,i, j, nbvrai;//, n_Non_Agglo;
  model *mod;
  time_t t_beg,t_end;
  div_t hour,min;
  // double **dist; //to stock the distances between sequences obtained from gg95.
  double prov; //to compute ts/tv
  double* freq;

  srand(time(NULL));
  fp_seq_out=NULL;
  tree = NULL;
  Init_Constant();

  fflush(stdout);

  input = (option *)Get_Input(argc,argv);

  fp_out = Openfile(input->out_file,input->out_file_open_mode);

  fprintf(fp_out,"\n- nhPhyML %s -\n\n", VERSION);

  //  printf("sizeofdouble : %d\n sizeoflongdouble : %d\n",sizeof(double),sizeof(double ));


  fp_tree_out = Openfile(input->tree_file,input->tree_file_open_mode);
  fp_tree_out_Eq = Openfile(input->tree_file_Eq,input->tree_file_open_mode);
  //If wanted, we can output internal sequences (not well tested)
  if (input->internalsequences) {
    fp_seq_out= Openfile(input->seq_out_file,input->out_file_open_mode);
  }

  //The file where we output pattern likelihoods
  fp_lnf_out = Openfile(input->lnf_out_file,input->out_file_open_mode);

  //The files where we output sequence distributions and composition
  fp_dist_out = Openfile(input->dist_out_file,input->out_file_open_mode);
  fp_comp_out = Openfile(input->comp_out_file,input->out_file_open_mode);
  
  n_data_sets = 0;

  fp_lk_out = fopen(input->lk_file,"w");
  
  n_data_sets++;
  
  time(&t_beg);
  
  n_otu = 0;
  
  if(n_data_sets > input->n_data_sets) 
    {
      data = NULL;
    }
  else
    {
      data = Get_Seq(input->fp_seq,
		     input->interleaved,
		     &n_otu,0,input->datatype);
      
    }

  if(data)
    {
      if(n_data_sets > 1) printf("\n. Data set [#%d]\n",n_data_sets);
      
      printf("\n. Compressing sequences...\n");
      
      alldata = Compact_Seq(data,n_otu,input);
      

      
      Free_Seq(data,alldata->n_otu);
      
      
      
      //  Make_All_Qmat_Struct(input->mod, n_otu); 
      /*Allocate memory 
	for all the instantaneous 
	substitution rate matrices 
	for all branches*/
     
      mod = Init_Model(alldata,input); /*End the initialization of the model*/
   
      
      Check_Ambiguities(alldata,input->datatype,input->stepsize);
      //printf("mod->n_pos : %d\n", mod->n_pos);
      
      printf("\n. Reading user tree...\n");
      
      tree = Read_Tree_File(input->fp_tree, alldata);
      
      if(!tree) 
	{
	  printf("\n. Missing tree for data set #%d\n",n_data_sets);
	  printf("  This data set is not analyzed.\n");
	  data = NULL;
	}
      
    }	    
  else {
    exit(0);
  }
  

  tree->mod        = mod;
  //printf("tree->mod->n_pos : %d\n", tree->mod->n_pos);
  tree->data       = alldata;
  tree->both_sides = 1;
  
  // printf("tree->mod->qmat_struct[1]->qmat[0] : %f\n",tree->mod->qmat_struct[1]->qmat[0]);

  //If we use a codon model, a set of codon equilibrium frequencies must be given to each node
  //At this time, equilibrium GC is the same for all nodes: we might save some time there. 
  /* if (tree->mod->ns > 4) {
    for (i=1;i<2*tree->n_otu-1;i++){
      printf("i :%d, tree->noeud[i]->b[0]->qmat_indice : %d\n",i, tree->noeud[i]->b[0]->qmat_indice);  
    }
    for (i=1;i<2*tree->n_otu-1;i++){
      if (i!=n_otu) { // No need to do it for he root node
	//We compute equilibrium codon frequencies
	Compute_Codon_Equilibrium_Frequencies(tree->noeud[i]->cpi, tree->noeud[i]->theta, tree->mod);
	printf("after CCEF\n");
	//We compute qmatrices from these frequencies
	

	Update_Qmat_Codon (tree->mod, tree->mod->qmat_struct[tree->noeud[i]->b[0]->qmat_indice], tree->noeud[i]->cpi);
      }
    }
  }
  */

  /*fout = Openfile("FirstChars",input->out_file_open_mode);
  for (j=0;j<60;j++) {
    for (i=0;i<alldata->n_otu;i++) {
      fprintf (fout,"%c",alldata->c_seq[i]->state[j]);
    }
    fprintf(fout,"\n");
  }
  */

  Order_Tree_CSeq(tree,alldata);
  Make_Tree_4_Lk(tree,alldata,alldata->init_len);
  
  printf("Parameter Alpha of the Gamma distribution : %f\n",(float)tree->mod->alpha);
  
  /* fout = Openfile("FirstChars",input->out_file_open_mode);
  for (j=0;j<60;j++) {
    for (i=0;i<alldata->n_otu;i++) {
      fprintf (fout,"%c",alldata->c_seq[i]->state[j]);
    }
    fprintf(fout,"\n");
  }
  */
 
  //We compute the ts/tv ratio according to Kimura
  // Note : it does not always work (nan), so we correct and set it at 4 if we find nan.
  
  
  if (input->s_opt->opt_kappa) {
    freq=(double*)mCalloc(16, sizeof(double));
    for (i=0;i<input->mod->n_pos;i++) {
      input->mod->kappa[i]=0.0;
    }
    nbvrai=n_otu*(n_otu-1)/2;
    for(i=0;i<n_otu;i++){
      for(j=i+1;j<n_otu;j++){
	prov=alsurbet_kim(alldata->c_seq[i], alldata->c_seq[j], freq, alldata);
	if(prov<-0.5) nbvrai--; else input->mod->kappa[0]+=prov;
      }
    }
    free(freq);
    input->mod->kappa[0]/=nbvrai;
  }
  
  //We default kappa to 4.
  //tree->mod->kappa[0]=4.;
  //Was before if ((isnan(tree->mod->kappa[0]))||(isinf(tree->mod->kappa[0]))) {
  //changed in :

  //printf("Kimura's Ts/Tv : %f\n",tree->mod->kappa[0]);
  if ((isnan(tree->mod->kappa[0]))||(tree->mod->kappa[0]>10)||(tree->mod->kappa[0]<0)) {
    tree->mod->kappa[0]=4.;
    printf("Setting Ts/Tv to : %f\n", (float)tree->mod->kappa[0]);
  }else {
    printf("Initial Ts/Tv : %f\n", (float)tree->mod->kappa[0]);
  }

  for (i=1;i<input->mod->n_pos;i++) {
    input->mod->kappa[i]= tree->mod->kappa[0];
  }

  /*
  for (i=0;i<2*tree->n_otu-2;i++) {
    printf("branch %d : %d\n", i, tree->t_edges[i]->rearr);
    
  }
  */



  //If we want to estimate the variance of the root G+C content estimate
  if (input->estimate_variance){

    tree->mod->omega[0]=input->lower_GC_bound;
    //Initial Likelihood  
    Lk(tree, tree->data);
    printf("Initial Likelihood : %f\n", (float)tree->tot_loglk);
    
    fprintf(fp_lnf_out,"    %d  %d  %d\n\n\n",(int)(100*input->upper_GC_bound)-(int)(100*input->lower_GC_bound)+2,tree->data->init_len, tree->data->crunch_len);
    for (i=(int)(100*input->lower_GC_bound);i<=(int)(100*input->upper_GC_bound+1);i++) {
      //optimizations  
      tree->mod->omega[0]=(double)i/100.;
      printf("\n\n\t\t\t\t**********\n\nTrying a root GC of: %f\n",(float)tree->mod->omega[0]);
      if (input->changetree) { 

	NR_Tot_First(tree); 

	Simu(tree, 1000); 

	if (!input->quick) {
	  printf("\n\n. Last optimization step...");
	  NR_Tot(tree); 
	}
	else {
	  printf("\n\n. No Final optimization.");
	}
      }
      else NR_Tot(tree);
      if (i!=(int)(100*input->upper_GC_bound+1)) {
	Update_Pis(tree);
	//Output
	s = Write_Tree(tree);
	sEq=Write_Tree_Equilibrium(tree);
	
	fprintf(fp_tree_out,"%s\n",s);
	
	fprintf(fp_tree_out_Eq,"%s\n",sEq);
      }
      fprintf(fp_dist_out,"##################Tree number %d##############\n",i);
      fprintf(fp_comp_out,"##################Tree number %d##############\n",i);
      Reconstruct_Ancestral_Distributions(tree, fp_dist_out, fp_comp_out);
      Fill_lnf_File_iter(tree,fp_lnf_out, i+2-100*input->lower_GC_bound);
    }
  }
  else {
    
    //Initial Likelihood  
    Lk(tree, tree->data);
    printf("Initial Likelihood : %f\n", (float)tree->tot_loglk);
    
    //optimizations  
    if (input->changetree) {
      NR_Tot_First(tree);
      Simu(tree, 1000);
      if (!input->quick) {
	printf("\n\n. Last optimization step...");
	NR_Tot(tree); 
      }
      else {
	printf("\n\n. No Final optimization.");
      }
    }
    else NR_Tot(tree);
    Reconstruct_Ancestral_Distributions(tree, fp_dist_out, fp_comp_out);
  }
  



  /* for (i=0;i<2*tree->n_otu-2;i++) {
   printf("branche %d: %f\n", i, tree->t_edges[i]->rght->theta[0]);
   }*/
  Update_Pis(tree);
  //Output
  s = Write_Tree(tree);
  sEq=Write_Tree_Equilibrium(tree);
  
  
  fprintf(fp_tree_out,"%s\n",s);
  
  fprintf(fp_tree_out_Eq,"%s\n",sEq);
  
  Free(s);
  Free(sEq);
  
  Unconstraint_Lk(tree);
  
  time(&t_end);
  
  hour = div(t_end-t_beg,3600);
  min  = div(t_end-t_beg,60  );
  
  min.quot -= hour.quot*60;
  Lk(tree,tree->data);
  if (input->n_data_sets==1)
    Print_Fp_Out(fp_out, t_beg, t_end, tree, input, n_data_sets);
  else
    Print_Fp_Out_Lines(fp_out, t_beg, t_end, tree, input, n_data_sets);
  
 
  
  printf("\n\n. Time used %dh%dm%ds\n", hour.quot,min.quot,(int)(t_end-t_beg)%60);
  printf("\n%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n\n");
  printf("Final Lk : %3f\n", (float)tree->tot_loglk);
  
  if (tree->mod->n_pos==1) {
    printf("Ancestral G+C : %2f\n",(float)(tree->root->pi[0][1]+tree->root->pi[0][2]));
    printf("Ratio Ts/Tv : %f\n", (float)tree->mod->kappa[0]);
  }
  else {
    for (i=0;i<tree->mod->n_pos;i++) {
      printf("Sites %d : Ancestral G+C : %2f\n", i, (float)(tree->root->pi[i][1]+tree->root->pi[i][2]));
      printf("Sites %d : Ratio Ts/Tv : %f\n", i, (float)tree->mod->kappa[i]);
    }
  }
  printf("Parameter Alpha of the Gamma distribution : %f\n",(float)tree->mod->alpha);
  fprintf(fp_lk_out,"%f\n",(float)tree->tot_loglk);
  if (tree->mod->n_pos==1) {
    fprintf(fp_lk_out,"Ancestral G+C : %2f\n",(float)(tree->root->pi[0][1]+tree->root->pi[0][2]));
  }
  else {
     for (i=0;i<tree->mod->n_pos;i++) {
       fprintf(fp_lk_out, "Sites %d : Ancestral G+C : %2f\n", i, (float)(tree->root->pi[i][1]+tree->root->pi[i][2]));
     }
  }
  
  if (input->internalsequences) {
    Print_Tree_Text (tree,fp_seq_out);
  }




  if (!input->estimate_variance) {
    Fill_lnf_File(tree,fp_lnf_out);
  }  
  
  fclose(fp_lnf_out);
  //fclose(fp_seq_out);
  fclose(fp_lk_out);
  fclose(fp_out);
  
  fclose(fp_dist_out);
  fclose(fp_comp_out);
  
  Free_Tree_Lk(tree);
  
  
  
  Free_Cseq(alldata);
  
  Free_Model(mod);
  
  
  Free_Tree(tree);
  
  return 0;
}











//FAILED TRY TO DO NJ DISTANCES

/*  

  //If we do not want to keep the length of the tree, we compute them under the model
  //of Galtier and Gouy 1995. For that purpose, we use routines from nhml3. Then we use the 
  // Neighbor-Joining algorithm to fit the computed distances on the topology.
  
  //We first compute branch length using gg95
  //First we allocate the needed memory space
  dist=(double**)mCalloc(2*n_otu,sizeof(double*));
  for(i=0;i<2*n_otu;i++) {
    dist[i]=(double*)mCalloc(2*n_otu,sizeof(double));//tree->data->init_len,sizeof(char));
    for (j=0;j<2*n_otu;j++) {
      dist[i][j]=0.;
    }
  }
  //then we compute the distances
  for(i=0;i<n_otu;i++){
    for(j=i;j<n_otu;j++){
      if(i==j)  dist[i][j]=0.;
      else dist[i][j]=dist[j][i]=gg95(alldata->c_seq[i], alldata->c_seq[j], alldata, tree->mod->kappa);
      //printf("nomi : %s, nomj : %s,distij : %f\n", tree->data->c_seq[i]->name, tree->data->c_seq[j]->name,dist[i][j]);
    } 
  } 

  // Neighbor-Joining tree fitting
  n_Non_Agglo=tree->n_otu;
  NJ_Tree_Fit (tree->noeud[tree->n_otu], tree->noeud[tree->n_otu]->v[0], tree, dist, &n_Non_Agglo);
  NJ_Tree_Fit (tree->noeud[tree->n_otu], tree->noeud[tree->n_otu]->v[1], tree, dist, &n_Non_Agglo);
  
  tree->noeud[tree->n_otu]->b[0]->l = tree->noeud[tree->n_otu]->b[1]->l = dist[tree->noeud[tree->n_otu]->v[0]->num][tree->noeud[tree->n_otu]->v[1]->num]/2;
  for(i=0;i<2*tree->n_otu-2;i++) {
    if (tree->t_edges[i]->l<=0.0) {
      tree->t_edges[i]->l=0.000001;
    }
  }
  free(dist);
  
  Adapt_Length(tree);
 */
