/*

nhPhyml :  a program that builds phylogenetic trees from DNA alignments using the non-homogeneous, non-stationary model of evolution conceived by Galtier and Gouy (1998) and the algorithmic structure of phyml (Guindon and Gascuel, 2003).

Copyright (C) Bastien Boussau. jul. 2005 onward

All parts of  the source except where indicated  are distributed under
the GNU public licence.  See http://www.opensource.org for details.

*/


#include "utilities.h"
#include "ml.h"
#include "model.h"
#include "free.h"


/*********************************************************/

void Make_Tree_4_Lk(arbre *tree, allseq *alldata, int n_site)
{
  int i,j,k,curr_site;
 
  tree->n_pattern = tree->data->crunch_len/tree->mod->stepsize;
  if(!tree->tot_loglk_sorted) tree->tot_loglk_sorted = (double *)mCalloc(tree->n_pattern, sizeof(double));
  if(!tree->tot_dloglk)       tree->tot_dloglk       = (double *)mCalloc(2*tree->n_otu-2,sizeof(double));
  if(!tree->tot_d2loglk)      tree->tot_d2loglk      = (double *)mCalloc(2*tree->n_otu-2,sizeof(double));
  if(!tree->tot_dloglk_domega_sorted) tree->tot_dloglk_domega_sorted = (double *)mCalloc(tree->n_pattern, sizeof(double));
  if(!tree->tot_d2loglk_domega2_sorted) tree->tot_d2loglk_domega2_sorted = (double *)mCalloc(tree->n_pattern, sizeof(double));
  if(!tree->tot_dloglk_dtheta_sorted) tree->tot_dloglk_dtheta_sorted = (double *)mCalloc(tree->n_pattern, sizeof(double));
  if(!tree->tot_d2loglk_dtheta2_sorted) tree->tot_d2loglk_dtheta2_sorted = (double *)mCalloc(tree->n_pattern, sizeof(double));
  if(!tree->tot_dloglk_dtheta)       tree->tot_dloglk_dtheta       = (double *)mCalloc(2*tree->n_otu-2,sizeof(double));
  if(!tree->tot_d2loglk_dtheta2)      tree->tot_d2loglk_dtheta2      = (double *)mCalloc(2*tree->n_otu-2,sizeof(double));
  if(!tree->tot_dloglk_dtot_sorted) tree->tot_dloglk_dtot_sorted = (double *)mCalloc(tree->n_pattern, sizeof(double));
  if(!tree->tot_d2loglk_dtot2_sorted) tree->tot_d2loglk_dtot2_sorted = (double *)mCalloc(tree->n_pattern, sizeof(double));
  if(!tree->tot_dloglk_dfrac_sorted) tree->tot_dloglk_dfrac_sorted = (double *)mCalloc(tree->n_pattern, sizeof(double));
  if(!tree->tot_d2loglk_dfrac2_sorted) tree->tot_d2loglk_dfrac2_sorted = (double *)mCalloc(tree->n_pattern, sizeof(double));
  if(!tree->tot_dloglk_dkappa_sorted) tree->tot_dloglk_dkappa_sorted = (double *)mCalloc(tree->n_pattern, sizeof(double));
  if(!tree->tot_d2loglk_dkappa2_sorted) tree->tot_d2loglk_dkappa2_sorted = (double *)mCalloc(tree->n_pattern, sizeof(double));
if(!tree->tot_dloglk_dl_coef_sorted) tree->tot_dloglk_dl_coef_sorted = (double *)mCalloc(tree->n_pattern, sizeof(double));
  if(!tree->tot_d2loglk_dl_coef2_sorted) tree->tot_d2loglk_dl_coef2_sorted = (double *)mCalloc(tree->n_pattern, sizeof(double));

  if(!tree->t_best_l)         tree->t_best_l         = (double *)mCalloc(2*tree->n_otu-2,sizeof(double));
  if(!tree->site_lk)          tree->site_lk          = (double *)mCalloc(alldata->crunch_len,sizeof(double));
 if(!tree->true_site_lk)          tree->true_site_lk          = (double *)mCalloc(alldata->crunch_len,sizeof(double));
 
  tree->n_p_lk_alloc = 1;
  tree->root = tree->noeud[tree->n_otu]; // This is the true root

  DiscreteGamma (tree->mod->r_proba, tree->mod->rr, tree->mod->alpha,
		 tree->mod->alpha,tree->mod->n_catg,0);  


 
  For(i,2*tree->n_otu-2)// For all the edges in the rooted case
    {
      if (tree->t_edges[i]) 
	{tree->t_best_l[i] = tree->t_edges[i]->l;

	}
      Make_Edge_Lk(tree->t_edges[i]->left,
		   tree->t_edges[i]->rght,
		   tree); // initialisation only

    }

  For(i,2*tree->n_otu-1)// For all the nodes in the rooted case !
    {
      Make_Node_Lk(tree->noeud[i]);
    }
  Alloc_All_P_Lk(tree);



  Fors(curr_site,tree->data->crunch_len,tree->mod->stepsize)
    {
      For(i,tree->n_otu)
	{
	  //Initialisation :
	  For(j,tree->mod->ns) tree->noeud[i]->b[0]->p_lk_rght[tree->data->pos[curr_site/tree->mod->stepsize]][curr_site/tree->mod->stepsize][0][j] = 0.0;
	  if (tree->mod->ns<5){
	    Init_Tips_At_One_Site_Nucleotides(tree->data->c_seq[i]->state[curr_site],
					      &tree->noeud[i]->b[0]->p_lk_rght[tree->data->pos[curr_site]][curr_site][0]);
	  }
	  else {
	    Init_Tips_At_One_Site_Codon(tree->data->c_seq[i]->state[curr_site],
					tree->data->c_seq[i]->state[curr_site+1],
					tree->data->c_seq[i]->state[curr_site+2],
					&tree->noeud[i]->b[0]->p_lk_rght[tree->data->pos[curr_site/tree->mod->stepsize]][curr_site/tree->mod->stepsize][0],
					tree->mod->c_code,
					tree->mod->ns);
	  }
	  if((tree->noeud[i]->b[0]->p_lk_rght) && (tree->noeud[i]->b[0]->get_p_lk_rght))
	    {	    
	      for(j=1;j<tree->mod->n_catg;j++)
		{
		  For(k,tree->mod->ns)
		    {
		      tree->noeud[i]->b[0]->p_lk_rght[tree->data->pos[curr_site]][curr_site][j][k]=
		      tree->noeud[i]->b[0]->p_lk_rght[tree->data->pos[curr_site]][curr_site][0][k];
		    }

		}
	    }
	  else
	    {
	      for(j=1;j<tree->mod->n_catg;j++)
		For(k,tree->mod->ns)
		  tree->noeud[i]->b[0]->p_lk_rght[tree->data->pos[curr_site]][curr_site][j][k]=
		  tree->noeud[i]->b[0]->p_lk_rght[tree->data->pos[curr_site]][curr_site][0][k];
	    }

	}
    }

}

/*********************************************************/
void Init_Tips_At_One_Site_Nucleotides(char state, double **p_lk)
{
  /* Initialize a vector of partial likelihood at a tip of the tree.
     'state' is the nucleotide observed at this tip */

  switch(state){
  case 'A' : (*p_lk)[0]=1.; (*p_lk)[1]=(*p_lk)[2]=(*p_lk)[3]=.0;
    break;
  case 'C' : (*p_lk)[1]=1.; (*p_lk)[0]=(*p_lk)[2]=(*p_lk)[3]=.0;
    break;
  case 'G' : (*p_lk)[2]=1.; (*p_lk)[1]=(*p_lk)[0]=(*p_lk)[3]=.0;
    break;
  case 'T' : (*p_lk)[3]=1.; (*p_lk)[1]=(*p_lk)[2]=(*p_lk)[0]=.0;
    break;
  case 'U' : (*p_lk)[3]=1.; (*p_lk)[1]=(*p_lk)[2]=(*p_lk)[0]=.0;
    break;
  case 'M' : (*p_lk)[0]=(*p_lk)[1]=1.; (*p_lk)[2]=(*p_lk)[3]=.0;
    break;
  case 'R' : (*p_lk)[0]=(*p_lk)[2]=1.; (*p_lk)[1]=(*p_lk)[3]=.0;
    break;
  case 'W' : (*p_lk)[0]=(*p_lk)[3]=1.; (*p_lk)[1]=(*p_lk)[2]=.0;
    break;
  case 'S' : (*p_lk)[1]=(*p_lk)[2]=1.; (*p_lk)[0]=(*p_lk)[3]=.0;
    break;
  case 'Y' : (*p_lk)[1]=(*p_lk)[3]=1.; (*p_lk)[0]=(*p_lk)[2]=.0;
    break;
  case 'K' : (*p_lk)[2]=(*p_lk)[3]=1.; (*p_lk)[0]=(*p_lk)[1]=.0;
    break;
  case 'B' : (*p_lk)[1]=(*p_lk)[2]=(*p_lk)[3]=1.; (*p_lk)[0]=.0;
    break;
  case 'D' : (*p_lk)[0]=(*p_lk)[2]=(*p_lk)[3]=1.; (*p_lk)[1]=.0;
    break;
  case 'H' : (*p_lk)[0]=(*p_lk)[1]=(*p_lk)[3]=1.; (*p_lk)[2]=.0;
    break;
  case 'V' : (*p_lk)[0]=(*p_lk)[1]=(*p_lk)[2]=1.; (*p_lk)[3]=.0;
    break;
  case 'N' : case 'X' : case '?' : case 'O' : case '-' : 
    (*p_lk)[0]=(*p_lk)[1]=(*p_lk)[2]=(*p_lk)[3]=1.;break;
  default : 
    {
      printf("\n. Unknown character state : %c\n",state);
      Exit("\n. Init failed (check the data type)\n"); 
      break;
    }
  }
}

/*********************************************************/
double Return_Lk(arbre *tree)
{
  Lk(tree,tree->data);
  return tree->tot_loglk;
}

/*********************************************************/
void Lk(arbre *tree,allseq *alldata)
{
  int br,site,catg,pos;


  edge *b;
 
  b = NULL;

  For (pos, tree->mod->n_pos){
    tree->noeud[tree->n_otu]->pi[pos][1]=tree->noeud[tree->n_otu]->pi[pos][2]=tree->mod->omega[pos]/2;
    tree->noeud[tree->n_otu]->pi[pos][0]=tree->noeud[tree->n_otu]->pi[pos][3]=(1-tree->mod->omega[pos])/2;
  }

  DiscreteGamma (tree->mod->r_proba, tree->mod->rr, tree->mod->alpha,
		tree->mod->alpha,tree->mod->n_catg,0);  
 
 
  
  For(br,2*tree->n_otu-2) // One more branch in the rooted case 
    {
      For(site,tree->n_pattern)
	{
	  tree->t_edges[br]->sum_scale_f_rght[site] = .0;
	  tree->t_edges[br]->sum_scale_f_left[site] = .0;
	}

      if(tree->t_edges[br]->l < BL_MIN) tree->t_edges[br]->l = BL_MIN;
      b=tree->t_edges[br];
    
      For (pos, tree->mod->n_pos){
	For (catg,tree->mod->n_catg)
	  {
	    Computes_Substitution_Probabilities(b, tree, catg, pos);
	    /*	  Computes_Derivative_Substitution_Probabilities(b, tree, catg);
	      Computes_Second_Derivative_Substitution_Probabilities (b, tree, catg);*/
	  }
      }
    }


  Post_Order_Lk(tree->noeud[tree->n_otu],
		tree->noeud[tree->n_otu]->v[0],
		tree); // Rooted case!
  Post_Order_Lk(tree->noeud[tree->n_otu],
		tree->noeud[tree->n_otu]->v[1],
		tree); // Rooted case!

  if (tree->n_otu > LIM_SCALE) {
    /*  Pre_Order_Lk_Scale(tree->noeud[tree->n_otu],
		       tree->noeud[tree->n_otu]->v[0],
		       tree);// Rooted case!
    Pre_Order_Lk_Scale(tree->noeud[tree->n_otu],
		       tree->noeud[tree->n_otu]->v[1],
		       tree);// Rooted case!
    
    */  Pre_Order_Lk(tree->noeud[tree->n_otu],
		 tree->noeud[tree->n_otu]->v[0],
		 tree);// Rooted case!
    Pre_Order_Lk(tree->noeud[tree->n_otu],
		 tree->noeud[tree->n_otu]->v[1],
		 tree);// Rooted case!
    
  }
  else {
    Pre_Order_Lk(tree->noeud[tree->n_otu],
		 tree->noeud[tree->n_otu]->v[0],
		 tree);// Rooted case!
    Pre_Order_Lk(tree->noeud[tree->n_otu],
		 tree->noeud[tree->n_otu]->v[1],
		 tree);// Rooted case!
  }

  tree->tot_loglk = .0;    
  tree->curr_catg =  0;
  tree->curr_site =  0;
  

  For(site,tree->n_pattern)
    {
      tree->curr_site=site;
      Site_Lk(tree,alldata, tree->data->pos[site]);
    }

  qksort(tree->tot_loglk_sorted, 0, tree->n_pattern-1);

  tree->tot_loglk = .0;
  For(site, tree->n_pattern)
    tree->tot_loglk += tree->tot_loglk_sorted[site];


  For(br,2*tree->n_otu-2) // One more branch in the rooted case 
    {
      if(tree->t_edges[br]->get_p_lk_left) tree->t_edges[br]->ud_p_lk_left = 1;
      if(tree->t_edges[br]->get_p_lk_rght) tree->t_edges[br]->ud_p_lk_rght = 1;
    }

}



/*********************************************************/
// Pre and Post_Order were inverted in Guindon's implementation.
// We are interested in node "fils"
void Post_Order_Lk(node *pere, node *fils, arbre *tree) 
{

  int i,dir1,dir2,dir3;
  
  dir1 = dir2 = dir3 = -1;

  if(fils->tax) return;
  else
    {
      For(i,3)
	{
	  if(fils->v[i] != pere)
	    {
	     
	      Post_Order_Lk(fils,fils->v[i],tree); // Compute the two conditional likelihoods of L and R subtrees
	
	      if(dir1 < 0) dir1 = i; // find the two branches to study
	      else dir2 = i;
	    }
	  else dir3 = i; // v[dir3] is the father node
	}


      /*  (tree->n_otu > LIM_SCALE)?
	  (Get_All_Partial_Lk_Scale(tree,fils->b[dir3],pere,fils)):*/
	(Get_All_Partial_Lk(tree,fils->b[dir3],pere,fils));
	}
}

/*********************************************************/

void Pre_Order_Lk(node *pere, node *fils, arbre *tree)
{
  int i,j,dir1,dir2,dir3, catg, site, pos;
  edge  *b1, *b2;
  double ****plk1, ****plk2, p1, p2;
  int ancestor;
  dir1 = dir2 = dir3 = -1;
  pos=0;
 
  //If we are at the root
  if (pere==tree->noeud[tree->n_otu]) {
    if (fils==tree->noeud[tree->n_otu]->v[0]) { //If the taxon is the first root's neighbour
	For(catg,tree->mod->n_catg)
	  {	  
	    For(site,tree->n_pattern)
	      {
		pos=tree->data->pos[site];
		For(i,tree->mod->ns) /*sort sum terms ? No global effect*/
		  {              
		    tree->noeud[tree->n_otu]->b[0]->p_lk_left[pos][site][catg][i]=0.;
		    For(j,tree->mod->ns) {
		      tree->noeud[tree->n_otu]->b[0]->p_lk_left[pos][site][catg][i] += 
			tree->noeud[tree->n_otu]->pi[pos][i] *
			tree->noeud[tree->n_otu]->b[1]->Pij_rr[pos][catg][i][j] * 
			tree->noeud[tree->n_otu]->b[1]->p_lk_rght[pos][site][catg][j];
		    }
		    
		  }
		/*	if (site==25){
		  printf("racine pere %d, fils %d, length : %f, site %d : %f %f %f %f\n",tree->noeud[tree->n_otu]->num,fils->num, tree->noeud[tree->n_otu]->b[0]->l, site,log(tree->noeud[tree->n_otu]->b[0]->p_lk_left[pos][site][catg][0]),log(tree->noeud[tree->n_otu]->b[0]->p_lk_left[pos][site][catg][1]),log(tree->noeud[tree->n_otu]->b[0]->p_lk_left[pos][site][catg][2]), log(tree->noeud[tree->n_otu]->b[0]->p_lk_left[pos][site][catg][3]));
		  //		  printf("noeud %d, racine, site %d : %f\n",tree->noeud[tree->n_otu]->num,site,log(tree->noeud[tree->n_otu]->b[0]->p_lk_left[pos][site][catg][i]));
		}
		else if (site==26){
		  // printf("noeud %d, racine,site %d : %f\n",tree->noeud[tree->n_otu]->num,site,log(tree->noeud[tree->n_otu]->b[0]->p_lk_left[pos][site][catg][i]));
		}
		*/
	      }
	  }

    }
    else if (fils==tree->noeud[tree->n_otu]->v[1]) { //If the taxon is the second root's neighbour
	
      	For(catg,tree->mod->n_catg)
	  {	  
	    For(site,tree->n_pattern)
	      {
		pos=tree->data->pos[site];
		For(i,tree->mod->ns) /*sort sum terms ? No global effect*/
		  {     
		    tree->noeud[tree->n_otu]->b[1]->p_lk_left[pos][site][catg][i]=0;
		    For(j,tree->mod->ns) {
		      tree->noeud[tree->n_otu]->b[1]->p_lk_left[pos][site][catg][i] += 
			tree->noeud[tree->n_otu]->pi[pos][i] *
			tree->noeud[tree->n_otu]->b[0]->Pij_rr[pos][catg][i][j] * 
			tree->noeud[tree->n_otu]->b[0]->p_lk_rght[pos][site][catg][j];
		    }
		  }
		/*	if (site==25){
		  printf("racine pere %d, fils %d, length %f, site %d : %f %f %f %f\n",tree->noeud[tree->n_otu]->num,tree->noeud[tree->n_otu]->v[1]->num,  tree->noeud[tree->n_otu]->b[1]->l, site,log(tree->noeud[tree->n_otu]->b[1]->p_lk_left[pos][site][catg][0]),log(tree->noeud[tree->n_otu]->b[1]->p_lk_left[pos][site][catg][1]),log(tree->noeud[tree->n_otu]->b[1]->p_lk_left[pos][site][catg][2]), log(tree->noeud[tree->n_otu]->b[1]->p_lk_left[pos][site][catg][3]));
		  //		  printf("noeud %d, racine, site %d : %f\n",tree->noeud[tree->n_otu]->num,site,log(tree->noeud[tree->n_otu]->b[0]->p_lk_left[pos][site][catg][i]));
		}
		else if (site==26){
		  // printf("noeud %d, racine,site %d : %f\n",tree->noeud[tree->n_otu]->num,site,log(tree->noeud[tree->n_otu]->b[0]->p_lk_left[pos][site][catg][i]));
		}
		*/
	      }
	  }
      }

    if (fils->tax) {return;} // If a taxon, stop here
    else { //There is a subtree, we keep on going down
      For(i,3)
	{
	  if(fils->v[i] != pere)
	    {
	    
	      // we go down to work upon the two sons of the current node 
	      Pre_Order_Lk(fils,fils->v[i],tree);
	    }
	}
    }
  }


  //If we are not at the root, but at a leaf
  else if(fils->tax) 
    {
      //We 've got to know in which edge we go reverse and in which one we go in the sense of evolution
      //ancestor=1 : node "pere" is the ancestor of the edge 1
      //ancestor=2 : node "pere" is the ancestor of the edge 2
      ancestor=0;

      //these ifs help us find which branches are the ones concerned with our computing the upper conditional lk
      if (pere->b[0]->rght==fils) 
	{b1=pere->b[1];b2=pere->b[2];} 
      else if (pere->b[1]->rght==fils) 
	{b1=pere->b[0];b2=pere->b[2];} 
      else   {b1=pere->b[0];b2=pere->b[1];}
      if (pere==b1->left) 
	{plk1=b1->p_lk_rght;ancestor=1;} 
      else {plk1=b1->p_lk_left;}
      if (pere==b2->left) 
	{plk2=b2->p_lk_rght;ancestor=2;} 
      else {plk2=b2->p_lk_left;}

	  For(catg,tree->mod->n_catg)
	    {	  
	      For(site,tree->n_pattern)
		{
		  pos=tree->data->pos[site];
		  For(i,tree->mod->ns) /*sort sum terms ? No global effect*/
		    {                  
		    
		      p1=0;
		      p2=0;
		      For(j,tree->mod->ns) {
			if (ancestor==1) {
			  p1+=b1->Pij_rr[pos][catg][i][j] * plk1[pos][site][catg][j];
			  p2+=b2->Pij_rr[pos][catg][j][i] * plk2[pos][site][catg][j]; 
			}
			else if (ancestor==2) {
			  p1+=b1->Pij_rr[pos][catg][j][i] * plk1[pos][site][catg][j];
			  p2+=b2->Pij_rr[pos][catg][i][j] * plk2[pos][site][catg][j];
			}
		      }
		      fils->b[0]->p_lk_left[pos][site][catg][i]=p1*p2;
		      
		    }
		  /*
		  if (site==25){
		    printf("noeud %d, fils de %d, length %f, site %d : %f %f %f %f\n",fils->num,pere->num,fils->b[0]->l, site, log(fils->b[0]->p_lk_left[pos][site][catg][0]),log(fils->b[0]->p_lk_left[pos][site][catg][1]),log(fils->b[0]->p_lk_left[pos][site][catg][2]),log(fils->b[0]->p_lk_left[pos][site][catg][3]));
		  }
		  else if (site==26) {
		    //		    printf("noeud %d, fils de %d, site %d : %f ",fils->num,pere->num,site, log(fils->b[0]->p_lk_left[pos][site][catg][i]));
		  }
		  */
		}
	    }
	  // printf("\n");
    }
  else  // If not at a leaf
    {
      
      ancestor=0;
      if (pere->b[0]->rght==fils) {b1=pere->b[1];b2=pere->b[2];} else if 
	    (pere->b[1]->rght==fils) {b1=pere->b[0];b2=pere->b[2];} else   {b1=pere->b[0];b2=pere->b[1];}
      if (pere==b1->left) {plk1=b1->p_lk_rght;ancestor=1;} else {plk1=b1->p_lk_left;}
      if (pere==b2->left) {plk2=b2->p_lk_rght;ancestor=2;} else {plk2=b2->p_lk_left;}

	  For(catg,tree->mod->n_catg)
	    {	  
	      For(site,tree->n_pattern)
		{
		  pos=tree->data->pos[site];
		  For(i,tree->mod->ns) /*sort sum terms ? No global effect*/
		    {                  
		 
		      p1=0;
		      p2=0;
		      For(j,tree->mod->ns) {
			if (ancestor==1) {
			  p1+=b1->Pij_rr[pos][catg][i][j] * plk1[pos][site][catg][j];
			  p2+=b2->Pij_rr[pos][catg][j][i] * plk2[pos][site][catg][j]; 
			}
			else if (ancestor==2) {
			  p1+=b1->Pij_rr[pos][catg][j][i] * plk1[pos][site][catg][j];
			  p2+=b2->Pij_rr[pos][catg][i][j] * plk2[pos][site][catg][j];
			}
		      }
		      fils->b[0]->p_lk_left[pos][site][catg][i]=p1*p2;
		      
		    }
		  /*	  if (site==25){
		    printf("noeud %d, fils de %d, length %f, site %d : %f %f %f %f\n",fils->num,pere->num,fils->b[0]->l, site, log(fils->b[0]->p_lk_left[pos][site][catg][0]),log(fils->b[0]->p_lk_left[pos][site][catg][1]),log(fils->b[0]->p_lk_left[pos][site][catg][2]),log(fils->b[0]->p_lk_left[pos][site][catg][3]));
			//printf("noeud %d, fils de %d, site %d : %f ",fils->num,pere->num,site, log(fils->b[0]->p_lk_left[pos][site][catg][i]));
		      }
		      else if (site==26) {
			//	printf("noeud %d, fils de %d, site %d : %f ",fils->num,pere->num,site, log(fils->b[0]->p_lk_left[pos][site][catg][i]));
		      }
*/
		}
	    }
	  //	  printf("\n");
      For(i,3)
	{
	  if(fils->v[i] != pere)
	    {
	  
	      // with this call, we keep on computing upper conditional lks for underlying nodes 
	      Pre_Order_Lk(fils,fils->v[i],tree);
	    }
	}
    }
}
/*********************************************************/

void Pre_Order_Lk_Scale(node *pere, node *fils, arbre *tree)
{
  int i,j,dir1,dir2,dir3, catg, site, pos;
  edge  *b1, *b2;
  double ****plk1, ****plk2, p1, p2;
  int ancestor;
  double max_p_lk;
  double sum_scale_d1,sum_scale_d2;
  dir1 = dir2 = dir3 = -1;
  pos=0;
  
 

  sum_scale_d1=0.0;
  sum_scale_d2=0.0;
  max_p_lk = MDBL_MIN;

  //If we are at the root
  if (pere==tree->noeud[tree->n_otu]) {
    if (fils==tree->noeud[tree->n_otu]->v[0]) { //If the taxon is the first root's neighbour
      For(site,tree->n_pattern)
	{
	  pos=tree->data->pos[site];
	  max_p_lk = MDBL_MIN;
	  sum_scale_d1 = tree->noeud[tree->n_otu]->b[1]->sum_scale_f_rght[site];
	  For(catg,tree->mod->n_catg)
	    {
	      pos=tree->data->pos[site];
	      For(i,tree->mod->ns) /*sort sum terms ? No global effect*/
		{              
		  tree->noeud[tree->n_otu]->b[0]->p_lk_left[pos][site][catg][i]=0.;
		  For(j,tree->mod->ns) {
		    tree->noeud[tree->n_otu]->b[0]->p_lk_left[pos][site][catg][i] += 
		      tree->noeud[tree->n_otu]->pi[pos][i] *
		      tree->noeud[tree->n_otu]->b[1]->Pij_rr[pos][catg][i][j] * 
		      tree->noeud[tree->n_otu]->b[1]->p_lk_rght[pos][site][catg][j];
		  }
		  if((tree->noeud[tree->n_otu]->b[0]->p_lk_left[pos][site][catg][i] > max_p_lk)) max_p_lk = tree->noeud[tree->n_otu]->b[0]->p_lk_left[pos][site][catg][i];
		}
	    }
	  if(max_p_lk < LIM_SCALE_VAL)
	    {
	      For(catg,tree->mod->n_catg)
		{
		  For(i,tree->mod->ns)
		    {
		      tree->noeud[tree->n_otu]->b[0]->p_lk_left[pos][site][catg][i] /= max_p_lk;
		      
		      if( tree->noeud[tree->n_otu]->b[0]->p_lk_left[pos][site][catg][i] > MDBL_MAX)
			{
			  Exit("\n. Numerical underflow ! (send me an e-mail : boussau@biomserv.univ-lyon1.fr)\n");
			}
		    }
		}
	      
	      //We store the scaling factor
	      tree->noeud[tree->n_otu]->b[0]->sum_scale_f_left[site]=sum_scale_d1+log(max_p_lk);
	    }
	  if(max_p_lk > (1./LIM_SCALE_VAL))
	    {
	      For(catg,tree->mod->n_catg)
		{
		  For(i,tree->mod->ns)
		    {
		       tree->noeud[tree->n_otu]->b[0]->p_lk_left[pos][site][catg][i]/= max_p_lk;
		      
		      if(tree->noeud[tree->n_otu]->b[0]->p_lk_left[pos][site][catg][i] < MDBL_MIN)
			{
			  Exit("\n. Numerical overflow ! (send me an e-mail : boussau@biomserv.univ-lyon1.fr)\n");
			}
		    }
		}
	      //We store the scaling factor
	      tree->noeud[tree->n_otu]->b[0]->sum_scale_f_left[site]=sum_scale_d1+log(max_p_lk);
	    }
	}
    }

    else if (fils==tree->noeud[tree->n_otu]->v[1]) { //If the taxon is the second root's neighbour
      For(site,tree->n_pattern)
	{
	  max_p_lk = MDBL_MIN;
	  sum_scale_d1 = tree->noeud[tree->n_otu]->b[0]->sum_scale_f_rght[site];
	  For(catg,tree->mod->n_catg)
	    {	
	      pos=tree->data->pos[site];
	      For(i,tree->mod->ns) /*sort sum terms ? No global effect*/
		{     
		  tree->noeud[tree->n_otu]->b[1]->p_lk_left[pos][site][catg][i]=0;
		  For(j,tree->mod->ns) {
		    tree->noeud[tree->n_otu]->b[1]->p_lk_left[pos][site][catg][i] += 
		      tree->noeud[tree->n_otu]->pi[pos][i] *
		      tree->noeud[tree->n_otu]->b[0]->Pij_rr[pos][catg][i][j] * 
		      tree->noeud[tree->n_otu]->b[0]->p_lk_rght[pos][site][catg][j];
		  }
		  if((tree->noeud[tree->n_otu]->b[1]->p_lk_left[pos][site][catg][i] > max_p_lk)) max_p_lk = tree->noeud[tree->n_otu]->b[1]->p_lk_left[pos][site][catg][i];
		}
	    }
	  if(max_p_lk < LIM_SCALE_VAL)
	    {
	      For(catg,tree->mod->n_catg)
		{
		  For(i,tree->mod->ns)
		    {
		      tree->noeud[tree->n_otu]->b[1]->p_lk_left[pos][site][catg][i] /= max_p_lk;
		      
		      if( tree->noeud[tree->n_otu]->b[1]->p_lk_left[pos][site][catg][i] > MDBL_MAX)
			{
			  Exit("\n. Numerical underflow ! (send me an e-mail : boussau@biomserv.univ-lyon1.fr)\n");
			}
		    }
		}
	      
	      //We store the scaling factor
	      tree->noeud[tree->n_otu]->b[1]->sum_scale_f_left[site]=sum_scale_d1+log(max_p_lk);
	    }
	  if(max_p_lk > (1./LIM_SCALE_VAL))
	    {
	      For(catg,tree->mod->n_catg)
		{
		  For(i,tree->mod->ns)
		    {
		       tree->noeud[tree->n_otu]->b[1]->p_lk_left[pos][site][catg][i]/= max_p_lk;
		      
		      if(tree->noeud[tree->n_otu]->b[1]->p_lk_left[pos][site][catg][i] < MDBL_MIN)
			{
			  Exit("\n. Numerical overflow ! (send me an e-mail : boussau@biomserv.univ-lyon1.fr)\n");
			}
		    }
		}
	      //We store the scaling factor
	      tree->noeud[tree->n_otu]->b[1]->sum_scale_f_left[site]=sum_scale_d1+log(max_p_lk);
	    }
	}
    }

    if (fils->tax) {return;} // If a taxon, stop here
    else { //There is a subtree, we keep on going down
      For(i,3)
	{
	  if(fils->v[i] != pere)
	    {
	    
	      // we go down to work upon the two sons of the current node 
	      Pre_Order_Lk_Scale(fils,fils->v[i],tree);
	    }
	}
    }
  }


  //If we are not at the root, but at a leaf
  else if(fils->tax) 
    {
      //We 've got to know in which edge we go reverse and in which one we go in the sense of evolution
      //ancestor=1 : node "pere" is the ancestor of the edge 1
      //ancestor=2 : node "pere" is the ancestor of the edge 2
      ancestor=0;

      //these ifs help us find which branches are the ones concerned with our computing the upper conditional lk
      if (pere->b[0]->rght==fils) 
	{b1=pere->b[1];b2=pere->b[2];} 
      else if (pere->b[1]->rght==fils) 
	{b1=pere->b[0];b2=pere->b[2];} 
      else   {b1=pere->b[0];b2=pere->b[1];}
      if (pere==b1->left) 
	{plk1=b1->p_lk_rght;ancestor=1;} 
      else {plk1=b1->p_lk_left;}
      if (pere==b2->left) 
	{plk2=b2->p_lk_rght;ancestor=2;} 
      else {plk2=b2->p_lk_left;}
	 	  
      For(site,tree->n_pattern)
	{
	  sum_scale_d1 = pere->b[0]->sum_scale_f_left[site];
	  if (pere->b[1]->rght==fils) 
	    {sum_scale_d2 = pere->b[2]->sum_scale_f_rght[site];} 
	  else if (pere->b[2]->rght==fils)
	    {sum_scale_d2 = pere->b[1]->sum_scale_f_rght[site];}
	  else {Exit("Branch 0 does not lead to the father node ! (send me an e-mail : boussau@biomserv.univ-lyon1.fr)\n");}
	  fils->b[0]->sum_scale_f_left[site]=sum_scale_d1+sum_scale_d2;

	  max_p_lk = MDBL_MIN;
	  For(catg,tree->mod->n_catg)
	    {
	      pos=tree->data->pos[site];
	      For(i,tree->mod->ns) /*sort sum terms ? No global effect*/
		{                  
		  
		  p1=0;
		  p2=0;
		  For(j,tree->mod->ns) {
		    if (ancestor==1) {
		      p1+=b1->Pij_rr[pos][catg][i][j] * plk1[pos][site][catg][j];
		      p2+=b2->Pij_rr[pos][catg][j][i] * plk2[pos][site][catg][j]; 
		    }
		    else if (ancestor==2) {
		      p1+=b1->Pij_rr[pos][catg][j][i] * plk1[pos][site][catg][j];
		      p2+=b2->Pij_rr[pos][catg][i][j] * plk2[pos][site][catg][j];
		    }
		  }
		  fils->b[0]->p_lk_left[pos][site][catg][i]=p1*p2;
		  if((fils->b[0]->p_lk_left[pos][site][catg][i] > max_p_lk)) max_p_lk = fils->b[0]->p_lk_left[pos][site][catg][i];
		}
	    }
	  if(max_p_lk < LIM_SCALE_VAL)
	    {
	      For(catg,tree->mod->n_catg)
		{
		  For(i,tree->mod->ns)
		    {
		      fils->b[0]->p_lk_left[pos][site][catg][i] /= max_p_lk;
		      
		      if( fils->b[0]->p_lk_left[pos][site][catg][i] > MDBL_MAX)
			{
			  Exit("\n. Numerical underflow ! (send me an e-mail : boussau@biomserv.univ-lyon1.fr)\n");
			}
		    }
		}
	      
	      //We store the scaling factor
	      fils->b[0]->sum_scale_f_left[site]+=log(max_p_lk);
	    }
	  if(max_p_lk > (1./LIM_SCALE_VAL))
	    {
	      For(catg,tree->mod->n_catg)
		{
		  For(i,tree->mod->ns)
		    {
		      fils->b[0]->p_lk_left[pos][site][catg][i] /= max_p_lk;
		      
		      if(fils->b[0]->p_lk_left[pos][site][catg][i] < MDBL_MIN)
			{
			  Exit("\n. Numerical overflow ! (send me an e-mail : boussau@biomserv.univ-lyon1.fr)\n");
			}
		    }
		}
	      //We store the scaling factor
	      fils->b[0]->sum_scale_f_left[site]+=log(max_p_lk);
	    }
	}
    }

  else  // If not at a leaf
    {
      
      ancestor=0;
      if (pere->b[0]->rght==fils) {b1=pere->b[1];b2=pere->b[2];} else if 
	    (pere->b[1]->rght==fils) {b1=pere->b[0];b2=pere->b[2];} else   {b1=pere->b[0];b2=pere->b[1];}
      if (pere==b1->left) {plk1=b1->p_lk_rght;ancestor=1;} else {plk1=b1->p_lk_left;}
      if (pere==b2->left) {plk2=b2->p_lk_rght;ancestor=2;} else {plk2=b2->p_lk_left;}
      For(site,tree->n_pattern)
	{
	  For(catg,tree->mod->n_catg)
	    {	  
	      sum_scale_d1 = pere->b[0]->sum_scale_f_left[site];
	      if (pere->b[1]->rght==fils) 
		{sum_scale_d2 = pere->b[2]->sum_scale_f_rght[site];} 
	      else if (pere->b[2]->rght==fils)
		{sum_scale_d2 = pere->b[1]->sum_scale_f_rght[site];}
	      else {Exit("Branch 0 does not lead to the father node ! (send me an e-mail : boussau@biomserv.univ-lyon1.fr)\n");}
	      fils->b[0]->sum_scale_f_left[site]=sum_scale_d1+sum_scale_d2;

	      max_p_lk = MDBL_MIN;
	      pos=tree->data->pos[site];
	      For(i,tree->mod->ns) /*sort sum terms ? No global effect*/
		{                  
		  
		  p1=0;
		  p2=0;
		  For(j,tree->mod->ns) {
		    if (ancestor==1) {
		      p1+=b1->Pij_rr[pos][catg][i][j] * plk1[pos][site][catg][j];
		      p2+=b2->Pij_rr[pos][catg][j][i] * plk2[pos][site][catg][j]; 
		    }
		    else if (ancestor==2) {
		      p1+=b1->Pij_rr[pos][catg][j][i] * plk1[pos][site][catg][j];
		      p2+=b2->Pij_rr[pos][catg][i][j] * plk2[pos][site][catg][j];
		    }
		  }
		  fils->b[0]->p_lk_left[pos][site][catg][i]=p1*p2;
		  if((fils->b[0]->p_lk_left[pos][site][catg][i] > max_p_lk)) max_p_lk = fils->b[0]->p_lk_left[pos][site][catg][i];
		}
	    }

	  if(max_p_lk < LIM_SCALE_VAL)
	    // printf ("max_p_lk :%f",max_p_lk);
	    {
	      For(catg,tree->mod->n_catg)
		{
		  For(i,tree->mod->ns)
		    {
		      fils->b[0]->p_lk_left[pos][site][catg][i] /= max_p_lk;
		      
		      if( fils->b[0]->p_lk_left[pos][site][catg][i] > MDBL_MAX)
			{
			  Exit("\n. Numerical underflow ! (send me an e-mail : boussau@biomserv.univ-lyon1.fr)\n");
			}
		    }
		}
	      
	      //We store the scaling factor
	      fils->b[0]->sum_scale_f_left[site]+=log(max_p_lk);
	    }
	  if(max_p_lk > (1./LIM_SCALE_VAL))
	    {
	      For(catg,tree->mod->n_catg)
		{
		  For(i,tree->mod->ns)
		    {
		      fils->b[0]->p_lk_left[pos][site][catg][i] /= max_p_lk;
		      
		      if(fils->b[0]->p_lk_left[pos][site][catg][i] < MDBL_MIN)
			{
			  Exit("\n. Numerical overflow ! (send me an e-mail : boussau@biomserv.univ-lyon1.fr)\n");
			}
		    }
		}
	      //We store the scaling factor
	      fils->b[0]->sum_scale_f_left[site]+=log(max_p_lk);
	    }


	}

      For(i,3)
	{
	  if(fils->v[i] != pere)
	    {
	  
	      // with this call, we keep on computing upper conditional lks for underlying nodes 
	      Pre_Order_Lk_Scale(fils,fils->v[i],tree);
	    }
	}
    }
}


/*********************************************************/
void Get_All_Partial_Lk(arbre *tree, edge *b_fcus, node *a, node *d)
{
  int i,j, pos;
  double p1_lk1,p2_lk2;
  double ****p_lk,****p_lk_v1,****p_lk_v2;
  int catg,site;
  double ****Pij1,****Pij2;
  int ancestor=0;

  if(d->tax) {
    return;
  }
  else
    {
      int dir1,dir2;
      
      dir1=dir2=-1;
      For(i,3) if(d->v[i] != a) (dir1<0)?(dir1=i):(dir2=i); // find the two branches to study which define the subtree for the conditional likelihood

      if(b_fcus->l < BL_MIN) b_fcus->l = BL_MIN;

      p_lk = 
	(d == b_fcus->left)? // is d at the left or right tip of the edge under focus ? 
	(b_fcus->p_lk_left): // left is "towards the root"
	(b_fcus->p_lk_rght); // right is "towards the leaves"
    
     
      /*There comes the difference between a reversible and non reversible model :
	With a non reversible model, one needs to differenciate the left part of 
	the edge from its right part.
      */

      //If the node d is closer to the root than its first "son"
      if (d == d->b[dir1]->left) {
	p_lk_v1 = d->b[dir1]->p_lk_rght; //Then p_lk_v1 equals the conditional likelihood of the lower part of the edge
      }                                  
      else {
	if (d->b[dir1]->left==tree->noeud[tree->n_otu]) {
	  ancestor=-1;
	  return;
	  // Do nothing, we have already calculated the left conditional likelihood in 
	  // Pre_Order_Lk
	}
	
	// There we need to differenciate cases where the edge
	// dir1 leads to the root from cases where it doesn't : 
	// the pump needs to be 
	// started. If dir1 leads to the root, we must define 
	// b[dir1]->p_lk_left differently, using the other edge
	// steming from the root.
	
	p_lk_v1 = (d->b[dir1]->p_lk_left); // else p_lk_v1 equals the conditional likelihood of the upper part of the edge
	ancestor = 1;
      }
	
      // Same stuff for the other edge steming from d
      if (d == d->b[dir2]->left) {         
	p_lk_v2 = (d->b[dir2]->p_lk_rght); 
      }
      else {
	if (d->b[dir2]->left==tree->noeud[tree->n_otu]) {
	  // Do nothing, we have already calculated the left conditional likelihood in 
	  // Pre_Order_Lk
	  ancestor=-1;
	  return;
	}
	p_lk_v2 = (d->b[dir2]->p_lk_left);
	ancestor = 2;
      }
      
            
      Pij1 = d->b[dir1]->Pij_rr;
      Pij2 = d->b[dir2]->Pij_rr;
      
     	  For(catg,tree->mod->n_catg)
	    {	  
	      For(site,tree->n_pattern)
		{
		  pos=tree->data->pos[site];
		  For(i,tree->mod->ns) /*sort sum terms ? No global effect*/
		    {                  /*site i at node a*/
		      p1_lk1 = p2_lk2 = .0;
		      For(j,tree->mod->ns) // We sum for all possible states j at node d
			{
			  if (ancestor==1) {
			    p1_lk1 += (Pij1[pos][catg][j][i] * p_lk_v1[pos][site][catg][j]); // going backwards in time : from base j to i
			    p2_lk2 += (Pij2[pos][catg][i][j] * p_lk_v2[pos][site][catg][j]); 
			  }
			  else if (ancestor==2) {
			    p1_lk1 += (Pij1[pos][catg][i][j] * p_lk_v1[pos][site][catg][j]);
			    p2_lk2 += (Pij2[pos][catg][j][i] * p_lk_v2[pos][site][catg][j]); // going backwards in time : from base j to i
			  }
			  else { // inferior conditional likelihood : following the way evolution goes
			    
			    p1_lk1 += (Pij1[pos][catg][i][j] * p_lk_v1[pos][site][catg][j]);
			    p2_lk2 += (Pij2[pos][catg][i][j] * p_lk_v2[pos][site][catg][j]); 
			  }
			}
		      
		      p_lk[pos][site][catg][i] = p1_lk1*p2_lk2;
		      if(p_lk[pos][site][catg][i] < MDBL_MIN) 
			{
			  //We do not print the warning message !!!!!
			  //printf("WARNING : scaling is required at site %d\n",site);
			}
		    }
		}
	    }
	}
    
}

/*********************************************************/

void Get_All_Partial_Lk_Scale(arbre *tree, edge *b_fcus, node *a, node *d)
{

  int i,j, pos;
  double p1_lk1,p2_lk2;
  double ****p_lk,****p_lk_v1,****p_lk_v2;
  int catg,site;
  double ****Pij1,****Pij2;
  double max_p_lk;
  double sum_scale_d1,sum_scale_d2;
  int ancestor=0;
  p1_lk1 = p2_lk2 = .0;
  if(d->tax) return;
  else
    {
      int dir1,dir2;
      
      dir1=dir2=-1;
      For(i,3) if(d->v[i] != a) (dir1<0)?(dir1=i):(dir2=i);

      if(b_fcus->l < BL_MIN) b_fcus->l = BL_MIN;

      p_lk = 
	(d == b_fcus->left)?
	(b_fcus->p_lk_left):
	(b_fcus->p_lk_rght);

      //There comes the difference between a reversible and non reversible model :
	//With a non reversible model, one needs to differenciate the left part of 
	//the edge from its right part.
      
      if (d == d->b[dir1]->left) {
	p_lk_v1 = d->b[dir1]->p_lk_rght; // We are interested in the node which is at the other  
      }                                   // tip of the edges coming from d.
      else {
	if (d->b[dir1]->left==tree->noeud[tree->n_otu]) {
	  ancestor=-1;
	  return;
	  //Do nothing, we can't calculate the root's left and right 
	  //conditional likelihoods yet : we will do that later 

	}
	p_lk_v1 = (d->b[dir1]->p_lk_left);
	ancestor = 1;
      }
	
	
      if (d == d->b[dir2]->left) {         // We are interested in the node which is at the other
	p_lk_v2 = (d->b[dir2]->p_lk_rght); // tip of the edges coming from d.
      }
      else {
	if (d->b[dir2]->left==tree->noeud[tree->n_otu]) {
	  //Do nothing, we can't calculate the root's left and right 
	  //conditional likelihoods yet : we will do that later

	  ancestor=-1;
	  return;
	  
	}
	p_lk_v2 = (d->b[dir2]->p_lk_left);
	ancestor = 2;
      }

      Pij1 = d->b[dir1]->Pij_rr;
      Pij2 = d->b[dir2]->Pij_rr;

    	  For(site,tree->n_pattern)
	    {	  
	      pos=tree->data->pos[site];
	      sum_scale_d1 = sum_scale_d2 = .0;
	      
	      (d == d->b[dir1]->left)?
		(sum_scale_d1 = d->b[dir1]->sum_scale_f_rght[site]):
		(sum_scale_d1 = d->b[dir1]->sum_scale_f_left[site]);
	      
	      (d == d->b[dir2]->left)?
		(sum_scale_d2 = d->b[dir2]->sum_scale_f_rght[site]):
		(sum_scale_d2 = d->b[dir2]->sum_scale_f_left[site]);
	      
	      (d == b_fcus->left)?
		(b_fcus->sum_scale_f_left[site] = sum_scale_d1 + sum_scale_d2):
		(b_fcus->sum_scale_f_rght[site] = sum_scale_d1 + sum_scale_d2);
	      
	      
	      max_p_lk = MDBL_MIN;
	      
	      
	      For(catg,tree->mod->n_catg)
		{
		  For(i,tree->mod->ns) //sort sum terms ? No global effect
		    {
		      p1_lk1 = p2_lk2 = .0;
		      For(j,tree->mod->ns)
			{
			  if (ancestor == 1 ) {
			    p1_lk1 += (Pij1[pos][catg][j][i] * p_lk_v1[pos][site][catg][j]); // going backwards in time : from base j to i
			    p2_lk2 += (Pij2[pos][catg][i][j] * p_lk_v2[pos][site][catg][j]); 
			  }
			  else if (ancestor==2) {
			    p1_lk1 += (Pij1[pos][catg][i][j] * p_lk_v1[pos][site][catg][j]);
			    p2_lk2 += (Pij2[pos][catg][j][i] * p_lk_v2[pos][site][catg][j]); // going backwards in time : from base j to i
			  }
			  else { // inferior condiional likelihood : following the way evolution goes
			    p1_lk1 += (Pij1[pos][catg][i][j] * p_lk_v1[pos][site][catg][j]);
			    p2_lk2 += (Pij2[pos][catg][i][j] * p_lk_v2[pos][site][catg][j]); 
			  }
			}
		 	      
		      p_lk[pos][site][catg][i]= p1_lk1*p2_lk2;
		      
		      
		      if(p_lk[pos][site][catg][i] > max_p_lk) { 
			max_p_lk = p_lk[pos][site][catg][i];
		      }
		      
		    }
		}
	      
	      
	      //Here we take care of the scaling by dividing by max_p_lk the obtained conditional probabilities if too small	      

	      if(max_p_lk < LIM_SCALE_VAL)
		{
		  For(catg,tree->mod->n_catg)
		    {
		      For(i,tree->mod->ns)
			{
			  p_lk[pos][site][catg][i] /= max_p_lk;
			  
			  if(p_lk[pos][site][catg][i] > MDBL_MAX)
			    {
			      Exit("\n. Numerical underflow ! (send me an e-mail : boussau@biomserv.univ-lyon1.fr)\n");
			    }
			}
		    }
		  
		  /*  (d == b_fcus->left)?
		      (b_fcus->sum_scale_f_left[site] += log(max_p_lk)):*/
		    (b_fcus->sum_scale_f_rght[site] += log(max_p_lk));
		}
	      
	      if(max_p_lk > (1./LIM_SCALE_VAL))
		{
		  For(catg,tree->mod->n_catg)
		    {
		      For(i,tree->mod->ns)
			{
			  p_lk[pos][site][catg][i] /= max_p_lk;
			  
			  if(p_lk[pos][site][catg][i] < MDBL_MIN)
			    {
			      Exit("\n. Numerical overflow ! (send me an e-mail : boussau@biomserv.univ-lyon1.fr)\n");
			    }
			}
		    }
		  /*
		  (d == b_fcus->left)?
		  (b_fcus->sum_scale_f_left[site] += log(max_p_lk)):*/
		    (b_fcus->sum_scale_f_rght[site] += log(max_p_lk));
		}
	    }
	
    }
}

/********************************************************/
/*
Computes the likelihood and its derivatives at one site.
Here in Phyml was defined as eroot the edge that is linked to leaf 0.
It should change in the non-reversible model : 
Phyml's Site_Lk uses the "pulley principle" by placing 
the root wherever is needed. A non-reversible model makes it 
impossible to use the pulley principle : we have once again to 
start from the true root.
*/


void Site_Lk(arbre *tree, allseq *alldata, int pos)
{
  int i,catg,j,k,l,m;
  double *site_dlk=NULL, *site_d2lk=NULL;
  double log_site_lk,site_lk,aux;
  //  edge *eroot;
  int is_ambigu;
  int  site = tree->curr_site;
  if(!alldata->wght[site]) return;

  if(tree->mod->s_opt->opt_bl)
    {
      site_dlk = (double *)mCalloc(2*tree->n_otu-2,sizeof(double));
      site_d2lk = (double *)mCalloc(2*tree->n_otu-2,sizeof(double));
    }
 

 
  is_ambigu = alldata->ambigu[site];
  
 
  
  if(tree->mod->s_opt->opt_bl)
    For(j,2*tree->n_otu-2) site_dlk[j] = site_d2lk[j] = 0.0;


  log_site_lk = site_lk = .0;      
  For(catg,tree->mod->n_catg)
    {
     
	  For(k,tree->mod->ns) //sort sum terms ? No global effect
	    {
	      For(l,tree->mod->ns)
		{
		  For(i,tree->mod->ns)
		    {
		      site_lk += 
			tree->mod->r_proba[catg] * // Rate probability
			tree->noeud[tree->n_otu]->pi[pos][k] *  // Proportion of nucleotide k at root
			tree->noeud[tree->n_otu]->b[0]->p_lk_rght[pos][site][catg][l] *   // Inferior Partial Likelihood
			tree->noeud[tree->n_otu]->b[0]->Pij_rr[pos][catg][k][l] *         // Substitution probability
			tree->noeud[tree->n_otu]->b[1]->p_lk_rght[pos][site][catg][i] *   // Inferior Partial Likelihood
			tree->noeud[tree->n_otu]->b[1]->Pij_rr[pos][catg][k][i]    ;     // Substitution probability

			    
		    }
		}
	    }

         

      if(tree->mod->s_opt->opt_bl)
	{
	  For(k,2*tree->n_otu-2)
	    {  
	      tree->t_edges[k]->site_dlk_rr[catg] = .0;
	      tree->t_edges[k]->site_d2lk_rr[catg] = .0;

	      For(l,tree->mod->ns) //sort sum terms ? No global effect
		{
		  For(m,tree->mod->ns)
		    {
		      //Pulley principle was applying here, in Phyml
		      tree->t_edges[k]->site_dlk_rr[catg] += 
			tree->noeud[tree->n_otu]->pi[pos][k] *
	
			tree->t_edges[k]->p_lk_left[pos][site][catg][l]*
			tree->t_edges[k]->p_lk_rght[pos][site][catg][m]*
			tree->t_edges[k]->dPij_rr[pos][catg][l][m];
		      

		      tree->t_edges[k]->site_d2lk_rr[catg] += 
			tree->noeud[tree->n_otu]->pi[pos][k] *
			tree->t_edges[k]->p_lk_left[pos][site][catg][l]*
			tree->t_edges[k]->p_lk_rght[pos][site][catg][m]*
			tree->t_edges[k]->d2Pij_rr[pos][catg][l][m];
		    }
		}

	      site_dlk[k]  += tree->t_edges[k]->site_dlk_rr[catg]*
		              tree->mod->r_proba[catg];
	      
	      site_d2lk[k] += tree->t_edges[k]->site_d2lk_rr[catg]*
		              tree->mod->r_proba[catg];
	    }
	}   
    }

  
  // log_site_lk = log(site_lk)       + tree->noeud[tree->n_otu]->b[0]->sum_scale_f_left[site] + tree->noeud[tree->n_otu]->b[0]->sum_scale_f_rght[site];
log_site_lk = log(site_lk) + tree->noeud[tree->n_otu]->b[1]->sum_scale_f_rght[site] + tree->noeud[tree->n_otu]->b[0]->sum_scale_f_rght[site];

  if(log_site_lk < -MDBL_MAX)
    {
      printf("%d %f %f %f %f\n",
	     site,
	     log_site_lk,
	     tree->mod->alpha,
	     tree->noeud[tree->n_otu]->b[0]->sum_scale_f_left[site],
	     tree->noeud[tree->n_otu]->b[0]->sum_scale_f_rght[site]);
    
      Exit("\nlog_site_lk < -MDBL_MAX\n");
    }


  tree->site_lk[site] = log_site_lk;
  tree->true_site_lk[site] = site_lk;
  
  if(tree->mod->s_opt->opt_bl)
    {
      For(k,2*tree->n_otu-2)
	{
	  aux = exp(tree->t_edges[k]->sum_scale_f_rght[site]+
		    tree->t_edges[k]->sum_scale_f_left[site]);
	  site_dlk[k]  *= aux;
	  site_d2lk[k] *= aux;
	  
	  tree->t_edges[k]->site_dlk = site_dlk[k];
	  tree->t_edges[k]->site_d2lk = site_d2lk[k];	  
	}
    }
  
  tree->tot_loglk_sorted[site] = alldata->wght[site]*log_site_lk;


  if((tree->mod->s_opt->opt_bl) &&
     (fabs(site_lk = exp(log_site_lk)) > sqrt(MDBL_MIN)))
    {
      For(k,2*tree->n_otu-2)
	{
	  aux = site_dlk[k] / site_lk;
	  tree->tot_dloglk[k]  += alldata->wght[site] 
	                       *  aux;
	  tree->tot_d2loglk[k] += alldata->wght[site]
	                       * (site_d2lk[k]/site_lk
			       -  aux*aux);
	}
    }

  if(tree->mod->s_opt->opt_bl)
    {
      Free(site_dlk);
      Free(site_d2lk);
      }
}

/*********************************************************/
double Lk_At_Given_Edge(arbre *tree, edge *b_fcus)
{
  int site,catg,pos,k,l,edge_num, ns= tree->mod->ns, i;
  double site_lk,log_site_lk,site_dlk,site_d2lk,aux;
 
  edge_num = b_fcus->num;
  tree->tot_loglk = log_site_lk = .0;
  tree->tot_dloglk[edge_num] = tree->tot_d2loglk[edge_num] = .0;

  if(b_fcus->l < BL_MIN) b_fcus->l = BL_MIN;


  For(pos,tree->mod->n_pos){
    For(catg,tree->mod->n_catg)
      {
	
	Computes_Substitution_Probabilities(b_fcus, tree, catg, pos);
	if(tree->mod->s_opt->opt_bl)
	  {
	    Computes_Derivative_Substitution_Probabilities (b_fcus, tree, catg, pos);
	    Computes_Second_Derivative_Substitution_Probabilities (b_fcus, tree, catg, pos);
	  }
      }
  }
  
  For(site,tree->n_pattern)
    {
      pos=tree->data->pos[site];
      if(tree->data->wght[site])
      	{
	  log_site_lk = site_lk = .0;
	  /* see equation (2) in phyml_tech_doc.pdf */

	  if ((b_fcus==tree->noeud[tree->n_otu]->b[0])||(b_fcus==tree->noeud[tree->n_otu]->b[1])) {
	    For(catg,tree->mod->n_catg)
	      {
		For(k,ns) /*sort sum terms ? No global effect*/
		  {
		    For(l,ns)
		      {
		
			site_lk += 
			  tree->mod->r_proba[catg] * // Rate probability
		
			  b_fcus->p_lk_left[pos][site][catg][k] *   // Left Partial Likelihood
			  b_fcus->Pij_rr[pos][catg][k][l] *         // Substitution probability
			  b_fcus->p_lk_rght[pos][site][catg][l];
		
		      }
		  }
	      }
	  }
	  else {
	    For(catg,tree->mod->n_catg)
	      {
		For(k,ns) /*sort sum terms ? No global effect*/
		  {
		    For(l,ns)
		      {
		
			  site_lk += 
			    tree->mod->r_proba[catg] * // Rate probability
			    b_fcus->p_lk_left[pos][site][catg][k] *   // Left Partial Likelihood
			    b_fcus->Pij_rr[pos][catg][k][l] *         // Substitution probability
			    b_fcus->p_lk_rght[pos][site][catg][l];
			
		      }
		  }
	      }
	    
	  }
	  /* code 2.3 begin*//* see equations in phyml_tech_doc.pdf */
	  /* compute log_site_lk */
	  
	  log_site_lk += log(site_lk) 
	    + b_fcus->sum_scale_f_left[site] + b_fcus->sum_scale_f_rght[site];
	  
	  
	  /* code 2.3 end*/

	      
	  
	  if(log_site_lk < -MDBL_MAX) Exit("\nlog_site_lk < -MDBL_MAX\n");

	  tree->site_lk[site] = log_site_lk;
	  tree->true_site_lk[site] = site_lk;
	 
	  tree->tot_loglk_sorted[site] = /* code 2.3 */
	    tree->data->wght[site]*log_site_lk;
	}
    }
  /* code 2.3 begin*/
  /* sort and add numbers from smallest to biggest */
  qksort(tree->tot_loglk_sorted, 0, tree->n_pattern-1);
  For(k, tree->n_pattern) tree->tot_loglk += tree->tot_loglk_sorted[k];
 
  /* code 2.3 end*/
  if(tree->mod->s_opt->opt_bl)
    {
   
      For(site,tree->n_pattern)
	{
	  pos=tree->data->pos[site];
	  if(tree->data->wght[site])
	    {
	      site_dlk = site_d2lk = .0;      
	      b_fcus->site_dlk = b_fcus->site_d2lk = .0; 
	      



	      if (b_fcus==tree->noeud[tree->n_otu]->b[0]) {
		For(catg,tree->mod->n_catg)
		  {
		    For(k,ns) /*sort sum terms ? No global effect*/
		      {
			For(i, ns) 
			  {
			    For(l,ns)
			      {
				aux = tree->mod->r_proba[catg] *
		
				  b_fcus->left->pi[pos][k] *  // Proportion of nucleotide k
				  b_fcus->left->b[1]->Pij_rr[pos][catg][k][i] *
				  b_fcus->left->b[1]->p_lk_rght[pos][site][catg][i];
							
				site_dlk +=
				  aux *
				  b_fcus->p_lk_rght[pos][site][catg][l] *
				  b_fcus->dPij_rr[pos][catg][k][l];
				
				site_d2lk +=
				  aux *
				  b_fcus->p_lk_rght[pos][site][catg][l] *
				  b_fcus->d2Pij_rr[pos][catg][k][l];
			      }
			  }
		      }
		  }
	      }

	      else if (b_fcus==tree->noeud[tree->n_otu]->b[1]) {
		For(catg,tree->mod->n_catg)
		  {
		    For(k,ns) /*sort sum terms ? No global effect*/
		      {
			For(i, ns) 
			  {
			    For(l,ns)
			      {
				aux = tree->mod->r_proba[catg] *
		
				  b_fcus->left->pi[tree->data->pos[site]][k] *  // Proportion of nucleotide k
				  b_fcus->left->b[0]->Pij_rr[pos][catg][k][i] *
				  b_fcus->left->b[0]->p_lk_rght[pos][site][catg][i];
			
				
				site_dlk +=
				  aux *
				  b_fcus->p_lk_rght[pos][site][catg][l] *
				  b_fcus->dPij_rr[pos][catg][k][l];
				
				site_d2lk +=
				  aux *
				  b_fcus->p_lk_rght[pos][site][catg][l] *
				  b_fcus->d2Pij_rr[pos][catg][k][l];
			      }
			  }
		      }
		  }
	      }



	      else {
		For(catg,tree->mod->n_catg)
		  {
		  
			For(i, ns) 
			  {
			    For(l,ns)
			      {
				aux = tree->mod->r_proba[catg] *
				  b_fcus->p_lk_left[pos][site][catg][i];
							
				site_dlk +=
				  aux *
				  b_fcus->p_lk_rght[pos][site][catg][l] *
				  b_fcus->dPij_rr[pos][catg][i][l];
				
				site_d2lk +=
				  aux *
				  b_fcus->p_lk_rght[pos][site][catg][l] *
				  b_fcus->d2Pij_rr[pos][catg][i][l];
				
			      }
			  }
		
		  }
		
	      }
	      if(tree->n_otu > LIM_SCALE)
		{
		  aux = exp(b_fcus->sum_scale_f_rght[site]+
			    b_fcus->sum_scale_f_left[site]);
		  site_dlk  *= aux;
		  site_d2lk *= aux;
		}

	      b_fcus->site_dlk = site_dlk;
	      b_fcus->site_d2lk = site_d2lk;
	    
	      
	        if(fabs(exp(log_site_lk)) > sqrt(MDBL_MIN))
		  {
		  if (tree->site_lk[site]==0.0) {} else {
		    aux = site_dlk /exp(tree->site_lk[site]);
		
		    tree->tot_dloglk[edge_num]  += tree->data->wght[site] 
		      * aux;
		    tree->tot_d2loglk[edge_num] += tree->data->wght[site] 
		      * (site_d2lk/ exp(tree->site_lk[site]) 
			 - aux*aux);
		  }
		
	    }
	    }

	}
    }

 
  return tree->tot_loglk;
}
  
/*********************************************************/
/*
Called by Update_Lk_At_Given_Edge when we need to update the 
likelihood of a tree when exploring the topological space.
*/
void Update_P_Lk(arbre *tree, edge *b_fcus, node *n)
{
/*  
           |
	   |<- b_fcus
	   |
	   n
          / \
       	 /   \
       	/     \
*/
// We needed to change it when the node we consider is the root
  int k,l, i, j, ancestor=0;
  int site, catg, pos;
  double ****p_lk, ****p_lk_v1, ****p_lk_v2;
  double **Pij1, **Pij2;
  double *n_scale_f, *d1_scale_f, *d2_scale_f;
  double p1_lk1,p2_lk2;
  double max_p_lk;
  edge *b1, *b2;
   
  b1 = b2  = NULL;
  p_lk = p_lk_v1 = p_lk_v2 = NULL;
  max_p_lk = MDBL_MIN;

  //If the edge under focus is at the root
  if (n==tree->noeud[tree->n_otu]) {
    if (b_fcus==n->b[0]) { //If it is root's branch 0
      For(catg,tree->mod->n_catg)
	{	  
	  For(site,tree->n_pattern)
	    {
	      pos=tree->data->pos[site];
	      For(i,tree->mod->ns) /*sort sum terms ? No global effect*/
		{                  
		  b_fcus->p_lk_left[pos][site][catg][i]=0.;
		  For(j,tree->mod->ns) {
		    b_fcus->p_lk_left[pos][site][catg][i] += 
		      tree->noeud[tree->n_otu]->pi[tree->data->pos[site]][i] *
		      tree->noeud[tree->n_otu]->b[1]->Pij_rr[pos][catg][i][j] * 
		      tree->noeud[tree->n_otu]->b[1]->p_lk_rght[pos][site][catg][j];
		  }
		}
	    }
	}
      
    }
    else { //If it is root's branch 1
      For(catg,tree->mod->n_catg)
	{	  
	  For(site,tree->n_pattern)
	    {
	      pos=tree->data->pos[site];
	      For(i,tree->mod->ns) /*sort sum terms ? No global effect*/
		{                  
		  b_fcus->p_lk_left[pos][site][catg][i]=0.;
		  For(j,tree->mod->ns) {
		    b_fcus->p_lk_left[pos][site][catg][i] += 
		      tree->noeud[tree->n_otu]->pi[tree->data->pos[site]][i] *
		      tree->noeud[tree->n_otu]->b[0]->Pij_rr[pos][catg][i][j] * 
		      tree->noeud[tree->n_otu]->b[0]->p_lk_rght[pos][site][catg][j];
		      }
		}
	    }
	}
      
    }
    return;
  }
  


  //If n is not the root, 
  //but is at the left end of b_fcus

  else if(n == b_fcus->left)
    {
      p_lk = b_fcus->p_lk_left;
      
      p_lk_v1 = 
      (n == n->b[b_fcus->l_v1]->left)?
      (n->b[b_fcus->l_v1]->p_lk_rght):
      (n->b[b_fcus->l_v1]->p_lk_left);

      p_lk_v2 = 
      (n == n->b[b_fcus->l_v2]->left)?
      (n->b[b_fcus->l_v2]->p_lk_rght):
      (n->b[b_fcus->l_v2]->p_lk_left);
      ancestor=(n == n->b[b_fcus->l_v1]->left)?1:2;

      n_scale_f = b_fcus->sum_scale_f_left;

      d1_scale_f = 
      (n == n->b[b_fcus->l_v1]->left)?
      (n->b[b_fcus->l_v1]->sum_scale_f_rght):
      (n->b[b_fcus->l_v1]->sum_scale_f_left);

      d2_scale_f = 
      (n == n->b[b_fcus->l_v2]->left)?
      (n->b[b_fcus->l_v2]->sum_scale_f_rght):
      (n->b[b_fcus->l_v2]->sum_scale_f_left);
    
      b_fcus->get_p_lk_left = 1;
      b_fcus->ud_p_lk_left  = 1;
      b1 = n->b[b_fcus->l_v1];
      b2 = n->b[b_fcus->l_v2];
      if(b_fcus->l < BL_MIN) b_fcus->l = BL_MIN;
      if(tree->n_otu <= LIM_SCALE)
	{
	  /* NO SCALE */
	  if(ancestor==1) {
	    For(site,tree->n_pattern)
	      {
		pos=tree->data->pos[site];
		if(tree->data->wght[site])
		  {
		    For(catg,tree->mod->n_catg)
		      {
			Pij1 = b1->Pij_rr[pos][catg];
			Pij2 = b2->Pij_rr[pos][catg];
			For(k,tree->mod->ns) /*sort sum terms ? No global effect*/
			  {
			    p1_lk1 = p2_lk2 = .0;
			    
			    For(l,tree->mod->ns)
			      {
				p1_lk1 += Pij1[l][k] * p_lk_v1[pos][site][catg][l];
				p2_lk2 += Pij2[k][l] * p_lk_v2[pos][site][catg][l];
			      }
			    p_lk[pos][site][catg][k] = p1_lk1 * p2_lk2;
			  }
		      }
		  }
	      }
	  }
	  else {
	    For(site,tree->n_pattern)
	      {
		pos=tree->data->pos[site];
		if(tree->data->wght[site])
		  {
		    For(catg,tree->mod->n_catg)
		      {
			Pij1 = b1->Pij_rr[pos][catg];
			Pij2 = b2->Pij_rr[pos][catg];
			For(k,tree->mod->ns) /*sort sum terms ? No global effect*/
			  {
			    p1_lk1 = p2_lk2 = .0;
			    
			    For(l,tree->mod->ns)
			      {
				p1_lk1 += Pij1[k][l] * p_lk_v1[pos][site][catg][l];
				p2_lk2 += Pij2[l][k] * p_lk_v2[pos][site][catg][l];
			      }
			    p_lk[pos][site][catg][k] = p1_lk1 * p2_lk2;
			  }
		      }
		  }
	      }
	  }
	}
      else
	{
	  /* SCALE */
	  if(ancestor==1) {
	    For(site,tree->n_pattern)
	      {
		pos=tree->data->pos[site];
		if(tree->data->wght[site])
		  {
		    For(catg,tree->mod->n_catg)
		      {
			Pij1 = b1->Pij_rr[pos][catg];
			Pij2 = b2->Pij_rr[pos][catg];
			if(!catg) 
			  {
			    n_scale_f[site] = d1_scale_f[site] + d2_scale_f[site];
			    max_p_lk = -MDBL_MAX;
			  }
			For(k,tree->mod->ns) /*sort sum terms ? No global effect*/
			  {
			    p_lk[pos][site][catg][k] = .0;
			    p1_lk1 = p2_lk2     = .0;
			    For(l,tree->mod->ns)
			      {
				p1_lk1 += Pij1[l][k] * p_lk_v1[pos][site][catg][l];
				p2_lk2 += Pij2[k][l] * p_lk_v2[pos][site][catg][l];
			      }
			    p_lk[pos][site][catg][k] = p1_lk1 * p2_lk2;
			    if((p_lk[pos][site][catg][k] > max_p_lk)) max_p_lk = p_lk[pos][site][catg][k];
			  }
		      }
		    if(max_p_lk < LIM_SCALE_VAL)
		      {
			For(catg,tree->mod->n_catg)
			  {
			    For(k,tree->mod->ns) 
			      {
				p_lk[pos][site][catg][k] /= max_p_lk;
				
				if(p_lk[pos][site][catg][k] > MDBL_MAX)
				  {
				    Exit("\n. Numerical overflow ! (send me an e-mail : boussau@biomserv.univ-lyon1.fr)\n");
				  }			  
			      }
			  }
			n_scale_f[site] += log(max_p_lk);
		      }
		    if(max_p_lk > (1./LIM_SCALE_VAL))
		      {
			For(catg,tree->mod->n_catg)
			  {
			    For(k,tree->mod->ns) 
			      {
				p_lk[pos][site][catg][k] /= max_p_lk;
				if(p_lk[pos][site][catg][k] < MDBL_MIN)
				  {
				    printf("\n. Numerical underflow ! (send me an e-mail : boussau@biomserv.univ-lyon1.fr)\n");
				  }			  
			      }
			  }
			n_scale_f[site] += log(max_p_lk);
		      }
		  }
	      }
	  }
	  else {
	    For(site,tree->n_pattern)
	      {
		pos=tree->data->pos[site];
		if(tree->data->wght[site])
		  {
		    For(catg,tree->mod->n_catg)
		      {
			Pij1 = b1->Pij_rr[pos][catg];
			Pij2 = b2->Pij_rr[pos][catg];
			if(!catg) 
			  {
			    n_scale_f[site] = d1_scale_f[site] + d2_scale_f[site];
			    max_p_lk = -MDBL_MAX;
			  }
			For(k,tree->mod->ns) /*sort sum terms ? No global effect*/
			  {
			    p_lk[pos][site][catg][k] = .0;
			    p1_lk1 = p2_lk2     = .0;
			    For(l,tree->mod->ns)
			      {
				p1_lk1 += Pij1[k][l] * p_lk_v1[pos][site][catg][l];
				p2_lk2 += Pij2[l][k] * p_lk_v2[pos][site][catg][l];
			      }
			    p_lk[pos][site][catg][k] = p1_lk1 * p2_lk2;
			    if((p_lk[pos][site][catg][k] > max_p_lk)) max_p_lk = p_lk[pos][site][catg][k];
			  }
		      }
		    if(max_p_lk < LIM_SCALE_VAL)
		      {
			For(catg,tree->mod->n_catg)
			  {
			    For(k,tree->mod->ns) 
			      {
				p_lk[pos][site][catg][k] /= max_p_lk;
				
				if(p_lk[pos][site][catg][k] > MDBL_MAX)
				  {
				    Exit("\n. Numerical overflow ! (send me an e-mail : s.guindon@auckland.ac.nz)\n");
				  }			  
			      }
			  }
			n_scale_f[site] += log(max_p_lk);
		      }
		    if(max_p_lk > (1./LIM_SCALE_VAL))
		      {
			For(catg,tree->mod->n_catg)
			  {
			    For(k,tree->mod->ns) 
			      {
				p_lk[pos][site][catg][k] /= max_p_lk;
				if(p_lk[pos][site][catg][k] < MDBL_MIN)
				  {
				    printf("\n. Numerical underflow ! (send me an e-mail : s.guindon@auckland.ac.nz)\n");
				  }			  
			      }
			  }
			n_scale_f[site] += log(max_p_lk);
		      }
		  }
	      }
	  }
	}
    }


  //or is at the right end of b_fcus
  else
    {
      p_lk = b_fcus->p_lk_rght;
      
      p_lk_v1 = 
      (n == n->b[b_fcus->r_v1]->left)?
      (n->b[b_fcus->r_v1]->p_lk_rght):
      (n->b[b_fcus->r_v1]->p_lk_left);

      p_lk_v2 = 
      (n == n->b[b_fcus->r_v2]->left)?
      (n->b[b_fcus->r_v2]->p_lk_rght):
      (n->b[b_fcus->r_v2]->p_lk_left);

      n_scale_f = b_fcus->sum_scale_f_rght;

      d1_scale_f = 
      (n == n->b[b_fcus->r_v1]->left)?
      (n->b[b_fcus->r_v1]->sum_scale_f_rght):
      (n->b[b_fcus->r_v1]->sum_scale_f_left);

      d2_scale_f = 
      (n == n->b[b_fcus->r_v2]->left)?
      (n->b[b_fcus->r_v2]->sum_scale_f_rght):
      (n->b[b_fcus->r_v2]->sum_scale_f_left);

      b_fcus->get_p_lk_rght = 1;
      b_fcus->ud_p_lk_rght  = 1;
      b1 = n->b[b_fcus->r_v1];
      b2 = n->b[b_fcus->r_v2];

      if(b_fcus->l < BL_MIN) b_fcus->l = BL_MIN;
      if(tree->n_otu <= LIM_SCALE)
	{
	  For(site,tree->n_pattern)
	    {
	      pos=tree->data->pos[site];
	      if(tree->data->wght[site])
		{
		  For(catg,tree->mod->n_catg)
		    {
		      
		      Pij1 = b1->Pij_rr[pos][catg];
		      Pij2 = b2->Pij_rr[pos][catg];
		      
		      
		      For(k,tree->mod->ns) /*sort sum terms ? No global effect*/
			{
			  p1_lk1 = p2_lk2 = .0;
			  
			  For(l,tree->mod->ns)
			    {
			      p1_lk1 += Pij1[k][l] * p_lk_v1[pos][site][catg][l];
			      p2_lk2 += Pij2[k][l] * p_lk_v2[pos][site][catg][l];
			    }
			  p_lk[pos][site][catg][k] = p1_lk1 * p2_lk2;
			}
		    }
		}
	    }
	}
      else
	{
	  /* SCALE */
	  
	  For(site,tree->n_pattern)
	    {
	      pos=tree->data->pos[site];
	      if(tree->data->wght[site])
		{
		  For(catg,tree->mod->n_catg)
		    {
		      
		      Pij1 = b1->Pij_rr[pos][catg];
		      Pij2 = b2->Pij_rr[pos][catg];
		      
		      if(!catg) 
			{
			  n_scale_f[site] = d1_scale_f[site] + d2_scale_f[site];
			  max_p_lk = -MDBL_MAX;
			}
		      
		      For(k,tree->mod->ns) /*sort sum terms ? No global effect*/
			{
			  p_lk[pos][site][catg][k] = .0;
			  p1_lk1 = p2_lk2     = .0;
			  
			  For(l,tree->mod->ns)
			    {
			      p1_lk1 += Pij1[k][l] * p_lk_v1[pos][site][catg][l];
			      p2_lk2 += Pij2[k][l] * p_lk_v2[pos][site][catg][l];
			    }
			  
			  p_lk[pos][site][catg][k] = p1_lk1 * p2_lk2;
			  
			  
			  if((p_lk[pos][site][catg][k] > max_p_lk)) max_p_lk = p_lk[pos][site][catg][k];
		      
			}
		    }
		  
		  if(max_p_lk < LIM_SCALE_VAL)
		    {
		      For(catg,tree->mod->n_catg)
			{
			  For(k,tree->mod->ns) 
			    {
			      p_lk[pos][site][catg][k] /= max_p_lk;
			      
			      if(p_lk[pos][site][catg][k] > MDBL_MAX)
				{
				  Exit("\n. Numerical overflow ! (send me an e-mail : s.guindon@auckland.ac.nz)\n");
				}			  
			    }
			}
		      n_scale_f[site] += log(max_p_lk);
		    }
		  
		  if(max_p_lk > (1./LIM_SCALE_VAL))
		    {
		      For(catg,tree->mod->n_catg)
			{
			  For(k,tree->mod->ns) 
			    {
			      p_lk[pos][site][catg][k] /= max_p_lk;
			      if(p_lk[pos][site][catg][k] < MDBL_MIN)
				{
				  printf("\n. Numerical underflow ! (send me an e-mail : s.guindon@auckland.ac.nz)\n");
				}			  
			    }
			}
		      n_scale_f[site] += log(max_p_lk);
		    }
		}
	    }
	}
    }

  if(tree->n_otu <= LIM_SCALE)
    {
      /* NO SCALE */

      For(site,tree->n_pattern)
	{
	  pos=tree->data->pos[site];
	  if(tree->data->wght[site])
	    {
	      For(catg,tree->mod->n_catg)
		{
		  
		  Pij1 = b1->Pij_rr[pos][catg];
		  Pij2 = b2->Pij_rr[pos][catg];
		  
		  
		  For(k,tree->mod->ns) /*sort sum terms ? No global effect*/
		    {
		      p1_lk1 = p2_lk2 = .0;
		      
		      For(l,tree->mod->ns)
			{
			  p1_lk1 += Pij1[k][l] * p_lk_v1[pos][site][catg][l];
			  p2_lk2 += Pij2[k][l] * p_lk_v2[pos][site][catg][l];
			}
		      p_lk[pos][site][catg][k] = p1_lk1 * p2_lk2;
		    }
		}
	    }
	}
    }
  else
    {
      /* SCALE */

      For(site,tree->n_pattern)
	{
	  pos=tree->data->pos[site];
	  if(tree->data->wght[site])
	    {
	      For(catg,tree->mod->n_catg)
		{
		  
		  Pij1 = b1->Pij_rr[pos][catg];
		  Pij2 = b2->Pij_rr[pos][catg];
		  
		  if(!catg) 
		    {
		      n_scale_f[site] = d1_scale_f[site] + d2_scale_f[site];
		      max_p_lk = -MDBL_MAX;
		    }
		  
		  For(k,tree->mod->ns) /*sort sum terms ? No global effect*/
		    {
		      p_lk[pos][site][catg][k] = .0;
		      p1_lk1 = p2_lk2     = .0;
		      		      
		      For(l,tree->mod->ns)
			{
			  p1_lk1 += Pij1[k][l] * p_lk_v1[pos][site][catg][l];
			  p2_lk2 += Pij2[k][l] * p_lk_v2[pos][site][catg][l];
			}

		      p_lk[pos][site][catg][k] = p1_lk1 * p2_lk2;
		      

		      if((p_lk[pos][site][catg][k] > max_p_lk)) max_p_lk = p_lk[pos][site][catg][k];
		      
		    }
		}
	         
	      if(max_p_lk < LIM_SCALE_VAL)
		{
		  For(catg,tree->mod->n_catg)
		    {
		      For(k,tree->mod->ns) 
			{
			  p_lk[pos][site][catg][k] /= max_p_lk;
			  
			  if(p_lk[pos][site][catg][k] > MDBL_MAX)
			    {
			      Exit("\n. Numerical overflow ! (send me an e-mail : s.guindon@auckland.ac.nz)\n");
			    }			  
			}
		    }
		  n_scale_f[site] += log(max_p_lk);
		}
	      
	      if(max_p_lk > (1./LIM_SCALE_VAL))
		{
		  For(catg,tree->mod->n_catg)
		    {
		      For(k,tree->mod->ns) 
			{
			  p_lk[pos][site][catg][k] /= max_p_lk;
			  if(p_lk[pos][site][catg][k] < MDBL_MIN)
			    {
			      printf("\n. Numerical underflow ! (send me an e-mail : s.guindon@auckland.ac.nz)\n");
			    }			  
			}
		    }
		  n_scale_f[site] += log(max_p_lk);
		}
	    }
	}
    }
}

/*********************************************************/
void Unconstraint_Lk(arbre *tree)
{
  int i;

  tree->unconstraint_lk = .0;

  For(i,tree->data->crunch_len)
    {
      tree->unconstraint_lk += 
	tree->data->wght[i]*log(tree->data->wght[i]);
    }
  tree->unconstraint_lk -= 
    tree->data->init_len*log(tree->data->init_len);
}

/*********************************************************/

void Site_Lk_Simple (arbre *tree, allseq *alldata) {

  int  catg, k, l, i, pos, site = tree->curr_site;
  double log_site_lk,site_lk;
  
 log_site_lk = site_lk = .0;      
 pos=tree->data->pos[site];
 
 For(catg,tree->mod->n_catg)
   {
     For(k,tree->mod->ns) //sort sum terms ? No global effect
       {
	 For(l,tree->mod->ns)
	   {
	     For(i,tree->mod->ns)
		    {
		      site_lk += 
			tree->mod->r_proba[catg] * // Rate probability
			tree->noeud[tree->n_otu]->pi[pos][k] *  // Proportion at root of nucleotide k
			tree->noeud[tree->n_otu]->b[0]->p_lk_rght[pos][site][catg][l] *   // Inferior Partial Likelihood
			tree->noeud[tree->n_otu]->b[0]->Pij_rr[pos][catg][k][l] *         // Substitution probability
			tree->noeud[tree->n_otu]->b[1]->p_lk_rght[pos][site][catg][i] *   // Inferior Partial Likelihood
			tree->noeud[tree->n_otu]->b[1]->Pij_rr[pos][catg][k][i]    ;     // Substitution probability
		      if(tree->noeud[tree->n_otu]->pi[pos][k] <0) {printf(" k : %i, tree->noeud[tree->n_otu]->pi[k] : %f", k,tree->noeud[tree->n_otu]->pi[pos][k]);}
		   
		    }
		}
	    }
}

  
  log_site_lk = log(site_lk) 
    + tree->noeud[tree->n_otu]->b[0]->sum_scale_f_left[site] + tree->noeud[tree->n_otu]->b[0]->sum_scale_f_rght[site];
  // printf("tree->noeud[tree->n_otu]->b[0]->sum_scale_f_left[site] : %f\n", tree->noeud[tree->n_otu]->b[0]->sum_scale_f_left[site]);
  if(log_site_lk < -MDBL_MAX)
    {
      printf("%d %f %f %f %f\n",
	     site,
	     log_site_lk,
	     tree->mod->alpha,
	     tree->noeud[tree->n_otu]->b[0]->sum_scale_f_left[site],
	     tree->noeud[tree->n_otu]->b[0]->sum_scale_f_rght[site]);
	     //	     tree->mod->pinvar);
      Exit("\nlog_site_lk < -MDBL_MAX\n");
    }


  tree->site_lk[site] = log_site_lk;
  tree->true_site_lk[site] = site_lk;
  tree->tot_loglk_sorted[site] = alldata->wght[site]*log_site_lk;
}

/*********************************************************/

void Lk_At_Root(arbre *tree,allseq *alldata)
{
  int site,br,catg, pos;
  edge *b;

  for (pos=0;pos<tree->mod->n_pos; pos++){
    tree->noeud[tree->n_otu]->pi[pos][1]=tree->noeud[tree->n_otu]->pi[pos][2]=tree->mod->omega[pos]/2.;
    tree->noeud[tree->n_otu]->pi[pos][0]=tree->noeud[tree->n_otu]->pi[pos][3]=(1-tree->mod->omega[pos])/2.;
  }

  DiscreteGamma (tree->mod->r_proba, tree->mod->rr, tree->mod->alpha,
		 tree->mod->alpha,tree->mod->n_catg,0);  

  For(br,2*tree->n_otu-2) // One more branch in the rooted case 
    {
      For(site,tree->n_pattern)
	{
	  tree->t_edges[br]->sum_scale_f_rght[site] = .0;
	  tree->t_edges[br]->sum_scale_f_left[site] = .0;
	}

      if(tree->t_edges[br]->l < BL_MIN) tree->t_edges[br]->l = BL_MIN;
      b=tree->t_edges[br];

      For(pos, tree->mod->n_pos){
	For(catg,tree->mod->n_catg)
	  {
	    
	    Computes_Substitution_Probabilities(b, tree, catg, pos);
	  }
      }
    }




  For(site,tree->n_pattern)
    {
      tree->curr_site=site;
 
      Site_Lk_Simple(tree,alldata);
    }
  qksort(tree->tot_loglk_sorted, 0, tree->n_pattern-1);
  
  tree->tot_loglk = .0;
  For(site, tree->n_pattern)
    tree->tot_loglk += tree->tot_loglk_sorted[site];

}
/*********************************************************/


double Return_Lk_At_Root(arbre *tree)
{
  
  Lk_At_Root(tree,tree->data);
  return tree->tot_loglk;
}

/*********************************************************/

void Init_Tips_At_One_Site_Codon(char state1, char state2, char state3, double **p_lk, code *c_code, int ns)
{
  /* Initialize a vector of partial likelihood at a tip of the tree.
     'state' is the codon observed at this tip */

  double *p_lk_state1, *p_lk_state2, *p_lk_state3;
  int i,j,k,l;
  int num_codon;

  p_lk_state1 = (double *)mCalloc(4,sizeof(double));
  p_lk_state2 = (double *)mCalloc(4,sizeof(double));
  p_lk_state3 = (double *)mCalloc(4,sizeof(double));

  Init_Tips_At_One_Site_Nucleotides(state1,&p_lk_state1);
  Init_Tips_At_One_Site_Nucleotides(state2,&p_lk_state2);
  Init_Tips_At_One_Site_Nucleotides(state3,&p_lk_state3);

  For(i,4)
    {
      if(p_lk_state1[i]) 
	{
	  For(j,4)
	    {
	      if(p_lk_state2[j])
		{
		  For(k,4)
		    {
		      if(p_lk_state3[k])
			{
			  num_codon = i*16+j*4+k;
			  if(c_code->from_64_2_61[num_codon] > -1) 
			    {
			      l = 0;
			      do
				{
				  (*p_lk)[c_code->from_64_2_61[num_codon]+
					  l*c_code->n_sense_c] = 1.0;
				  l++;
				}
			      while(l*c_code->n_sense_c < ns);
			    }
			  else
			    {
			      if(
				 (!Is_Ambigu(&state1,0,1)) &&
				 (!Is_Ambigu(&state2,0,1)) &&
				 (!Is_Ambigu(&state3,0,1))
				 )
				{
				  int i;
				  printf("\n. WARNING : terminating codon replaced by unknown state\n");
				  l = 0;
				  do
				    {
				      For(i,61) (*p_lk)[i+l*c_code->n_sense_c] = 1.0;
				      l++;
				    }
				  while(l*c_code->n_sense_c < ns);
				}
			    }
			}
		    }
		}
	    }
	}
    }
  Free(p_lk_state1);
  Free(p_lk_state2);
  Free(p_lk_state3);
}

/*********************************************************/
