/*

nhPhyml :  a program that builds phylogenetic trees from DNA alignments using the non-homogeneous, non-stationary model of evolution conceived by Galtier and Gouy (1998) and the algorithmic structure of phyml (Guindon and Gascuel, 2003).

Copyright (C) Bastien Boussau. jul. 2005 onward

All parts of  the source except where indicated  are distributed under
the GNU public licence.  See http://www.opensource.org for details.

*/

#include "utilities.h"
#include "model.h"
#include "free.h"
#include "eigen.h"
#include "eigenmb.h"



model *Init_Model(allseq *data, option *input)
{
  model *mod;
  int i, pos;
  int ns;

  mod = input->mod;
  
  mod->s_opt = input->s_opt;
  

  mod->stepsize = input->stepsize;
   mod->unity=1.0;

  if(!data) mod->seq_len = input->seq_len;
  else      mod->seq_len = data->init_len;

  //  mod->invar  = input->invar;
  //  mod->pinvar = input->pinvar;
  
  if (input->model) {
    mod->n_pos=3;
  }
  else {
    mod->n_pos=1;
  }
  
  mod->n_catg = input->ncatg;

   
  mod->r_proba  = (double *)mCalloc(mod->n_catg,sizeof(double));
  mod->rr       = (double *)mCalloc(mod->n_catg,sizeof(double));
  mod->kappa = (double *)mCalloc(mod->n_pos,sizeof(double));
  mod->omega = (double *)mCalloc(mod->n_pos,sizeof(double));
  mod->kappa_old = (double *)mCalloc(mod->n_pos,sizeof(double));
  mod->omega_old = (double *)mCalloc(mod->n_pos,sizeof(double));
  mod->l_coef = (double *)mCalloc(mod->n_pos,sizeof(double));
  mod->ThetaValues = (double *)mCalloc(mod->NumEqFreq,sizeof(double));
  
  Make_Thetas(mod);

  /*
  for(i=0;i<mod->NumEqFreq;i++) {
    printf("theta %d : %f\n", i, mod->ThetaValues[i]); 
  }
  */

  mod->alpha    = input->alpha;
  // mod->omega=0.5;
  if (input->estimate_variance) {
     for (pos=0;pos<mod->n_pos;pos++) {
       mod->omega[pos]=input->lower_GC_bound;
       mod->l_coef[pos] = 1.0;
     }
  }
  else {

    for (pos=0;pos<mod->n_pos;pos++) {
      mod->omega[pos]=data->b_frq[1]+data->b_frq[2];
      mod->l_coef[pos] = 1.0;
      //    printf("Omega position %d: %f\n",pos, mod->omega[pos]);
    }
  }
  //mod->rr_param = (double *)mCalloc(5,sizeof(double));

  //  For(i,5) mod->rr_param[i] = 1.0;

  For(i,mod->n_catg) 
    {
      mod->rr[i]      = 1.0;
      mod->r_proba[i] = 1.0;
    }

  //  if(!input->invar) For(i,data->crunch_len) data->invar[i] = 0;


  //mod->ns = mod->datatype = -1;
  ns = 4;
  mod->ns = ns;
  // mod->datatype = 1;                // Pas sr


  // if(!mod->datatype)
  // { 
     
  if (!input->nt_or_cd) {
    /* init for nucleotides */
    ns = 4;
    mod->ns = ns;

    mod->R = 0.5;
    mod->Y = 0.5;
    //  mod->lambda = 1.;
    for (pos=0;pos<mod->n_pos;pos++) {
      mod->kappa[pos] = input->titv;
      if (input->estimate_variance) {
	mod->omega[pos]=input->lower_GC_bound;
      }
      else{
	mod->omega[pos]=data->b_frq[1]+data->b_frq[2];
      }
      // printf("Kappa position %d: %f\n",pos, mod->kappa[pos]);
    }

  
    // } 
  }
  else {
    /*init for codons : only one kappa for the moment*/
    ns = mod->c_code->n_sense_c;
    mod->ns = ns;
    mod->kappa[0] = input->titv;
    Get_Codon_Freqs(data,mod);
   
     
    //The following line fills model pi values and is useful if we want to use the observed frequencies for codon equilibrium frequencies.
    For(i,ns) mod->pi[i] = data->c_frq[i];
    
    //let's allocate the space necessary for codon equilibrium frequencies

    mod->update_eigen  = 1;
    // mod->analytical    = 0; We know that for nucleotides, it's analytical, and for codons, it's not
    /*
      Shouldn't be useful for we do not allow changes in dn/ds ratio
    Init_T_Omega(mod);
    
    Init_T_Omega_Proba(mod);
    
    We can't call this function here as it depends upon an equilibrium GC (a node)
    Update_Qmat_Codon(mod,mod->qmat_struct);
    */
  }
	  


	  mod->alpha_old = mod->alpha; 
	  for (pos=0;pos<mod->n_pos;pos++) {
	    mod->omega_old[pos] = mod->omega[pos];
	    mod->kappa_old[pos] = mod->kappa[pos];
	  }
	  mod->omega_changed = 1;
  //  mod->pinvar_old = mod->pinvar;


  // free(dr);free(di);free(space);
  return mod;
}

/*********************************************************/

int PMat_Numeric(double l, model *mod, double **Pij, qmat *qmat_struct)
{
  /* Compute change probabilities using eigen vectors and eigen values */

  int n;
  int i, j, k;
  double *U,*V,*R;
  double *expt;
  double *uexpt;
  int dim1,dim2,neg;


  dim2 = mod->ns*mod->ns;
  dim1 = mod->ns;

  n = mod->ns;

  expt  = (double *)mCalloc(n,  sizeof(double));
  uexpt = (double *)mCalloc(n*n,sizeof(double));


  U = qmat_struct->u_mat      ;// + dim2*qmat_struct->curr_qmat_cat;
  V = qmat_struct->v_mat      ;// + dim2*qmat_struct->curr_qmat_cat;
  R = qmat_struct->expD_mr_vct;// + dim1*qmat_struct->curr_qmat_cat;


  /* Compute pow(exp(D/mr),l) into mat_eDmrl */
  For(k,n)  expt[k] = pow(R[k],l);

  /* Multiply Vr*pow(exp(D/mr),l)*Vi into Pij */
  For(i,n) For(k,n) uexpt[i*n+k] = U[i*n+k] * expt[k];
  
  neg = 0;
  For(i,n)
    {      
      For(j,n)
	{
	  Pij[i][j] = .0;
	  
	  For(k,n) 
	    Pij[i][j] += uexpt[i*n+k] * V[k*n+j];

	  if(Pij[i][j] < MDBL_MIN)
	    {
	      Pij[i][j] = 1.E-30;
	    }
	  if(Pij[i][j] < 0.0)
	    {
	      if(Pij[i][j] < 0.0-1.E-5) neg = 1;
	      Pij[i][j] = 1.E-60;
	    }
	}
    }


  Free(expt); Free(uexpt);
  
  if(neg) 
    {
      printf("neg encountered  l=%f kappa=%f",// mr=%f\n, qmatcat = %d\n",
	     l,
	     mod->kappa[0]);
	     // mod->mr);
      //,qmat_struct->curr_qmat_cat);
      return 0;
    }
  return 1;
}


/*********************************************************/

void Update_Eigen(int dim, double *Q, double *U, double *V, double *Root)
{
    /* Call the eigenvalue decomposition routine with correction if
       round-off errors occured when inverting the instantaneous rate 
       matrix 
    */
    
    int i;
    int result;
    double *di,*space,*Q_buff;
    double scalar;
    int n_iter,*spaceint;
    double **Q_buff_sq, **U_sq;
    
    
    di        = (double *)mCalloc(dim,sizeof(double));
    space     = (double *)mCalloc(2*dim,sizeof(double));
    spaceint  = (int    *)mCalloc(2*dim,sizeof(int));
    Q_buff    = (double *)mCalloc(dim*dim,sizeof(double));
    Q_buff_sq = (double **)mCalloc(dim,sizeof(double *));
    U_sq      = (double **)mCalloc(dim,sizeof(double *));
    
    For(i,dim) 
        {
            Q_buff_sq[i] = (double *)mCalloc(dim,sizeof(double ));
            U_sq[i]      = (double *)mCalloc(dim,sizeof(double ));
        }
    
    
    For(i,dim*dim) 
        {
            Q_buff[i] = Q[i];
            Q_buff_sq[i/dim][i%dim] = Q[i];
        }
    
    scalar   = 1.0;
    n_iter   = 0;
    result   = 0;
    
    result   = EigenRealGeneral(dim,Q_buff_sq,Root,di,U_sq,spaceint,space);
    For(i,dim*dim) U[i] = U_sq[i/dim][i%dim];


    /*       result   = eigen(1,Q_buff,dim,Root,di,U,V,space); */
    if(!result)
        {      
            /* Compute inverse(Vr) into Vi */
            For (i,dim*dim) V[i] = U[i];
            while(!Matinv(V, dim, dim, space))
                {
                    printf("\n. Trying Q<-Q*scalar and then Root<-Root/scalar to fix this...\n");
                    Q_buff   = (double *)mCalloc(dim*dim,sizeof(double));
                    scalar += scalar / 3.;
                    For(i,dim*dim) {Q_buff[i] = Q[i];}	  
                    For(i,dim*dim) Q_buff[i] *= scalar;
                    result = eigen(1,Q_buff,dim,Root,di,U,V,space);
                    if (result == -1)
                        Exit("\n. Eigenvalues/vectors computation does not converge : computation cancelled\n");
                    else if (result == 1)
                        Exit("\n. Complex eigenvalues/vectors : computation cancelled\n");
                    
                    For (i,dim*dim) V[i] = U[i];
                    n_iter++; 
                    if(n_iter > 100) Exit("\n. Cannot work out eigen vectors\n");
                };
            For(i,dim) Root[i] /= scalar;
        }
    else
        {
            if (result==-1)
                printf("\n. Eigenvalues/vectors computation does not converge : computation cancelled");
            else if (result==1)
                printf("\n. Complex eigenvalues/vectors : computation cancelled");
        }
    
    
    Free(Q_buff);
    For(i,dim) { Free(Q_buff_sq[i]); Free(U_sq[i]);}
    Free(U_sq);
    Free(Q_buff_sq);
    Free(di);
    Free(space);
    Free(spaceint);
}

/*********************************************************/
int Matinv (double *x, int n, int m, double *space)
{
  /* Invert matrix x.  x[n*m]  ... m>=n */

   int i,j,k;
   int *irow;
   double ee, t,t1,xmax;
   double det;
   
   ee = 1.0E-10;
   det = 1.0;

   irow = (int *)mCalloc(n,sizeof(int));

   For (i,n)  
     {
       xmax = 0.;
       for (j=i; j<n; j++)
         if (xmax < fabs(x[j*m+i])) 
	   { 
	     xmax = fabs(x[j*m+i]); 
	     irow[i]=j; 
	   }

      det *= xmax;
      if (xmax < ee)   
	{
	  printf("\n. Cannot invert the matrix of eigen vectors.\n");
	  return(0);
	}

      if (irow[i] != i) 
	{
	  For (j,m) 
	    {
	      t = x[i*m+j];
	      x[i*m+j] = x[irow[i]*m+j];
	      x[irow[i]*m+j] = t;
	    }
	}
      t = 1./x[i*m+i];
      For (j,n) 
	{
	  if (j == i) continue;
	  t1 = t*x[j*m+i];
	  For(k,m)  x[j*m+k] -= t1*x[i*m+k];
	  x[j*m+i] = -t1;
	}
      For(j,m)   x[i*m+j] *= t;
      x[i*m+i] = t;
   }                            /* i  */
   for (i=n-1; i>=0; i--) 
     {
       if (irow[i] == i) continue;
       For(j,n)  
	 {
	   t = x[j*m+i];
	   x[j*m+i] = x[j*m + irow[i]];
	   x[j*m + irow[i]] = t;
	 }
     }

   free(irow);
   return (1);
}

/*********************************************************/

/*Of no use for a codon model 
void Update_Qmat_GG98(model *mod, double theta)
{
// Fill the instantaneous 4*4 rate matrix for the GG98 model 

  int i,j;
  double *U, *V, *Root;


  U      = (double *)mCalloc(mod->ns*mod->ns,sizeof(double));
  V      = (double *)mCalloc(mod->ns*mod->ns,sizeof(double));
  Root   = (double *)mCalloc(mod->ns,sizeof(double));

  if((mod->kappa < 0.0001) || (mod->kappa > 1.0000))
    Exit("\n. Err. in Update_Qmat_GG98\n");

  //We use kappa from the model and the given theta (equilibrium GC) 
  //to compute substitution probabilities
  
  mod->qmat_struct->qmat[0*4+1] = theta/2;// A->C 
  mod->qmat_struct->qmat[0*4+2] = mod->kappa * theta/2 ;// A->G 
  mod->qmat_struct->qmat[0*4+3] = 1-(theta/2);// A->T 
  
  mod->qmat_struct->qmat[1*4+0] = 1-(theta/2);// C->A 
  mod->qmat_struct->qmat[1*4+2] = theta/2;// C->G 
  mod->qmat_struct->qmat[1*4+3] = mod->kappa * (1-(theta/2));// C->T 
  
  mod->qmat_struct->qmat[2*4+0] = mod->kappa * (1-(theta/2));// G->A 
  mod->qmat_struct->qmat[2*4+1] = theta/2;// G->C 
  mod->qmat_struct->qmat[2*4+3] = 1-(theta/2);// G->T 
  
  mod->qmat_struct->qmat[3*4+0] = 1-(theta/2);// T->A 
  mod->qmat_struct->qmat[3*4+1] = mod->kappa * theta/2;// T->C 
  mod->qmat_struct->qmat[3*4+2] = theta/2;// T->G 

  mod->qmat_struct->qmat[0*4+0] = -(1+mod->kappa * theta)/2;//A->A
  mod->qmat_struct->qmat[1*4+1] = -(1+mod->kappa * (1-theta))/2;//C->C
  mod->qmat_struct->qmat[2*4+2] = -(1+mod->kappa * (1-theta))/2;//G->G
  mod->qmat_struct->qmat[3*4+3] = -(1+mod->kappa * theta)/2;//T->T


  //We divide by the mean substitution rate, which is obtained by adding minus the diagonal terms
  mod->mr = .0;

  mod->mr = - mod->qmat_struct->qmat[0] * (1-theta)/2 - mod->qmat_struct->qmat[1] * theta/2 - mod->qmat_struct->qmat[2] * theta/2 - mod->qmat_struct->qmat[3] * (1-theta)/2;


 
  For(i,mod->ns*mod->ns) mod->qmat_struct->qmat[i] /= mod->mr;


  Update_Eigen(mod->ns,mod->qmat_struct->qmat,U,V,Root);
  
  For(i,mod->ns) For(j,mod->ns) 
    {
      mod->qmat_struct->u_mat[i*mod->ns+j] = U[i*mod->ns+j];
      mod->qmat_struct->v_mat[i*mod->ns+j] = V[i*mod->ns+j];
    }
  
  For(i,mod->ns)
    mod->qmat_struct->expD_mr_vct[i] = exp(Root[i]);

  
  Free(U);
  Free(V);
  Free(Root);

}
*/
/*********************************************************/
void Update_Qmat_Codon (model *mod, qmat *qmat_branch, double* cpi)
{
  /* Fill the instantaneous rate matrix (codon-based model 
     of substitutions) using codons equilibrium frequencies cpi
  */

  int i,j;
  int aa1,aa2;
  double sum;
  int num_codon1, num_codon2;
  int dim2,dim3;
  int n_catq;
  double w;
  double *U, *V, *Root;
  double div;
  int codon_64_i, codon_64_j;
 
  div = 0.0;

  U      = (double *)mCalloc(mod->ns*mod->ns,sizeof(double));
  V      = (double *)mCalloc(mod->ns*mod->ns,sizeof(double));
  Root   = (double *)mCalloc(mod->ns,sizeof(double));
 
  w = qmat_branch->dnds;

  n_catq   = 0;
  //mod->n_catq;
  dim2     = mod->ns;
  dim3     = mod->ns*mod->ns;


  num_codon1 = num_codon2 = -1;


  For(i,mod->ns)
    {
      codon_64_i = i%mod->ns_codon; //rest of the division of i by 61 for instance ; when i<61, returns i. 
      num_codon1 = mod->c_code->sense_c[codon_64_i]; //Composition code of codon 1

      for(j=i;j<mod->ns;j++)
	{
	  codon_64_j = j%mod->ns_codon;
	  num_codon2 = mod->c_code->sense_c[codon_64_j]; //Composition code of codon 2
	
	  if((i/mod->ns_codon) != (j/mod->ns_codon)) //if 2 codons differ ?
	     {
	       if((i%mod->ns_codon) == (j%mod->ns_codon)) //one sense codon below 61, one nonsense codon above 61 
		 {
		    //Filling the great 61*61 matrix 
		      //qmat is one-dimensional
		   

		       if(qmat_branch->qmat[i*mod->ns+j] < 0.0)
			 {
			   printf(" %f\n",
				  //	  qmat_branch->theta[(int)i/mod->ns_codon+(int)j/mod->ns_codon-1],
				  qmat_branch->qmat[i*mod->ns+j]);
			   Exit("\n. Q(i!=j) < 0.0 !?!\n");
			 }

		     
		 }
	       else //Not possible to switch between sense and nonsense codons
		 {

		       qmat_branch->qmat[i*mod->ns+j] = 0.0;
		       qmat_branch->qmat[j*mod->ns+i] = qmat_branch->qmat[i*mod->ns+j];

		 }
	     }
	  else
	    {
	      //If only one base substitution between 2 codons
	      if(mod->c_code->n_diff_b_2_codons[codon_64_i*mod->ns_codon+codon_64_j] == 1)
		{
		  aa1 = Get_AA(mod->c_code,num_codon1);
		  aa2 = Get_AA(mod->c_code,num_codon2);

	
		      qmat_branch->qmat[i*mod->ns+j]=
			(aa1 == aa2)?
			(
			 (mod->c_code->tstvtable[codon_64_i*mod->ns_codon+codon_64_j]==1)?
			 (mod->kappa[0]):
			 (1.0)
			 ):
			(
			 (mod->c_code->tstvtable[codon_64_i*mod->ns_codon+codon_64_j]==1)?
			 (qmat_branch->t_dnds*mod->kappa[0]):
			 (qmat_branch->t_dnds)
			 );

                      /**/

/*                       (aa1 == aa2)? */
/*                           (*(mod->c_code->gtr_4_codon[codon_64_i*mod->ns_codon+codon_64_j])): */
/*                           (*(qmat_branch->t_omega[k*mod->n_omega+(int)i/mod->ns_codon])* */
/*                           (*(mod->c_code->gtr_4_codon[codon_64_i*mod->ns_codon+codon_64_j]))); */
                      
                        /**/
		      qmat_branch->qmat[j*mod->ns+i] = qmat_branch->qmat[i*mod->ns+j];

		      if(qmat_branch->qmat[i*mod->ns+j] < 0.0)
			{
			  printf("%f %f %f\n",
				 mod->kappa[0],
				 qmat_branch->t_dnds,
				 qmat_branch->qmat[i*mod->ns+j]);
			  Exit("\n. Q(i!=j) < 0.0 !?!\n");
			}
		}
	      else //if more than 1 base substitution between the two codons
		{  
		  qmat_branch->qmat[i*dim2+j] = 0.0;
		  qmat_branch->qmat[j*mod->ns+i] = qmat_branch->qmat[i*mod->ns+j]; 
		}
	    }
	}
    }



  /* Build equilibrium frequencies into the rate matrix */

      For(i,mod->ns)
	{
	  qmat_branch->qmat[i*mod->ns+i] = 0.0;
               
	  For(j, mod->ns)
	    {
	      if((i/mod->ns_codon) != (j/mod->ns_codon))
		{
		  /*	  qmat_branch->qmat[i*dim2+j] *=
		    (*(qmat_branch->t_omega_proba[(int)j/mod->ns_codon]));
		  */
		}
	      else
		{
		  qmat_branch->qmat[i*dim2+j] *=
		    cpi[j%mod->ns_codon];
		}
	    }
	}


  /* Scaling factor so that the expected number of codon substitutions per time unit = 1 */
  mod->mr = .0;
 
      For(i,mod->ns)
	{
	  For(j,mod->ns)
	    {
	      if((int)(i/mod->ns_codon) == (int)(j/mod->ns_codon))
		{
		  mod->mr +=
		    cpi[i%mod->ns_codon] *
		    //   (*(qmat_branch->t_omega_proba[(int)i/mod->ns_codon])) *
		    //qmat_branch->qmat_proba[k] *
		    qmat_branch->qmat[i*dim2+j];
		}
	    }
	}
   

  /*   printf("SCALER = %f\n",mod->mr); */


  /* Scale matrix */

      For(i,mod->ns)
	{
	  For(j,mod->ns)
	    {
	      if((int)(i/mod->ns_codon) == (int)(j/mod->ns_codon))
		{
		  qmat_branch->qmat[i*dim2+j] /= mod->mr;
		}
	    }
	}


  /* Diagonal  terms */

      For(i,mod->ns)
	{
	  sum = .0;
	  For(j,mod->ns)
	    sum += qmat_branch->qmat[i*dim2+j];
	  qmat_branch->qmat[i*dim2+i] = -sum;
	}




 
      Update_Eigen(mod->ns,qmat_branch->qmat,U,V,Root);

      For(i,mod->ns) For(j,mod->ns)
	{
	  qmat_branch->u_mat[i*dim2+j]  = U[i*dim2+j];
	  qmat_branch->v_mat[i*dim2+j]  = V[i*dim2+j];
	  if(V[i*dim2+j] > 1.E+10) 
	    Exit("\n. Numerical overflow during the calculation of eigen vectors\n");
	}

      For(i,mod->ns) 
	{
	  qmat_branch->expD_mr_vct[i] = exp(Root[i]);
	}
   

  Free(U);
  Free(V);
  Free(Root);
  
}
