/*

nhPhyml :  a program that builds phylogenetic trees from DNA alignments using the non-homogeneous, non-stationary model of evolution conceived by Galtier and Gouy (1998) and the algorithmic structure of phyml (Guindon and Gascuel, 2003).

Copyright (C) Bastien Boussau. jul. 2005 onward

All parts of  the source except where indicated  are distributed under
the GNU public licence.  See http://www.opensource.org for details.

*/

#include "utilities.h"
#include "optimiz.h"
#include "simu.h"
#include "ml.h"
#include "free.h"

double UNLIKELY;
double ROUND_MAX;




/*********************************************************/
/*Used in Simu*/
void Optimiz_Ext_Br(arbre *tree)
{
  int i, pos;
  edge *b;
  double l_infa,l_max,l_infb,l_init, theta_init[tree->mod->n_pos];
  double lk_infa,lk_max,lk_infb,lk, lk_init;
  //  printf("OptimizExtBr\n");
  lk_init = tree->tot_loglk;
  
  For(i,2*tree->n_otu-2) //One more edge in the rooted case
    {
     
      b = tree->t_edges[i];
      // We optimize leaf edges and also the root edges, as well as constrained branches.
      if((b->left->tax) || (b->rght->tax)||(b->left->num==tree->n_otu)||(b->rearr==0))
    
	{
	  l_init = b->l;
	  for (pos=0;pos<tree->mod->n_pos; pos++) {
	    theta_init[pos] = b->rght->theta[pos];
	  }
	  l_infa = BL_MIN;
	  l_max  = b->l;
	  l_infb = 4.*b->l;
	  if(b->l <= BL_MIN) {l_infa = -BL_START; l_max = .0; l_infb = BL_START;}
	  lk_infa = lk_max = lk_infb = .0;
 
	  if (tree->mod->EqFreq) {
	    lk=Try_Thetas(tree, b, 1);
	  }
	  else {
	    lk=Br_Len_Theta_NR_Slow(tree,tree->data, b, 100, b->l, b->rght->theta);
	  }

	  b->ql[0] = b->l;
	  for (pos=0;pos<tree->mod->n_pos; pos++) {
	    b->rght->qtheta[pos][0] = b->rght->theta[pos];
	    b->rght->theta[pos] = theta_init[pos];
	  }
	  b->best_conf = 1;
	  b->l = l_init;
	
	}
    }
  tree->tot_loglk = lk_init; 
}




/*********************************************************/
/*
Recursive function that permits the exploration of the whole tree : 
we start at the root and then get down to the leaves.
*/
void Optimize_Param_Serie(node *a, node *d, edge *b_fcus, 
			  arbre *tree,allseq *alldata, int n_passes)
{
  int i;
  double l_infa,l_max,l_infb;
  double lk_infa,lk_max,lk_infb;
  l_infa = BL_MIN;
  l_max  = b_fcus->l;
  l_infb = 4.*b_fcus->l;
  if(b_fcus->l <= BL_MIN) {l_infa = -BL_START; l_max = .0; l_infb = BL_START;}
  lk_infa = lk_max = lk_infb = .0;
  
  Br_Len_NR_End(tree,alldata, b_fcus, 100);
  // printf("bl2 : %f\n", b_fcus->l);

  if(d->tax) return;
  else For(i,3) if(d->v[i] != a)
    {
      Update_P_Lk(tree,d->b[i],d);
      Optimize_Param_Serie(d,d->v[i],d->b[i],tree,alldata,n_passes);
    }
  For(i,3) if((d->v[i] == a) && !(d->v[i]->tax)) Update_P_Lk(tree,d->b[i],d);
}






/*********************************************************/
double Br_Len_NR_End(arbre *tree,allseq *alldata, edge *b_fcus, int n_passes)
{
  int n_iter, n_iter_mov,edge_num, pos;
  double step_edge,lk_new,lk_old;
  double l_branch_best;
  int l_r,r_l, br, site, catg;
  double mult;
  int old_opt;
  edge *b;
 
  old_opt =              tree->mod->s_opt->opt_bl;
  tree->mod->s_opt->opt_bl = tree->both_sides = 1;
  edge_num =                     b_fcus->num;
  l_branch_best =                  b_fcus->l;
  step_edge =                             .0;

  l_r = b_fcus->l_r;
  r_l = b_fcus->r_l;

  lk_new = lk_old = UNLIKELY;

  n_iter = 0;
  //A OPTIMISER
  // Lk(tree, tree->data);

  

For(br,2*tree->n_otu-2) // One more branch in the rooted case 
    {
      For(site,tree->n_pattern)
	{
	  tree->t_edges[br]->sum_scale_f_rght[site] = .0;
	  tree->t_edges[br]->sum_scale_f_left[site] = .0;
	}

      if(tree->t_edges[br]->l < BL_MIN) tree->t_edges[br]->l = BL_MIN;
      b=tree->t_edges[br];

      For(pos,tree->mod->n_pos){
	For(catg,tree->mod->n_catg)
	  {
	    Computes_Substitution_Probabilities(b, tree, catg, pos);
	  }
      }
    }


  Post_Order_Lk(tree->noeud[tree->n_otu],
		tree->noeud[tree->n_otu]->v[0],
		tree); // Rooted case!
  Post_Order_Lk(tree->noeud[tree->n_otu],
		tree->noeud[tree->n_otu]->v[1],
		tree); // Rooted case!

 
  Pre_Order_Lk(tree->noeud[tree->n_otu],
	       tree->noeud[tree->n_otu]->v[0],
	       tree);// Rooted case!
  Pre_Order_Lk(tree->noeud[tree->n_otu],
	       tree->noeud[tree->n_otu]->v[1],
	       tree);// Rooted case!





  //  printf ("lktree : %f\n", tree->tot_loglk);
  for(;;)
    {
    
      lk_new = Lk_At_Given_Edge(tree,b_fcus);
      //      printf("LKNEW : %f\n", lk_new);
      if(fabs(lk_new - lk_old) < MIN_DIFF_LK)
	break;
      else
	{
	  if((lk_new > lk_old+MIN_DIFF_LK) || (!n_iter))
	    {
	      lk_old = lk_new;
	      l_branch_best = b_fcus->l;
	    }

	  else if((lk_new < lk_old-MIN_DIFF_LK) && (lk_old != 0.0))
	    {
	      tree->mod->s_opt->opt_bl = tree->both_sides = 0;
	      n_iter_mov = 0;
	      while(lk_new < lk_old-MIN_DIFF_LK)
		{
		  //		  printf("step_edge before: %f\n", step_edge);
		  step_edge *= .5;
		  //	  printf("step_edge after : %f\n", step_edge);
		  b_fcus->l = l_branch_best + step_edge;
		  if(b_fcus->l < BL_MIN) b_fcus->l = BL_MIN;
		  lk_new = Lk_At_Given_Edge(tree,b_fcus);
		  // printf("lknew2 : %f\n",lk_new);
		  n_iter_mov++;
		  if(n_iter_mov > 10)
		    {
		      b_fcus->l = l_branch_best;
		      lk_new = Lk_At_Given_Edge(tree,b_fcus);
		      tree->mod->s_opt->opt_bl = tree->both_sides = old_opt;
		      return lk_new;
		    }
		}
	      tree->mod->s_opt->opt_bl = tree->both_sides = 1;
	      lk_new = Lk_At_Given_Edge(tree,b_fcus);
	      tree->tot_dloglk[edge_num] = .0;
	    }
	}

      mult = 0;
    
      step_edge = - tree->tot_dloglk[edge_num] / tree->tot_d2loglk[edge_num];
      //      printf("length : %f, b_fcus->rght->num :%d,b_fcus->rght->name :%s \n",b_fcus->l, b_fcus->rght->num,b_fcus->rght->name ); 
      /*  printf("STEPEDGE : %f\n", step_edge);
      printf("tree->tot_dloglk[edge_num] : %f, tree->tot_d2loglk[edge_num] : %f\n",tree->tot_dloglk[edge_num],tree->tot_d2loglk[edge_num]);
      printf("length : %f\n\n",b_fcus->l);
      */
      if(tree->tot_d2loglk[edge_num] > .0)
	{
	  b_fcus->l /= 1.2;
	  step_edge = 0;
	}

      while(fabs(step_edge) > b_fcus->l) step_edge /= 1.2;
      b_fcus->l += step_edge;

      if(b_fcus->l < BL_MIN)
	{
	  b_fcus->l = BL_MIN;
	  break;
	}

      n_iter++;
      if(n_iter > n_passes) {
	//	printf("NITER\n");
	b_fcus->l = l_branch_best;
	break;
      }
    }

  tree->mod->s_opt->opt_bl = tree->both_sides = old_opt;
  return lk_new;
}

/*********************************************************/

double Br_Len_NR(arbre *tree,allseq *alldata, edge *b_fcus, int n_passes, double initl)
{
  int n_iter, n_iter_mov,edge_num, pos;
  double step_edge,lk_new,lk_old;
  double l_branch_best;
  int l_r,r_l, br, site, catg;
  double mult;
  int old_opt;
  edge *b;

  old_opt =              tree->mod->s_opt->opt_bl;
  tree->mod->s_opt->opt_bl = tree->both_sides = 1;
  edge_num =                     b_fcus->num;
  l_branch_best =                  b_fcus->l;
  step_edge =                             .0;

  l_r = b_fcus->l_r;
  r_l = b_fcus->r_l;

  lk_new = lk_old = UNLIKELY;
 
  
  For(br,2*tree->n_otu-2) // One more branch in the rooted case 
    {
      For(site,tree->n_pattern)
	{
	  tree->t_edges[br]->sum_scale_f_rght[site] = .0;
	  tree->t_edges[br]->sum_scale_f_left[site] = .0;
	}

      if(tree->t_edges[br]->l < BL_MIN) tree->t_edges[br]->l = BL_MIN;
      b=tree->t_edges[br];
      
      For(pos,tree->mod->n_pos){
	For(catg,tree->mod->n_catg)
	  {
	    
	    Computes_Substitution_Probabilities(b, tree, catg, pos);
	  }
      }
    }


  Post_Order_Lk(tree->noeud[tree->n_otu],
		tree->noeud[tree->n_otu]->v[0],
		tree); // Rooted case!
  Post_Order_Lk(tree->noeud[tree->n_otu],
		tree->noeud[tree->n_otu]->v[1],
		tree); // Rooted case!

 
  Pre_Order_Lk(tree->noeud[tree->n_otu],
	       tree->noeud[tree->n_otu]->v[0],
	       tree);// Rooted case!
  Pre_Order_Lk(tree->noeud[tree->n_otu],
	       tree->noeud[tree->n_otu]->v[1],
	       tree);// Rooted case!

  




  n_iter = 0;
  for(;;)
    {
      lk_new = Lk_At_Given_Edge(tree,b_fcus);

      if(fabs(lk_new - lk_old) < MIN_DIFF_LK)
	break;
      else
	{
	  if((lk_new > lk_old+MIN_DIFF_LK) || (!n_iter))
	    {
	      lk_old = lk_new;
	      l_branch_best = b_fcus->l;
	    }

	  else if((lk_new < lk_old-MIN_DIFF_LK) && (lk_old != 0.0))
	    {
	      tree->mod->s_opt->opt_bl = tree->both_sides = 0;
	      n_iter_mov = 0;
	      while(lk_new < lk_old-MIN_DIFF_LK)
		{
		  step_edge *= .5;
		  b_fcus->l = l_branch_best + step_edge;
		  if(b_fcus->l < BL_MIN) b_fcus->l = BL_MIN;
		  lk_new = Lk_At_Given_Edge(tree,b_fcus);
		  n_iter_mov++;
		  if(n_iter_mov > 10)
		    {
		      //We optimize the length partially
		      b_fcus->l = (l_branch_best+16*initl)/17;
		      lk_new = Lk_At_Given_Edge(tree,b_fcus);
		      tree->mod->s_opt->opt_bl = tree->both_sides = old_opt;
		      return lk_new;
		    }
		}
	      tree->mod->s_opt->opt_bl = tree->both_sides = 1;
	      lk_new = Lk_At_Given_Edge(tree,b_fcus);
	      tree->tot_dloglk[edge_num] = .0;
	    }
	}

      mult = 0;
      step_edge = - tree->tot_dloglk[edge_num] / tree->tot_d2loglk[edge_num];
     
      
      if(tree->tot_d2loglk[edge_num] > .0)
	{
	  b_fcus->l /= 1.2;
	  step_edge = 0;
	}

      while(fabs(step_edge) > b_fcus->l) step_edge /= 1.2;
      b_fcus->l += step_edge;

      if(b_fcus->l < BL_MIN)
	{
	  b_fcus->l = BL_MIN;
	  break;
	}

      n_iter++;
      if(n_iter > n_passes) 
	{
	  b_fcus->l = l_branch_best;
	  break;
	}

    }
  /*To avoid getting trapped in a local maximum during the exploration of the space of topologies */
 
  b_fcus->l=(b_fcus->l+16*initl)/17;
 
  lk_new = Lk_At_Given_Edge(tree,b_fcus);

  tree->mod->s_opt->opt_bl = tree->both_sides = old_opt;
  return lk_new;
}

/*********************************************************/
/*Optimization procedure as in nhml */
void NR_Tot (arbre *tree) {
  /*
    tables containing the proposed optimized values for the parameters.
  */
  double *prop_omega, **prop_theta, *prop_bl, *prop_kappa, *old_omega, **old_theta, *old_bl, *old_kappa, prop_alpha, old_alpha;
  double lk_old, lk_new, div;
  int i, n_iter, pos;

  Lk(tree, tree->data);
  prop_theta=NULL;
  prop_bl=NULL;
  old_theta=NULL;
  old_bl=NULL;
  
  old_alpha=0.;
  prop_alpha=0.;
 
  old_omega=(double *)mCalloc(tree->mod->n_pos, sizeof(double));
  prop_omega=(double *)mCalloc(tree->mod->n_pos, sizeof(double));
  old_kappa=(double *)mCalloc(tree->mod->n_pos, sizeof(double));
  prop_kappa=(double *)mCalloc(tree->mod->n_pos, sizeof(double));

  
  
  for (pos=0;pos<tree->mod->n_pos;pos++) {
    old_omega[pos]=0.;
    prop_omega[pos]=0.;
    old_kappa[pos]=0.;
    prop_kappa[pos]=0.;   
  }
  
  lk_new=tree->tot_loglk;
  lk_old=UNLIKELY;



  if (tree->mod->s_opt->opt_theta) {
    prop_theta=(double **)mCalloc(tree->mod->n_pos, sizeof(double *));
    old_theta=(double **)mCalloc(tree->mod->n_pos, sizeof(double *));
    for (pos=0;pos<tree->mod->n_pos;pos++) {
      prop_theta[pos]=(double *)mCalloc(2*tree->n_otu-2, sizeof(double));
      old_theta[pos]=(double *)mCalloc(2*tree->n_otu-2, sizeof(double));
      for (i=0;i<2*tree->n_otu-2;i++) {
	prop_theta[pos][i]=0;
	old_theta[pos][i]=0.;
      }
    }
  }
  //If the number of equilibrium frequencies is limited, we search for the best assignation before we optimize the other parameters
  if (tree->mod->EqFreq) {
    Optimize_Branch_Thetas(tree, 1000);
  }
 

  if (tree->mod->s_opt->opt_bl) {
    prop_bl=(double *)mCalloc(2*tree->n_otu-2, sizeof(double));
    old_bl=(double *)mCalloc(2*tree->n_otu-2, sizeof(double));
    for (i=0;i<2*tree->n_otu-2;i++) {
      prop_bl[0]=0;
      old_bl[0]=0.;
    }
  }
  

  //We first optimize length coefficients if there is more than one type of position
  if (tree->mod->n_pos>1) {
    for (pos=0;pos<tree->mod->n_pos-1;pos++) { //l_coef[last]=1.0
      Optimize_L_Coef(tree, pos);
    }
  }
  
 

   //As long as the likelihood can increase
  while (lk_new>lk_old+MIN_DIFF_LK*100) {
    lk_old=lk_new=tree->tot_loglk;
 
    //propose new values for the parameters
    if (tree->mod->s_opt->opt_omega) {
      NR_Omega(prop_omega, tree);
      for (pos=0;pos<tree->mod->n_pos;pos++) {
	old_omega[pos] = tree->mod->omega[pos];
      }
    }

    if (tree->mod->s_opt->opt_kappa) {
      NR_Kappa(prop_kappa, tree);
      for (pos=0;pos<tree->mod->n_pos;pos++) {
	old_kappa[pos] = tree->mod->kappa[pos];
      }
    }
    
    if (tree->mod->s_opt->opt_alpha) {
      NR_Alpha(&prop_alpha, tree);
      old_alpha = tree->mod->alpha;
    }
    
    //If the set of equilibrium frequencies is not limited
    if (tree->mod->s_opt->opt_theta) {
      NR_Theta(prop_theta, tree);
     }
 
    if (tree->mod->s_opt->opt_bl) {
      NR_Bl(prop_bl, tree);
    }
 
    for (i=0;i<2*tree->n_otu-2;i++) {
         if (tree->mod->s_opt->opt_theta) {
	   for (pos=0;pos<tree->mod->n_pos;pos++) {
	     old_theta[pos][i]=tree->t_edges[i]->rght->theta[pos];
	   }
	}
	if (tree->mod->s_opt->opt_bl) {
	  old_bl[i]=tree->t_edges[i]->l;
	}
      }

    n_iter=0;

    //Churchill-Felsenstein like technique : if the likelihood does not increase
    //when applying all the improvements at once, make smaller improvements.
    lk_new=UNLIKELY;
    while ((lk_new<lk_old) && (n_iter<100)){
    
            div=pow(2,n_iter);
	  
	    for (pos=0;pos<tree->mod->n_pos;pos++) {
	      if (tree->mod->s_opt->opt_omega) {
		tree->mod->omega[pos]=old_omega[pos]+(prop_omega[pos])/div;
		
		if (tree->mod->omega[pos]>=1.) {
		  tree->mod->omega[pos]=0.999999;
		} 
		if (tree->mod->omega[pos]<=0.) {
		  tree->mod->omega[pos]=0.000001;
		} 
		if (isnan(tree->mod->omega[pos])) {
		  tree->mod->omega[pos]=old_omega[pos];
		}
	      }
	      
	      
	      if (tree->mod->s_opt->opt_kappa) {
		tree->mod->kappa[pos]=old_kappa[pos]+(prop_kappa[pos])/div;
		if (tree->mod->kappa[pos]<=0.) {
		  tree->mod->kappa[pos]=0.000001;
		} 
		if (isnan(tree->mod->kappa[pos])) {
		  tree->mod->kappa[pos]=old_kappa[pos];
		}
	      }
	      
	    }
 
      if (tree->mod->s_opt->opt_alpha) {
	tree->mod->alpha=old_alpha+(prop_alpha)/div;
	if (tree->mod->alpha<=0.) {
	  tree->mod->alpha=0.000001;
	} 
	if (isnan(tree->mod->alpha)) {
	  tree->mod->alpha=old_alpha;
	}
      }
     
      for (i=0;i<2*tree->n_otu-2;i++) {
	if (tree->mod->s_opt->opt_theta) {
	  for (pos=0;pos<tree->mod->n_pos;pos++) {
	
	    tree->t_edges[i]->rght->theta[pos] = old_theta[pos][i]+(prop_theta[pos][i])/div;
	    
	    if (tree->t_edges[i]->rght->theta[pos]>=1.) {
	      tree->t_edges[i]->rght->theta[pos]=0.999999;
	    } 
	    if (tree->t_edges[i]->rght->theta[pos]<=0.) {
	      tree->t_edges[i]->rght->theta[pos]=0.000001;
	    }
	  }
	  /*	  printf("Dividing : tree->t_edges[i]->rght->theta : %f\n",tree->t_edges[i]->rght->theta[0]);
	  if  (i==4) {
	    printf("old_theta[pos][i] : %f prop_theta[pos][i] : %f\n",old_theta[0][i],prop_theta[0][i]);  
	    } 
	  */
	}
	if (tree->mod->s_opt->opt_bl) {
	  tree->t_edges[i]->l =  old_bl[i]+(prop_bl[i])/div;

	  if (tree->t_edges[i]->l<=0.) {
	    tree->t_edges[i]->l=0.000001;
	  } 

	}
      }
      n_iter++;
      lk_new=Return_Lk(tree);
      if (isnan(lk_new))   lk_new=Return_Lk(tree);
    

      //  printf("kappa : %f, omega : %f\n", tree->mod->kappa[0],tree->mod->omega[0]); 
    
       }
    printf("\nLikelihood : %f", lk_new);
    if (n_iter==100) {
      for (pos=0;pos<tree->mod->n_pos;pos++) {
	if (tree->mod->s_opt->opt_omega) {
	  tree->mod->omega[pos]=old_omega[pos];
	}
	if (tree->mod->s_opt->opt_kappa) {
	  tree->mod->kappa[pos]=old_kappa[pos];
	}
      }
      if (tree->mod->s_opt->opt_alpha) {
	tree->mod->alpha=old_alpha;
      }
     	for (i=0;i<2*tree->n_otu-2;i++) {
	  if (tree->mod->s_opt->opt_theta) {
	    for (pos=0;pos<tree->mod->n_pos;pos++) {	
	      tree->t_edges[i]->rght->theta[pos] = old_theta[pos][i];
	    }
	  }
	  if (tree->mod->s_opt->opt_bl) {
	    tree->t_edges[i]->l =  old_bl[i];
	  }
	}
	lk_new=lk_old;
    }
  }

}
/*********************************************************/
/*Fills the table prop_omega with optimized values obtained from
  the Newton_Raphson method */
double NR_Omega (double *prop_omega, arbre *tree) {
 
  double step_edge;
  int pos;


  for (pos=0;pos<tree->mod->n_pos;pos++) {

    step_edge = .0;
    
    Compute_Derivatives_Domega(tree, pos);
    
    step_edge =  - tree->tot_dloglk_domega / tree->tot_d2loglk_domega2;
    
    prop_omega[pos] = step_edge;
    
    if(tree->tot_d2loglk_domega2 > .0)
      {
	prop_omega[pos] = 0.0;
      }
  }
    //   printf("omega deriv : %f, deriv2 : %f step_edge : %f\n",tree->tot_dloglk_domega, tree->tot_d2loglk_domega2, *prop_omega);
    return tree->tot_loglk;

  
  
}
/*********************************************************/
/*Fills the table prop_theta with optimized values obtained from
  the Newton_Raphson method */

void NR_Theta(double **prop_theta, arbre *tree) {
  int i, pos;
  for (pos=0;pos<tree->mod->n_pos; pos++) {
    for (i=0;i<2*tree->n_otu-2;i++) {
      prop_theta[pos][i]=NR_Theta_Br(tree->t_edges[i], tree, pos);
    }
  }
}

/*********************************************************/
/*Applies Newton-Raphson to a single branch, in order to optimize its theta*/
double NR_Theta_Br(edge *b_fcus, arbre *tree, int pos) {
double  prop_theta, step_edge;
 

  step_edge = .0;
  
  
  
  Compute_Derivatives_Dtheta(tree, b_fcus, pos);
  
  step_edge = - tree->tot_dloglk_dtheta[b_fcus->num] / tree->tot_d2loglk_dtheta2[b_fcus->num];
  
  prop_theta=step_edge;
  
  if(tree->tot_d2loglk_dtheta2[b_fcus->num] > .0)
    {
      prop_theta=0.0;
    }

  //printf("theta%d deriv : %f, deriv2 : %f step_edge : %f\n", b_fcus->num,tree->tot_dloglk_dtheta[b_fcus->num], tree->tot_d2loglk_dtheta2[b_fcus->num], prop_theta);


  return prop_theta;
}


/*********************************************************/
/*Fills the table prop_bl with optimized values obtained from
  the Newton_Raphson method */

void NR_Bl(double *prop_bl, arbre *tree) {
  int i, rb1, rb2;
  /*We optimize the root edges length differently from the other edges*/
  rb1=rb2=-1;

  for (i=0;i<2*tree->n_otu-2;i++) {
    if (tree->t_edges[i]->left==tree->noeud[tree->n_otu]) {
      if (rb1<0){
	rb1=i;
      }
      else {rb2=i;}
    }
    else {/*Now we optimize the other edges length*/
      prop_bl[i]=NR_Bl_Br(tree->t_edges[i], tree);
    
    }
  }


  prop_bl[rb1]=NR_Bl_Br(tree->t_edges[rb1], tree);
  prop_bl[rb2]=NR_Bl_Br(tree->t_edges[rb2], tree);
  
  
}

/*********************************************************/

/*Applies Newton-Raphson to a single branch, in order to optimize its length*/
double NR_Bl_Br(edge *b_fcus, arbre *tree) {
 
  double prop_bl, step_edge;



 
  step_edge = .0;
 
   //Lk_At_Given_Edge provides the derivatives
  Lk_At_Given_Edge(tree,b_fcus);



     
      step_edge = - tree->tot_dloglk[b_fcus->num] / tree->tot_d2loglk[b_fcus->num];
   
      prop_bl = step_edge;
   
  
      if (tree->tot_d2loglk[b_fcus->num]>0.0) {

	prop_bl = 0.0;
	
      }
      //     printf("bl[%d] deriv :%f, deriv2 : %f, step_edge : %f\n", b_fcus->num,tree->tot_dloglk[b_fcus->num], tree->tot_d2loglk[b_fcus->num], prop_bl);
      return prop_bl;

}





/*********************************************************/
/*Fills the table prop_kappa with optimized values obtained from
  the Newton_Raphson method */
double NR_Kappa (double *prop_kappa, arbre *tree) {
  double step_edge, lk_new;
  int pos;
  step_edge = .0;
  
  lk_new = Return_Lk_At_Root(tree);
  
  for (pos=0;pos<tree->mod->n_pos;pos++) {
  
  
    Compute_Derivatives_Dkappa(tree, pos);
    
    
    
    step_edge =  - tree->tot_dloglk_dkappa / tree->tot_d2loglk_dkappa2;
    prop_kappa[pos] = step_edge;
    
    if (tree->tot_d2loglk_dkappa2 > .0)
      {
	/*
	  tree->mod->kappa=4.0;
	  NR_Kappa(prop_kappa,tree);
	*/
	prop_kappa[pos] = 0.0;
      }
  }

    //   printf("kappa deriv : %f, deriv2 : %f, step_edge : %f\n", tree->tot_dloglk_dkappa, tree->tot_d2loglk_dkappa2, *prop_kappa);
    return lk_new;
    

   
  
}
/*********************************************************/
/* We compute numerically the first and second derivatives of the likelihood 
   in respect to alpha.
   We compute the likelihood values lk1 at alpha-dalpha-dalpha and lk2 at alpha-dalpha+dalpha.
   A first derivative is then equal to dlk1=(lk2-lk1)/dalpha.
   We compute the likelihood values lk1bis at alpha+dalpha-dalpha and lk2bis at alpha+dalpha+dalpha.
   A first derivative is then equal to dlk2=(lk1bis-lk2bis)/dalpha.
   We compute the second derivative with : d2lk=(dlk2-dlk1)/(dalpha).
   Step_edge = d2lk/dlk, dlk being computed as the mean of the two first derivatives previously obtained.
*/
double NR_Alpha (double *prop_alpha, arbre *tree) {
  double dalpha, lk1, lk2, lk1bis, lk2bis, dlk1, dlk2, d2lk, dlk;
  dalpha=0.000001;
  Lk(tree, tree->data);


  lk1bis=lk2 = Return_Lk(tree);
  tree->mod->alpha=tree->mod->alpha-2*dalpha;
  lk1 = Return_Lk(tree);
  
  tree->mod->alpha=tree->mod->alpha+4*dalpha;
  lk2bis = Return_Lk(tree);
  
  dlk1=(lk2-lk1)/dalpha;
  dlk2=(lk2bis-lk1bis)/dalpha;
  
  d2lk=(dlk2-dlk1)/dalpha;
  dlk=(dlk2+dlk1)/2;
  tree->mod->alpha=tree->mod->alpha-2*dalpha; // Back to the original value

  *prop_alpha=- dlk/d2lk;

  //Was before  if (isinf(*prop_alpha)) {
  // Changed because of unix in :
  if (*prop_alpha>1000000) {
    *prop_alpha=1000.;
  }
  Lk(tree, tree->data); 

  return (*prop_alpha);
}


/*********************************************************/
/*Optimization procedure as in nhml, used when the topology is to be optimized, before simu */
void NR_Tot_First (arbre *tree) {
  /*
    tables containing the proposed optimized values for the parameters.
  */
  double *prop_omega, **prop_theta, *prop_bl, *prop_kappa, *old_omega, **old_theta, *old_bl, *old_kappa, prop_alpha, old_alpha;
  double lk_old, lk_new, div;
  int i, n_iter, pos;

  Lk(tree, tree->data);
  prop_theta=NULL;
  prop_bl=NULL;
  old_theta=NULL;
  old_bl=NULL;
  old_alpha=0.;
  prop_alpha=0.;


  old_omega=(double *)mCalloc(tree->mod->n_pos, sizeof(double));
  prop_omega=(double *)mCalloc(tree->mod->n_pos, sizeof(double));
  old_kappa=(double *)mCalloc(tree->mod->n_pos, sizeof(double));
  prop_kappa=(double *)mCalloc(tree->mod->n_pos, sizeof(double));
  
  for (pos=0; pos <tree->mod->n_pos; pos++) {
    prop_omega[pos]=0.;
    prop_kappa[pos]=0.;
    old_omega[pos]=0.;
    old_kappa[pos]=0.;
  } 
  
  
  
  
  lk_new=tree->tot_loglk;
  lk_old=UNLIKELY;
 

  //We do not optimize thetas at the beginning
 
  tree->mod->s_opt->opt_theta=0;
  prop_theta=(double **)mCalloc(tree->mod->n_pos, sizeof(double));
  old_theta=(double **)mCalloc(tree->mod->n_pos, sizeof(double));

  if (tree->mod->s_opt->opt_theta) {
    //  printf("Optimizing Thetas\n");
    for (pos=0; pos <tree->mod->n_pos; pos++) {
      prop_theta[pos]=(double *)mCalloc(2*tree->n_otu-2, sizeof(double));
      old_theta[pos]=(double *)mCalloc(2*tree->n_otu-2, sizeof(double));
      for (i=0;i<2*tree->n_otu-2;i++) {
	prop_theta[pos][i]=0;
	old_theta[pos][i]=0.;
      }
    }
  }
  if (tree->mod->s_opt->opt_bl) {
  
    prop_bl=(double *)mCalloc(2*tree->n_otu-2, sizeof(double));
    old_bl=(double *)mCalloc(2*tree->n_otu-2, sizeof(double));
    for (i=0;i<2*tree->n_otu-2;i++) {
      prop_bl[0]=0;
      old_bl[0]=0.;
    }
  }


  //If the number of equilibrium frequencies is limited, we search for the best assignation before we optimize the other parameters
  if (tree->mod->EqFreq) {
    for (i=0;i<2*tree->n_otu-2;i++) {
      for (pos=0;pos<tree->mod->n_pos;pos++) {
	tree->t_edges[i]->rght->theta[pos]=tree->mod->ThetaValues[(int)(tree->mod->NumEqFreq/2)];
      }
     }
    //  printf("tree->mod->EqFreq : %d\n", tree->mod->EqFreq);
    Optimize_Branch_Thetas(tree, 1);
  }
  /* for (i=0;i<2*tree->n_otu-2;i++) {
   printf("branche %d: %f\n", i, tree->t_edges[i]->rght->theta[0]);
   }*/



 //We first optimize length coefficients if there is more than one type of position
  if (tree->mod->n_pos>1) {
    for (pos=0;pos<tree->mod->n_pos-1;pos++) { //l_coef[last]=1.0
      Optimize_L_Coef(tree, pos);
    }
  }
  


//We test if only one or two steps of optimization before the exploration yelds a better exploration :
// The final likelihood is then lower than if we optimize the tree with the while...
//for(j=0;j<2;j++) {   
  while (lk_new>lk_old+20.) {
   
    lk_old=lk_new=tree->tot_loglk;
 
   
    if (tree->mod->s_opt->opt_omega) {
      NR_Omega(prop_omega, tree);
      for (pos=0; pos <tree->mod->n_pos; pos++) {
	old_omega[pos] = tree->mod->omega[pos];
      }
    }

    if (tree->mod->s_opt->opt_kappa) {
      NR_Kappa(prop_kappa, tree);
      for (pos=0; pos <tree->mod->n_pos; pos++) {
	old_kappa[pos] = tree->mod->kappa[pos];
      }
    }
    
    if (tree->mod->s_opt->opt_alpha) {
      NR_Alpha(&prop_alpha, tree);
      old_alpha = tree->mod->alpha;
    }
    
    if (tree->mod->s_opt->opt_theta) {
      NR_Theta(prop_theta, tree);
     }
 
    if (tree->mod->s_opt->opt_bl) {
      NR_Bl(prop_bl, tree);
    }
 
    
    for (i=0;i<2*tree->n_otu-2;i++) {
      for (pos=0; pos <tree->mod->n_pos; pos++) {
	if (tree->mod->s_opt->opt_theta) {
	  old_theta[pos][i]=tree->t_edges[i]->rght->theta[pos];
	  
	}
      }
      if (tree->mod->s_opt->opt_bl) {
	old_bl[i]=tree->t_edges[i]->l;
	
      }
    }
    

    n_iter=0;

    lk_new=UNLIKELY;
    
    while ((lk_new<lk_old) && (n_iter<100)){
    
      div=pow(2,n_iter);

      if (tree->mod->s_opt->opt_omega) {
	for (pos=0; pos <tree->mod->n_pos; pos++) {
	  tree->mod->omega[pos]=old_omega[pos]+(prop_omega[pos])/div;
	  
	  if (tree->mod->omega[pos]>=1.) {
	    tree->mod->omega[pos]=0.999999;
	  } 
	  if (tree->mod->omega[pos]<=0.) {
	    tree->mod->omega[pos]=0.000001;
	  } 
	  if (isnan(tree->mod->omega[pos])) {
	    tree->mod->omega[pos]=old_omega[pos];
	  }
	}
      }
      
      if (tree->mod->s_opt->opt_kappa) {
	for (pos=0; pos <tree->mod->n_pos; pos++) {
	  tree->mod->kappa[pos]=old_kappa[pos]+(prop_kappa[pos])/div;
	  if (tree->mod->kappa[pos]<=0.) {
	    tree->mod->kappa[pos]=0.000001;
	  } 
	  if (isnan(tree->mod->kappa[pos])) {
	    tree->mod->kappa[pos]=old_kappa[pos];
	  }
	}
      }
 
      if (tree->mod->s_opt->opt_alpha) {
	tree->mod->alpha=old_alpha+(prop_alpha)/div;
	if (tree->mod->alpha<=0.) {
	  tree->mod->alpha=0.000001;
	} 
	if (isnan(tree->mod->alpha)) {
	  tree->mod->alpha=old_alpha;
	}
      }
     

      for (i=0;i<2*tree->n_otu-2;i++) {
	if (tree->mod->s_opt->opt_theta) {
	  for (pos=0; pos <tree->mod->n_pos; pos++) {
	    tree->t_edges[i]->rght->theta[pos] = old_theta[pos][i]+(prop_theta[pos][i])/div;
	    
	    if (tree->t_edges[i]->rght->theta[pos]>=1.) {
	      tree->t_edges[i]->rght->theta[pos]=0.999999;
	    } 
	    if (tree->t_edges[i]->rght->theta[pos]<=0.) {
	      tree->t_edges[i]->rght->theta[pos]=0.000001;
	    }
	    
	  }
	}
	
	if (tree->mod->s_opt->opt_bl) {
	  tree->t_edges[i]->l =  old_bl[i]+(prop_bl[i])/div;

	  if (tree->t_edges[i]->l<=0.) {
	    tree->t_edges[i]->l=0.000001;
	  } 

	}
      }
      n_iter++;
      lk_new=Return_Lk(tree);
      if (isnan(lk_new))   lk_new=Return_Lk(tree);
      //printf("kappa : %f, omega : %f\n", tree->mod->kappa[0],tree->mod->omega[0]); 
      //  printf("Likelihood : %f\n", lk_new);
       }
    if (n_iter==100) {
      if (tree->mod->s_opt->opt_omega) {
	for (pos=0; pos <tree->mod->n_pos; pos++) {
	  tree->mod->omega[pos]=old_omega[pos];
	}
      }
      if (tree->mod->s_opt->opt_kappa) {
	for (pos=0; pos <tree->mod->n_pos; pos++) {
	  tree->mod->kappa[pos]=old_kappa[pos];
	}
      }
      if (tree->mod->s_opt->opt_alpha) {
	tree->mod->alpha=old_alpha;
      }
      for (i=0;i<2*tree->n_otu-2;i++) {
	for (pos=0; pos <tree->mod->n_pos; pos++) {
	  if (tree->mod->s_opt->opt_theta) {
	    tree->t_edges[i]->rght->theta[pos] = old_theta[pos][i];
	  }
	}
	if (tree->mod->s_opt->opt_bl) {
	  tree->t_edges[i]->l =  old_bl[i];
	}
      }
      lk_new=lk_old;
    }
      }
  tree->mod->s_opt->opt_theta=1;
}
/*********************************************************/
/* 
Function used to optimize the central branch length and theta when computing the approximate likelihood
of a topology obtained by a NNI.

*/
double Br_Len_Theta_NR(arbre *tree,allseq *alldata, edge *b_fcus, int n_passes, double initl, double *inittheta)
{
  int n_iter, edge_num, pos;
  double step_edge,lk_new,lk_old, prop_theta[tree->mod->n_pos], prop_bl,div;
  double l_branch_best;
  int l_r,r_l, br, site, catg;
  int old_opt;
  edge *b;

  old_opt =              tree->mod->s_opt->opt_bl;
  tree->mod->s_opt->opt_bl = tree->both_sides = 1;
  edge_num =                     b_fcus->num;
  l_branch_best =                  b_fcus->l;
  step_edge =                             .0;
  for (pos=0;pos<tree->mod->n_pos;pos++) {
    prop_theta[pos] = .0;
  }
  l_r = b_fcus->l_r;
  r_l = b_fcus->r_l;

  lk_new = lk_old = UNLIKELY;

 
  
  For(br,2*tree->n_otu-2) // One more branch in the rooted case 
    {
      For(site,tree->n_pattern)
	{
	  tree->t_edges[br]->sum_scale_f_rght[site] = .0;
	  tree->t_edges[br]->sum_scale_f_left[site] = .0;
	}

      if(tree->t_edges[br]->l < BL_MIN) tree->t_edges[br]->l = BL_MIN;
      b=tree->t_edges[br];
    
     
      For(pos,tree->mod->n_pos){
	For(catg,tree->mod->n_catg)
	  {
	    
	    Computes_Substitution_Probabilities(b, tree, catg, pos);
	    if(tree->mod->s_opt->opt_bl)
	      {
		Computes_Derivative_Substitution_Probabilities (b, tree, catg, pos);
		Computes_Second_Derivative_Substitution_Probabilities (b, tree, catg, pos);
	      }
	  }
      }
    }


  Post_Order_Lk(tree->noeud[tree->n_otu],
		tree->noeud[tree->n_otu]->v[0],
		tree); // Rooted case!
  Post_Order_Lk(tree->noeud[tree->n_otu],
		tree->noeud[tree->n_otu]->v[1],
		tree); // Rooted case!

 
  Pre_Order_Lk(tree->noeud[tree->n_otu],
	       tree->noeud[tree->n_otu]->v[0],
	       tree);// Rooted case!
  Pre_Order_Lk(tree->noeud[tree->n_otu],
	       tree->noeud[tree->n_otu]->v[1],
	       tree);// Rooted case!
   
  for (pos=0; pos <tree->mod->n_pos; pos++) {
    prop_theta[pos]=NR_Theta_Br(b_fcus, tree, pos);
    // printf("prop_theta : %f\n", prop_theta[pos]);
  }
  prop_bl=NR_Bl_Br(b_fcus, tree);
  //printf("prop_bl : %f\n", prop_bl);
  n_iter=10;
  
  lk_new=UNLIKELY;
  Lk_At_Given_Edge(tree, b_fcus);
  // Lk(tree, tree->data);
  lk_old=tree->tot_loglk;
  while ((lk_new<lk_old) && (n_iter<100)){
    
    div=pow(2,n_iter);
    n_iter++;
    for (pos=0;pos<tree->mod->n_pos;pos++) {
      b_fcus->rght->theta[pos] = inittheta[pos]+(prop_theta[pos])/div;
      if (b_fcus->rght->theta[pos]<=0.0) {
	b_fcus->rght->theta[pos]=0.000001;
      }
      if (b_fcus->rght->theta[pos]>=1.0) {
	b_fcus->rght->theta[pos]=0.999999;
      }
    }
    
    b_fcus->l = initl+(prop_bl)/div;
    if (b_fcus->l<BL_MIN) {
      b_fcus->l=BL_MIN;
    }
  
    Lk_At_Given_Edge(tree, b_fcus);
    lk_new = tree->tot_loglk;
    if (n_iter==100){
      for (pos=0;pos<tree->mod->n_pos;pos++) {
	b_fcus->rght->theta[pos] = inittheta[pos];
      }
      b_fcus->l = initl;
      lk_new=lk_old;
    }
  }
  return lk_new;
}





/*********************************************************/
/* 
Function used to optimize the central branch length and theta when computing the approximate likelihood
of a topology obtained by a NNI.

*/
double Br_Len_Theta_NR_Slow(arbre *tree,allseq *alldata, edge *b_fcus, int n_passes, double initl, double *inittheta)
{
  int n_iter, edge_num, pos;
  double step_edge,lk_new,lk_old, prop_theta[tree->mod->n_pos], prop_bl,div;
  double l_branch_best;
  int l_r,r_l, br, site, catg;
  int old_opt;
  edge *b;

  old_opt =              tree->mod->s_opt->opt_bl;
  tree->mod->s_opt->opt_bl = tree->both_sides = 1;
  edge_num =                     b_fcus->num;
  l_branch_best =                  b_fcus->l;
  step_edge =                             .0;

  For(pos,tree->mod->n_pos){
    prop_theta[pos] = .0;
  }
  l_r = b_fcus->l_r;
  r_l = b_fcus->r_l;

  lk_new = lk_old = UNLIKELY;

 
  
  For(br,2*tree->n_otu-2) // One more branch in the rooted case 
    {
      For(site,tree->n_pattern)
	{
	  tree->t_edges[br]->sum_scale_f_rght[site] = .0;
	  tree->t_edges[br]->sum_scale_f_left[site] = .0;
	}

      if(tree->t_edges[br]->l < BL_MIN) tree->t_edges[br]->l = BL_MIN;
      b=tree->t_edges[br];
      For(pos,tree->mod->n_pos){
	For(catg,tree->mod->n_catg)
	  {
	    
	    Computes_Substitution_Probabilities(b, tree, catg, pos);
	    if(tree->mod->s_opt->opt_bl)
	      {
		Computes_Derivative_Substitution_Probabilities (b, tree, catg, pos);
		Computes_Second_Derivative_Substitution_Probabilities (b, tree, catg, pos);
	      }
	  }
      }
    }


  Post_Order_Lk(tree->noeud[tree->n_otu],
		tree->noeud[tree->n_otu]->v[0],
		tree); // Rooted case!
  Post_Order_Lk(tree->noeud[tree->n_otu],
		tree->noeud[tree->n_otu]->v[1],
		tree); // Rooted case!

 
  Pre_Order_Lk(tree->noeud[tree->n_otu],
	       tree->noeud[tree->n_otu]->v[0],
	       tree);// Rooted case!
  Pre_Order_Lk(tree->noeud[tree->n_otu],
	       tree->noeud[tree->n_otu]->v[1],
	       tree);// Rooted case!
   
  
  for (pos=0; pos <tree->mod->n_pos; pos++) {
    prop_theta[pos]=NR_Theta_Br(b_fcus, tree, pos);
  }
  prop_bl=NR_Bl_Br(b_fcus, tree);
 
  n_iter=10;
  
  lk_new=UNLIKELY;
  Lk_At_Given_Edge(tree, b_fcus);
  // Lk(tree, tree->data);
  lk_old=tree->tot_loglk;
  while ((lk_new<lk_old) && (n_iter<100)){
    
    div=pow(2,n_iter);
    n_iter++;
    For(pos,tree->mod->n_pos){
      b_fcus->rght->theta[pos] = inittheta[pos]+(prop_theta[pos])/div;
      if (b_fcus->rght->theta[pos]<=0.0) {
	b_fcus->rght->theta[pos]=0.000001;
      }
      if (b_fcus->rght->theta[pos]>=1.0) {
	b_fcus->rght->theta[pos]=0.999999;
      }
    }
    
    b_fcus->l = initl+(prop_bl)/div;
    if (b_fcus->l<BL_MIN) {
      b_fcus->l=BL_MIN;
    }
  
    Lk_At_Given_Edge(tree, b_fcus);
    lk_new = tree->tot_loglk;
    if (n_iter==100){
      For(pos,tree->mod->n_pos){
	b_fcus->rght->theta[pos] = inittheta[pos];
      }
      b_fcus->l = initl;
      lk_new=lk_old;
    }
  }
  return lk_new;
}




/*********************************************************/
/*
Function periodically called by simu to optimize free parameters.

*/
void Optimiz_All_Free_Param_But_Theta_Bl_NR(arbre *tree) {


  /*
    tables containing the proposed optimized values for the parameters.
  */
  double *prop_omega, **prop_theta, *prop_bl, *prop_kappa, *old_omega, **old_theta, *old_bl, *old_kappa, prop_alpha, old_alpha;
  double lk_old, lk_new, div;
  int i, n_iter, pos;
  int starting_Opt_Theta, starting_Opt_Bl;

  Lk(tree, tree->data);
  prop_theta=NULL;
  prop_bl=NULL;
  old_theta=NULL;
  old_bl=NULL;
  old_alpha=0.;
  prop_alpha=0.0;

  old_omega=(double *)mCalloc(tree->mod->n_pos, sizeof(double));
  prop_omega=(double *)mCalloc(tree->mod->n_pos, sizeof(double));
  old_kappa=(double *)mCalloc(tree->mod->n_pos, sizeof(double));
  prop_kappa=(double *)mCalloc(tree->mod->n_pos, sizeof(double));
  

  for(pos=0;pos<tree->mod->n_pos;pos++) {
    prop_omega[pos]=0.;
    old_omega[pos]=0.;
    old_kappa[pos]=0.;
    prop_kappa[pos]=0.;
  }
  
  starting_Opt_Theta=tree->mod->s_opt->opt_theta;
  starting_Opt_Bl=tree->mod->s_opt->opt_bl;

  //Remove these if you want to optimize all the parameters
  tree->mod->s_opt->opt_theta=0;
  tree->mod->s_opt->opt_bl=0;

  lk_new=tree->tot_loglk;
  lk_old=UNLIKELY;
 
  

  prop_theta=(double **)mCalloc(tree->mod->n_pos, sizeof(double));
  old_theta=(double **)mCalloc(tree->mod->n_pos, sizeof(double));

  if (tree->mod->s_opt->opt_theta) {
    for (pos=0; pos <tree->mod->n_pos; pos++) {
      prop_theta[pos]=(double *)mCalloc(2*tree->n_otu-2, sizeof(double));
      old_theta[pos]=(double *)mCalloc(2*tree->n_otu-2, sizeof(double));
      for (i=0;i<2*tree->n_otu-2;i++) {
	prop_theta[pos][i]=0;
	old_theta[pos][i]=0.;
      }
    }
  }

  
  if (tree->mod->s_opt->opt_bl) {
    prop_bl=(double *)mCalloc(2*tree->n_otu-2, sizeof(double));
    old_bl=(double *)mCalloc(2*tree->n_otu-2, sizeof(double));
    for (i=0;i<2*tree->n_otu-2;i++) {
      prop_bl[0]=0;
      old_bl[0]=0.;
    }
  }
  /*
 //If the number of equilibrium frequencies is limited, we search for the best assignation before we optimize the other parameters
  if (tree->mod->EqFreq) {
    Optimize_Branch_Thetas(tree, 1000);
  }
  */

  //We first optimize length coefficients if there is more than one type of position
  if (tree->mod->n_pos>1) {
    for (pos=0;pos<tree->mod->n_pos-1;pos++) { //l_coef[last]=1.0
      Optimize_L_Coef(tree, pos);
    }
  }
  


    lk_old=lk_new=tree->tot_loglk;
 

    if (tree->mod->s_opt->opt_omega) {
      NR_Omega(prop_omega, tree);
      for(pos=0;pos<tree->mod->n_pos;pos++) {
	old_omega[pos] = tree->mod->omega[pos];
      }
      }


    if (tree->mod->s_opt->opt_kappa) {
      NR_Kappa(prop_kappa, tree);
      for(pos=0;pos<tree->mod->n_pos;pos++) {
	old_kappa[pos] = tree->mod->kappa[pos];
      }
    }


    if (tree->mod->s_opt->opt_alpha) {
      NR_Alpha(&prop_alpha, tree);
      old_alpha = tree->mod->alpha;
    }

    if (tree->mod->s_opt->opt_theta) {
      NR_Theta(prop_theta, tree);
     }
 
    if (tree->mod->s_opt->opt_bl) {
      NR_Bl(prop_bl, tree);
    }
 
    for (i=0;i<2*tree->n_otu-2;i++) {
         if (tree->mod->s_opt->opt_theta) {
	   for(pos=0;pos<tree->mod->n_pos;pos++) {
	     old_theta[pos][i]=tree->t_edges[i]->rght->theta[pos];
	   }
	 }
	if (tree->mod->s_opt->opt_bl) {
	  old_bl[i]=tree->t_edges[i]->l;
	  
	}
      }

    n_iter=1;

    lk_new=UNLIKELY;
    while ((lk_new<lk_old) && (n_iter<100)){

      div=pow(2,n_iter);
     
      if (tree->mod->s_opt->opt_omega) {
	 for(pos=0;pos<tree->mod->n_pos;pos++) {
	   tree->mod->omega[pos]=old_omega[pos]+(prop_omega[pos]-old_omega[pos])/div;
	   
	   if (tree->mod->omega[pos]>=1.) {
	     tree->mod->omega[pos]=0.999999;
	   } 
	   if (tree->mod->omega[pos]<=0.) {
	     tree->mod->omega[pos]=0.000001;
	   } 
	 }
      }

      
      if (tree->mod->s_opt->opt_kappa) {
	for(pos=0;pos<tree->mod->n_pos;pos++) {
	  tree->mod->kappa[pos]=old_kappa[pos]+(prop_kappa[pos]-old_kappa[pos])/div;
	  if (tree->mod->kappa[pos]<=0.) {
	    tree->mod->kappa[pos]=0.000001;
	  } 
	}
      }

      if (tree->mod->s_opt->opt_alpha) {
	tree->mod->alpha=old_alpha+(prop_alpha-old_alpha)/div;
	if (tree->mod->alpha<=0.) {
	  tree->mod->alpha=0.000001;
	} 
      }

      for (i=0;i<2*tree->n_otu-2;i++) {
	if (tree->mod->s_opt->opt_theta) {
	  for(pos=0;pos<tree->mod->n_pos;pos++) {
	    tree->t_edges[i]->rght->theta[pos] = old_theta[pos][i]+(prop_theta[pos][i]-old_theta[pos][i])/div;
	    
	    if (tree->t_edges[i]->rght->theta[pos]>=1.) {
	      tree->t_edges[i]->rght->theta[pos]=0.999999;
	    } 
	    if (tree->t_edges[i]->rght->theta[pos]<=0.) {
	      tree->t_edges[i]->rght->theta[pos]=0.000001;
	    }
	  }
	}
	if (tree->mod->s_opt->opt_bl) {
	  tree->t_edges[i]->l =  old_bl[i]+(prop_bl[i]-old_bl[i])/div;
	
	  if (tree->t_edges[i]->l<=0.) {
	    tree->t_edges[i]->l=0.000001;
	  } 

	}
      }

      n_iter++;
      lk_new=Return_Lk(tree);
      if (isnan(lk_new))   lk_new=Return_Lk(tree);


       }
    if ((n_iter==100)||isnan(lk_new)) {
      if (tree->mod->s_opt->opt_omega) {
	 for(pos=0;pos<tree->mod->n_pos;pos++) {
	   tree->mod->omega[pos]=old_omega[pos];
	 }
      }
      if (tree->mod->s_opt->opt_kappa) {
	for(pos=0;pos<tree->mod->n_pos;pos++) {	
	  tree->mod->kappa[pos]=old_kappa[pos];
	}
      }
      if (tree->mod->s_opt->opt_alpha) {
	tree->mod->alpha=old_alpha;
      }
      for (i=0;i<2*tree->n_otu-2;i++) {
	if (tree->mod->s_opt->opt_theta) {
	  for(pos=0;pos<tree->mod->n_pos;pos++) {
	    tree->t_edges[i]->rght->theta[pos] = old_theta[pos][i];
	  }
	}
	if (tree->mod->s_opt->opt_bl) {
	  tree->t_edges[i]->l =  old_bl[i];
	}
      }
      lk_new=lk_old;
    }
 //Remove these if you want to optimize all the parameters
    tree->mod->s_opt->opt_theta=starting_Opt_Theta;
    tree->mod->s_opt->opt_bl=starting_Opt_Bl;


}



/*********************************************************/
void Optimize_L_Coef(arbre *tree, int pos)
{

  double ax,bx,cx,old;
  double fa, fb, fc;
  double lk_init;

  lk_init = tree->tot_loglk;

  tree->mod->s_opt->opt_bl =
  tree->both_sides    = 0;

  ax =  .05;
  bx =  tree->mod->l_coef[pos];
  old = tree->mod->l_coef[pos];
  cx = 1000.;

  L_Coef_Brak(&ax,&bx,&cx,
              &fa,&fb,&fc,
              tree, pos);

  L_Coef_Golden(ax,bx,cx,
                1.e-5,&(tree->mod->l_coef[pos]),
                tree,tree->data,old, pos);

  if(tree->tot_loglk < lk_init) tree->tot_loglk = lk_init;
}

/*********************************************************/
double L_Coef_Golden(double ax, double bx, double cx, double tol,
		     double *xmin, arbre *tree, allseq *alldata, double old, int pos)
{
   double f1,f2,x0,x1,x2,x3;

   x0=ax;
   x3=cx;
   if (fabs(cx-bx) > fabs(bx-ax))
     {
       x1=bx;
       x2=bx+GOLDEN_C*(cx-bx);
     }
   else
     {
       x2=bx;
       x1=bx-GOLDEN_C*(bx-ax);
     }
   tree->mod->l_coef[pos]=x1;
   Lk(tree,alldata);
   f1=-tree->tot_loglk;
   tree->mod->l_coef[pos]=x2;
   Lk(tree,alldata);
   f2=-tree->tot_loglk;
   while (fabs(x3-x0) > tol*(fabs(x1)+fabs(x2)))
     {
       if (f2 < f1)
         {
           SHFT3(x0,x1,x2,GOLDEN_R*x1+GOLDEN_C*x3)
           tree->mod->l_coef[pos]=x2;
           Lk(tree,alldata);
           SHFT2(f1,f2,-tree->tot_loglk)
             }
       else
         {
           SHFT3(x3,x2,x1,GOLDEN_R*x2+GOLDEN_C*x0)
             tree->mod->l_coef[pos]=x1;
           Lk(tree,alldata);
           SHFT2(f2,f1,-tree->tot_loglk)
             }
     }
   if (f1 < f2)
     {
       *xmin=(x1+old)/2;
       return f1;
     }
   else
     {
       *xmin=(x2+old)/2;
       return f2;
  }
}

/*********************************************************/
int L_Coef_Brak(double *ax, double *bx, double *cx,
                double *fa, double *fb, double *fc,
                arbre *tree, int pos)
{
   double ulim,u,r,q,fu,dum;

   tree->mod->l_coef[pos] = *ax; // One limit to begin with
   *fa=-Return_Lk(tree);
   tree->mod->l_coef[pos] = *bx; // The other limit
   *fb=-Return_Lk(tree);
   if (*fb > *fa) {
      SHFT(dum,*ax,*bx,dum)
      SHFT(dum,*fb,*fa,dum)
   }
   *cx=(*bx)+MNBRAK_GOLD*(*bx-*ax);
   tree->mod->l_coef[pos] = fabs(*cx);
   *fc=-Return_Lk(tree);
   while (*fb > *fc)
     {

       r=(*bx-*ax)*(*fb-*fc);
       q=(*bx-*cx)*(*fb-*fa);
       u=(*bx)-((*bx-*cx)*q-(*bx-*ax)*r)/
               (2.0*SIGN(MAX(fabs(q-r),MNBRAK_TINY),q-r));
       ulim=(*bx)+MNBRAK_GLIMIT*(*cx-*bx);

       if ((*bx-u)*(u-*cx) > 0.0)
         {
           tree->mod->l_coef[pos] = fabs(u);
           fu=-Return_Lk(tree);
           if (fu < *fc)
             {
               *ax=(*bx);
               *bx=u;
               *fa=(*fb);
               *fb=fu;
               (*ax)=fabs(*ax);
               (*bx)=fabs(*bx);
               (*cx)=fabs(*cx);
               return(0);
             }
           else if (fu > *fb)
             {
               *cx=u;
               *fc=fu;
               (*ax)=fabs(*ax);
               (*bx)=fabs(*bx);
               (*cx)=fabs(*cx);
               return(0);
             }
           u=(*cx)+MNBRAK_GOLD*(*cx-*bx);
           tree->mod->l_coef[pos] = fabs(u);
           fu=-Return_Lk(tree);
    }
       else if ((*cx-u)*(u-ulim) > 0.0)
         {
           tree->mod->l_coef[pos] = fabs(u);
           fu=-Return_Lk(tree);
           if (fu < *fc)
             {
               SHFT(*bx,*cx,u,*cx+MNBRAK_GOLD*(*cx-*bx))
               tree->mod->l_coef[pos] = fabs(u);
               SHFT(*fb,*fc,fu,-Return_Lk(tree))
             }
         }
       else if ((u-ulim)*(ulim-*cx) >= 0.0)
         {
           u=ulim;
           tree->mod->l_coef[pos] = fabs(u);
           fu=-Return_Lk(tree);
         }
       else
         {
           u=(*cx)+MNBRAK_GOLD*(*cx-*bx);
           tree->mod->l_coef[pos] = fabs(u);
           fu=-Return_Lk(tree);
         }
       SHFT(*ax,*bx,*cx,u)
       SHFT(*fa,*fb,*fc,fu)
      }
   (*ax)=fabs(*ax);
   (*bx)=fabs(*bx);
   (*cx)=fabs(*cx);
   return(0);
}

/*********************************************************/
/*In case we use a limited set of equilibrium frequencies, this function traverses the tree and tries, for each branch, all the possible equilibrium frequencies (thetas). For each theta, the branch length is optimized. We finally take the most likely combination of thetas for all the branches using a procedure much like the search for more likely topologies : we apply a proportion lambda of all the changes at once ; if they do not improve the likelihood, we only apply half of them, the ones permitting the highest gain in likelihood, and so on, until the likelihood improves.*/
double Optimize_Branch_Thetas (arbre *tree, int n_step_max) {
  int i, step, n_neg, n_tested=0, each;
  double lambda=1.0;//0.75;
  double old_loglk = UNLIKELY; 
  int n_without_change     = 0;
  edge **sorted_b,**tested_b;
  int starting_Opt_Bl;

  sorted_b = (edge **)mCalloc(2*tree->n_otu-2,sizeof(edge *));
  tested_b = (edge **)mCalloc(2*tree->n_otu-2,sizeof(edge *));
  step=0;
  each=2;
  
  printf ("\nOptimizing the affectation of Equilibrium Frequencies...\n");
  do {
    ++step;
    
    if (step > n_step_max) break;
    each--;

    starting_Opt_Bl=tree->mod->s_opt->opt_bl;
    tree->mod->s_opt->opt_bl = 0;
    tree->both_sides    = 1;
    Lk(tree,tree->data);
    // printf("Likl 1 : %f\n",tree->tot_loglk); 
    tree->mod->s_opt->opt_bl = starting_Opt_Bl;

    /*  if(old_loglk < UNLIKELY+1)
 	printf("\n. Log(lk) :               * -> %15.6f ",tree->tot_loglk);
      else
 	printf("\n. Log(lk) : %15.6f -> %15.6f ",old_loglk,tree->tot_loglk);
  
    if(old_loglk > UNLIKELY+1)
	{
	  if(n_tested > 1) printf("%3d swaps done ",n_tested);
	  else             printf("%3d swap  done", n_tested);
	}
   
	fflush(NULL);*/
      if((fabs(old_loglk-tree->tot_loglk) < 1.E-03)|| (n_without_change == 1))  break; //If the increase in likelihood is insufficient or if no theta change has been performed

      if(tree->tot_loglk < old_loglk) //If the likelihood decreases
	{
	  /*  printf("\nLikelihood decreases\n");
	      printf("\n\n. Moving backward (thetas) \n");*/
	  UnMake_Theta_Changes(tree,tested_b,0,n_tested);
	  n_tested=(int)ceil((double)n_tested/2.);
	  Make_Theta_Changes(tree,tested_b,0,n_tested);
	  if(n_tested > 0) n_without_change = 0;
	  else             n_without_change++;
	}
      else //Likelihood does not decrease
	{ 
	  // Every four steps, perform parameters optimization
	  if(!each) 
	    {
	      each = 2;
	      Optimiz_All_Free_Param_But_Theta_Bl_NR(tree);
	      //Optimiz_All_Free_Param(tree,1,0);
	      tree->mod->s_opt->opt_bl = 0;
	      tree->both_sides    = 1;
	      Lk(tree,tree->data);
	      tree->mod->s_opt->opt_bl = starting_Opt_Bl;
	    }
	  
	  old_loglk = tree->tot_loglk;
	  For(i,2*tree->n_otu-2) {
	    Try_Thetas(tree,tree->t_edges[i], 0); 
	    // printf("Likl 2 : %f\n",tree->tot_loglk);
	  }
	  //  printf("Likl 3 : %f\n",tree->tot_loglk); 
	  //Now we have a set of thetas and associated bl and lk
	   Select_Edges_Theta_To_Change(tree,sorted_b,&n_neg);
	   // printf("n_neg : %d\n", n_neg);
	   Sort_Edges_Diff_Lk(tree,sorted_b,n_neg);
	    
	   n_tested = 0;
	   // the order of tested_b is the sorted order
	   For(i,(int)ceil((double)n_neg*(lambda)))
	     tested_b[n_tested++] = sorted_b[i];
	   Make_Theta_Changes(tree,tested_b,0,n_tested);

	   if(n_tested > 0) n_without_change = 0;
	   else             n_without_change++;
	}
  }
  while(1);
  Free(sorted_b);
  Free(tested_b);
  return (tree->tot_loglk);
}
/*********************************************************/


