/*

nhPhyml :  a program that builds phylogenetic trees from DNA alignments using the non-homogeneous, non-stationary model of evolution conceived by Galtier and Gouy (1998) and the algorithmic structure of phyml (Guindon and Gascuel, 2003).

Copyright (C) Bastien Boussau. jul. 2005 onward

All parts of  the source except where indicated  are distributed under
the GNU public licence.  See http://www.opensource.org for details.

*/

#include "utilities.h"
#include "free.h"


int  T_MAX_FILE;
double MDBL_MIN;
double UNLIKELY;



#define N_SEQUENCEFILE 1
//#define N_DATATYPE 2
#define N_FORMAT 2
#define N_DATASETS 3
//#define N_MODELNAME 5
#define N_KAPPA 4
//#define N_PROPORTIONINVAR 6 /*same as kappa*/
#define N_NBCATG 5
#define N_ALPHA 6
#define N_CHANGETREE 7
#define N_INTERNALSEQS 8
#define N_STARTINGTREE 9
#define N_NT_OR_CD 10

#define N_NB_PARAMS_DNA 11
//#define N_NB_PARAMS_AA 9



void Set_Defaults(option* input)
 {
  input->datatype          = 0; /*DNA*/
  input->nt_or_cd          = 0; /*Nucleotides*/
  input->n_data_sets       = 1;
  input->ncatg             = 1;
  input->interleaved       = 1;
  input->titv              = 4.0;
  input->end_opt           = 0;
  input->alpha             = 1.0;
  input->lambda            = 1.0;
  input->positions         = 0; // 0 : 123, 1 : 1, 2 : 12, 3 : 13, 4 : 2, 5 : 23, 6 : 3
  input->s_opt->opt_alpha  = 0;
  input->s_opt->opt_kappa  = 1;
  input->s_opt->opt_bl     = 1;
  input->s_opt->opt_omega  = 1;
  input->s_opt->opt_theta  = 1;
  input->s_opt->opt_RootPos  = 1;
  // input->s_opt->opt_pinvar = 0;
  input->inputtree         = 0;
  input->changetree        = 1;
  input->internalsequences = 0;
  input->tree              = NULL;
  //  input->invar             = 0;
  // input->pinvar            = 0.0;
  input->stepsize          = 1;
  input->model             = 0; /*Uniform model : no influence of the position in the codon on the model */
  input->estimate_variance = 0; /*We do not estimate the root GC variance */
  input->lower_GC_bound    = 0.65;
  input->upper_GC_bound    = 0.85;
  

}







/*********************************************************/
void Get_Input_Interactive(option *input)
{
  char choix;
  char *s    = (char *)mCalloc(T_MAX_LINE,sizeof(char));
  char *buff = (char *)mCalloc(T_MAX_LINE,sizeof(char));
  char *pos = (char *)mCalloc(T_MAX_LINE,sizeof(char));
  int n_trial;



  printf("Enter the sequence file name > "); fflush(NULL);
  Getstring_Stdin(input->seqfile);
  input->fp_seq = OpenSeqs(input,0);

  printf("Enter the tree file name > "); fflush(NULL);
  Getstring_Stdin(input->inputtreefile);
  input->fp_tree = OpenTree(input,0);
  printf("\n");
 
#ifdef WIN32

    if(Filexists(input->out_file))

  
#elif UNIX

  if(Filexists(input->out_file))

#endif
    {
      printf("\n");

      printf("A file '%s' already exists\n",input->out_file);

      printf("Do you want to Replace it or Append to it ?\n");
      n_trial = 0;
      do
        {
          printf("Please type R or A > ");
          scanf("%c",&choix);
          if(choix == '\n') choix = 'r';
          else getchar();
          if(++n_trial>10) Exit("\n");
          Uppercase(&choix);
        }
      while((choix != 'R') && (choix != 'A'));
      if(choix == 'R') input->out_file_open_mode = 1;
      else             input->out_file_open_mode = 2;
    }
    /*
#ifdef WIN32

  if(Filexists(input->tree_file))

#elif UNIX

  if(Filexists(input->tree_file))

#endif
    {
      printf("\n");

      printf("A file '%s' already exists\n",input->tree_file);

      printf("Do you want to Replace it or Append to it ?\n");
      n_trial = 0;
      do
        {
          printf("Please type R or A > ");
          scanf("%c",&choix);
          if(choix == '\n') choix = 'X';
          else getchar();
          Uppercase(&choix);

if(++n_trial>10) Exit("\n");
        }
      while((choix != 'R') && (choix != 'A'));
      if(choix == 'R') input->tree_file_open_mode = 1;
      else             input->tree_file_open_mode = 2;
    }
    */

  choix                    = 0;
  Set_Defaults(input);


  do
    {
#ifdef WIN32
      system("cls");
#elif UNIX
      printf("\033[2J\033[H");
#endif



      printf("\n - nhPHYML %s - \n\n\n",VERSION);


      printf("Settings for this run:\n\n");


     

    
      printf("  I "
             "       Input sequences interleaved (or sequential) "
             " %-15s \n",
             (input->interleaved)?("interleaved"):("sequential"));
    

      printf("  U "
             "         Unit of evolution : nucleotide or codon ? "
             " %-15s \n",
             (input->nt_or_cd)?("cd"):("nt"));


      if (input->nt_or_cd) {
	printf("  G "
	       "                                      Genetic code "
	       " %-15s \n",
	       input->mod->c_code->name);
      }
      else {

	if (input->positions==0) {
	  pos="123"; 
	}
	else if (input->positions==1) {
	  pos="1"; 
	}
	else if (input->positions==2) {
	  pos="12"; 
	}
	else if (input->positions==3) {
	  pos="13"; 
	}
	else if (input->positions==4) {
	  pos="2"; 
	}
	else if (input->positions==5) {
	  pos="23"; 
	}
	else if (input->positions==6) {
	  pos="3"; 
	}

	printf("  P "
	       "                                Selected positions "
	       " %-15s \n",
	       pos);

	printf("  M "
             "  Model of evolution : Uniform or Codon Position ? "
             " %-15s \n",
             (input->model)?("cdp"):("uni"));
      } 


      /* strcpy(s,"");
      sprintf(s," (%d sets)",input->n_data_sets);
      strcpy(buff,(input->n_data_sets > 1)?("yes"):("no"));
      buff=strcat(buff,(input->n_data_sets > 1)?(s):("\0"));
      printf("  S "
             "                        Analyze multiple data sets "
             " %-15s \n",buff);
      */


      
          strcpy(s,(input->s_opt->opt_kappa)?("estimated"):("fixed"));
          (input->s_opt->opt_kappa)?(strcat(s, "")):(strcat(s," (ts/tv = "));
          (input->s_opt->opt_kappa)?(strcat(s, "")):((char *)sprintf(s+(int)strlen(s),"%3.2f)",input->titv));

          printf("  T "
                 "                     Ts/tv ratio (fixed/estimated) "
                 " %-15s \n",s);
        

	  /*
      (input->s_opt->opt_pinvar)?(strcpy(s,"estimated")):(strcpy(s,"fixed"));
      (input->s_opt->opt_pinvar)?(strcat(s,"")):(strcat(s," (p-invar = "));
      (input->s_opt->opt_pinvar)?(strcat(s,"")):((char *)sprintf(s+strlen(s),"%3.2f)",input->pinvar));
      printf("  V  "
             " Proportion of invariable sites (fixed/estimated)"
             "  %-15s \n",s);
	  */

      printf("  R "
             "        One category of substitution rate (yes/no) "
             " %-15s \n",
             (input->ncatg > 1)?("no"):("yes"));
      if(input->ncatg > 1)
        {
          printf("  C "
                 "            Number of substitution rate categories "
" %-15d \n",
                 input->ncatg);
        }


      if(input->ncatg > 1)
        {
          strcpy(s,(input->s_opt->opt_alpha)?("estimated"):("fixed"));
          (input->s_opt->opt_alpha)?(strcat(s, "")):(strcat(s," (alpha = "));
          (input->s_opt->opt_alpha)?(strcat(s, "")):((char *)sprintf(s+strlen(s),"%3.2f)",input->alpha));

          printf("  A "
                 "    Gamma distribution parameter (fixed/estimated) "
                 " %-15s \n",s);
        }


      strcpy(s,(input->changetree)?("optimized"):("fixed"));
      
      printf("  O "
             "               Optimize Topology (fixed/Optimized) "
             " %-15s \n",
             s);

      strcpy(s,(input->mod->EqFreq)?("limited"):("Unlimited"));
      
      printf("  E "
             "Set of Equilibrium Frequencies (limited/Unlimited) "
             " %-15s \n",
             s);
      
      if (input->mod->EqFreq) {
	printf("  F "
	       "    Number of Equilibrium Frequencies "
	       " %d \n",input->mod->NumEqFreq);
      }

      if (input->positions==0){
      strcpy(s,(input->internalsequences)?("yes"):("No"));
      
      printf("  S "
             "                         Output Sequences (yes/No) "
             " %-15s \n",
             s);
      }
      
      printf("  Q                                              Quit");
      



      printf("\n");

      printf("\nAre these settings correct? "
             "(type  Y  or letter for one to change)  ");

      scanf("%c",&choix);
      if(choix == '\n') choix = 'X';
      else getchar(); /* \n */

      Uppercase(&choix);

      if ((choix == 'Y') || (choix == 'y'))
        break;

      switch(choix)
        {

	  /*  case 'U' :
          {
            if(!input->inputtree) input->inputtree = 1;
            else
              {
                input->inputtree = 0;
                printf("Enter the name of the tree file > ");
                Getstring_Stdin(input->inputtreefile);
                input->fp_tree = Openfile(input->inputtreefile,0);
              }
            break;
          }
	  */


	case 'U' :
            {
              (input->nt_or_cd)?(input->nt_or_cd=0):(input->nt_or_cd=1);
	      if (input->nt_or_cd) 
		{
		  input->mod->ns       = input->mod->c_code->n_sense_c;
		  input->mod->ns_codon = input->mod->c_code->n_sense_c;
		  input->stepsize=3;
		}
	      else 
		{
		  input->stepsize=1;
		}
              break;
            }

	case 'G' :
	  {
	    int code_num;
	    
	    code_num = input->mod->c_code->num_curr_code;
	    Free_Code(input->mod->c_code);
	    
	    if(code_num == 16)
	      code_num = 0;
	    else code_num++;
	    input->mod->c_code   = Make_Code(code_num);
	    input->mod->ns       = input->mod->c_code->n_sense_c;
	    input->mod->ns_codon = input->mod->c_code->n_sense_c;
	    
	    break;
	  }
        

	case 'R' :
            {
              (input->ncatg == 1)?(input->ncatg = 4):(input->ncatg = 1);
              break;
            }

        case 'C' :
          {
            char *c;
            printf("Enter your number of categories > ");
            c = (char *)mCalloc(T_MAX_LINE,sizeof(char));
            Getstring_Stdin(c);
            n_trial = 0;
            while((!atoi(c)) || (atoi(c) < 0))
              {
                if(++n_trial > 10) Exit("\nErr : the number of categories must be a positive integer\n");
                printf("\nThe number of categories must be a positive integer\n");
                printf("Enter a new value > ");
                Getstring_Stdin(c);
              }
            input->ncatg = atoi(c);
            Free(c);
            break;
          }


        case 'A' :
          {
            char answer;

            switch(input->s_opt->opt_alpha)
              {
              case 0 :
                {
                  printf("Optimize alpha ? [Y/n] ");
                  scanf("%c",&answer);
                  if(answer == '\n') answer = 'Y';
                  else getchar();
                  break;
                }
              case 1 :
                {
                  printf("Optimize alpha ? [N/y] ");
                  scanf("%c",&answer);
                  if(answer == '\n') answer = 'N';
		  else getchar();
                  break;
                }
              default : Exit("\n");
              }

            n_trial = 0;
            while((answer != 'Y') && (answer != 'y') &&
                  (answer != 'N') && (answer != 'n'))
              {
                if(++n_trial > 10) Exit("\nErr : wrong answers !");
                printf("Optimize alpha ? [N/y] ");
                scanf("%c",&answer);
                if(answer == '\n') answer = 'N';
                else getchar();
              }

            switch(answer)
              {
              case 'Y' : case 'y' :
                {
                  input->s_opt->opt_alpha = 1;
                  break;
                }
              case 'N' : case 'n' :
                {
                  char *a;
                  a = (char *)mCalloc(T_MAX_LINE,sizeof(char));
                  input->alpha = 1.0;
                  input->s_opt->opt_alpha = 0;
                  printf("Enter your value of alpha > ");
                  Getstring_Stdin(a);
                  n_trial = 0;
                  while((!atof(a)) || (atof(a) < .0))
                    {
                      if(++n_trial > 10) Exit("\nErr : alpha must be a positive number\n");
                      printf("Alpha must be a positive number\n");
                      printf("Enter a new value > ");
                      Getstring_Stdin(a);
                    }
                  input->alpha = (double)atof(a);
                  Free(a);
                  input->s_opt->opt_alpha  = 0;
                  break;
                }
              }
            break;
          }

        case 'T' :
          {
            char answer;
	    if((input->datatype)   ||
               (input->model == 1) ||
               (input->model == 3) ||
               (input->model == 7)) Exit("\n 'T' is not a valid choice for this model\n");

            switch(input->s_opt->opt_kappa)
              {
              case 0 :
                {
                  printf("Optimize ts/tv ratio ? [Y/n] ");
                  scanf("%c", &answer);
                  if(answer == '\n') answer = 'Y';
                  else getchar();
                  break;
                }
              case 1 :
                {
                  printf("Optimize ts/tv ratio ? [N/y] ");
                  scanf("%c", &answer);
                  if(answer == '\n') answer = 'N';
                  else getchar();
                  break;
                }
              default : Exit("\n");
              }

            n_trial = 0;
            while((answer != 'Y') && (answer != 'y') &&
                  (answer != 'N') && (answer != 'n'))
              {
                if(++n_trial > 10) Exit("\nErr : wrong answers !");
                printf("Optimize ts/tv ratio ? [N/y] ");
                scanf("%c", &answer);
                if(answer == '\n') answer = 'N';
                else getchar();
              }

            switch(answer)
              {
              case 'Y' : case 'y' :
                {
                  input->titv = 4.0;
                  input->s_opt->opt_kappa = 1;
                  if(input->model == 6)
                    input->s_opt->opt_lambda = 1;
                  break;
                }
              case 'N' : case 'n' :
                {
                  char *t;
                  t = (char *)mCalloc(T_MAX_LINE,sizeof(char));
		  input->s_opt->opt_kappa = 0;
                  printf("Enter your value of the ts/tv ratio > ");
                  Getstring_Stdin(t);
                  n_trial = 0;
                  while((!atof(t)) || (atof(t) < .0))
                    {
                      if(++n_trial > 10) Exit("\nErr : the ti/tv ratio must be a positive number\n");
                      printf("The ratio must be a positive number\n");
                      printf("Enter a new value > ");
                      Getstring_Stdin(t);
                    }
                  input->titv = (double)atof(t);
                  input->s_opt->opt_kappa  = 0;
                  input->s_opt->opt_lambda = 0;
                  Free(t);
                  break;
                }
              }
            break;
          }

        case 'I' :
          {
            if(input->interleaved)
              input->interleaved = 0;
            else input->interleaved = 1;
            break;
          }
	  
	case 'O' :
          {
            if(input->changetree)
              input->changetree = 0;
            else input->changetree = 1;
	  //If we want to leave the root position unchanged 
	    //    input->s_opt->opt_RootPos  = input->changetree;
            break;
          }

	case 'S' :
          {
            if(input->internalsequences)
              input->internalsequences = 0;
            else input->internalsequences = 1;
            break;
          }

	case 'Q' :
          {
            Exit("\n\nQ permits to quit\n\n");
          }

	case 'P' :
          {
	    if (input->positions<6) {
	    input->positions++;
	    input->internalsequences = 0;
	    }
	    else {
	      input->positions=0;
	    }
	    
          }

	case 'M' :
          {
	    if (input->model==0) {
	    input->model=1;
	    }
	    else {
	      input->model=0;
	    }
	  }

	case 'E' :
          {
	    if (input->mod->EqFreq==0) {
	      input->mod->EqFreq=1;
	      input->s_opt->opt_theta=0;
	    }
	    else {
	      input->mod->EqFreq=0;
	      input->s_opt->opt_theta=1;
	    }
	  }

	case 'F' :
          {
            char *c;
            printf("Enter your number of Equilibrium Frequencies > ");
            c = (char *)mCalloc(T_MAX_LINE,sizeof(char));
            Getstring_Stdin(c);
            n_trial = 0;
            while((!atoi(c)) || (atoi(c) < 0))
              {
                if(++n_trial > 10) Exit("\nErr : the number of categories must be a positive integer\n");
                printf("\nThe number of categories must be a positive integer\n");
                printf("Enter a new value > ");
                Getstring_Stdin(c);
              }
            input->mod->NumEqFreq = atoi(c);
            Free(c);
            break;
          }



        default :
          {
            printf("Not a valid choice\n");
            break;
          }
        }
    }while(1);
  /*
  if((input->model == 1) || (input->model == 3))
    {
      input->s_opt->opt_kappa  = 0;
      input->s_opt->opt_lambda = 0;
    }
  */
  if(input->model != 6) input->s_opt->opt_lambda = 0;


  Free(s);
  Free(buff);

}


/*********************************************************/
void Get_Input_CommandLine(option *input, int argc, char **argv)
{
  int i, j, gottree, gotseqs, code;
  char *p, *sp, *origin_p;
  origin_p = (char*)mCalloc(T_MAX_LINE,sizeof(char));
  Set_Defaults(input);

  code=gottree=gotseqs=0;
  for (i=1;i<argc;i++) {
    p = argv[i];
    strcpy( origin_p,p);
    for (j=0;j<strlen(p);j++) {
      if (isalpha(p[j])){
	if (isupper(p[j])) {
	  p[j]=tolower(p[j]);
	}
      }
    }

    if (strstr(p,"-format=")) {
       sp=p+8;
       if (strcmp(sp, "s")==0) {
	 input->interleaved=0; 
       }
       else if (strcmp(sp, "i")==0) {
	  input->interleaved=1; 
       }
       else {
	 printf("\n\nFormat unknown : %s\n", sp);
	 Exit("");
       }
     }
     
     if (strstr(p,"-tstv=")) {
        sp=p+6;

	if (strcmp(sp, "e")==0) {
	  input->s_opt->opt_kappa=1; 
	}
	else {
	  if (strtod(sp, NULL)){
	    input->titv = strtod(sp, NULL);
	    input->s_opt->opt_kappa=0;
	  }
	  else {
	     printf("\n\nTstv unknown : %s\n", sp);
	     Exit("");
	  }
	}
     }

     if (strstr(p,"-rates=")) {
        sp=p+7;

	if (strtol(sp, NULL, 10)){input->ncatg = strtol(sp, NULL, 10);}
	else {
	  printf("\n\nNumber of rates unknown : %s\n", sp);
	  Exit("");
	}
     }
     
     if (strstr(p,"-alpha=")) {
        sp=p+7;

	if (strcmp(sp, "e")==0) {
	  input->s_opt->opt_alpha=1; 
	}
	else {
	  if (strtod(sp, NULL)){
	    input->alpha =strtod(sp, NULL);
	    input->s_opt->opt_alpha=0;
	  }
	  else {
	     printf("Alpha parameter unknown : %s\n", sp);
	     Exit("");
	  }
	}
     }

     if (strstr(p,"-topology=")) {
        sp=p+10;


	if (strcmp(sp, "e")==0) {
	  input->changetree=1; 
	}
	else if (strcmp(sp, "k")==0) {
	  input->changetree=0; 
	}
     }

     if (strstr(p,"-outseqs=")) {
        sp=p+9;

	if (strcmp(sp, "y")==0) {
	  input->internalsequences=1; 
	}
	if (strcmp(sp, "n")==0) {
	  input->internalsequences=0; 
	}
     }

    


     if (strstr(p,"-sequences=")) {
        sp=origin_p+11;

	gotseqs=0;
	strcpy(input->seqfile, sp);
	input->fp_seq = OpenSeqs(input,0);
	gotseqs=1;
	input->out_file_open_mode = 1; /* stands for the "R" (Replacement) interactive option */
	input->tree_file_open_mode = 1; /* stands for the "R" (Replacement) interactive option */

	
     }

     if (strstr(p,"-tree=")) {
        sp=origin_p+6;

	gottree=0;
	input->inputtree = 0;
	strcpy(input->inputtreefile, sp);
	input->fp_tree = Openfile(input->inputtreefile,0);
	gottree=1;
     }
     
      if (strstr(p,"-positions=")) {
        sp=p+11;


	if (strcmp(sp, "123")==0) {
	  input->positions=0; 
	}
	else if (strcmp(sp, "1")==0) {
	  input->positions=1; 
	}
	else if (strcmp(sp, "12")==0) {
	  input->positions=2; 
	}
	else if (strcmp(sp, "13")==0) {
	  input->positions=3; 
	}
	else if (strcmp(sp, "2")==0) {
	  input->positions=4; 
	}
	else if (strcmp(sp, "23")==0) {
	  input->positions=5; 
	}
	else if (strcmp(sp, "3")==0) {
	  input->positions=6; 
	}
	else {
	  printf("\n\nUnknown positions : %s\n", sp);
	  Exit("");
	}
     }

      if (strstr(p,"-unit=")) {
	sp=p+6;
	if (strcmp(sp, "cd")==0) {
	  input->nt_or_cd=1; 
	  input->mod->ns       = input->mod->c_code->n_sense_c;
	  input->mod->ns_codon = input->mod->c_code->n_sense_c;
	  input->stepsize=3; 
	}
	else if (strcmp(sp, "nt")==0) {
	  input->nt_or_cd=0;
	  input->stepsize=1; 
	}
     }

      if (strstr(p,"-model=")) {
	sp=p+7;
	if (strcmp(sp, "cdp")==0) {
	  input->model=1; 
	}
	else if (strcmp(sp, "uni")==0) {
	  input->model=0;
	}
     }

      if (strstr(p,"-eqfreq=")) {
	sp=p+8;
	if (strcmp(sp, "lim")==0) {
	  input->mod->EqFreq=1; 
	  input->s_opt->opt_theta=0;
	}
	else if (strcmp(sp, "unlim")==0) {
	  input->mod->EqFreq=0;
	  input->s_opt->opt_theta=1;
	}
      }
      
      if (strstr(p,"-numeqfreq=")) {
	sp=p+11;
	if (strtol(sp, NULL, 10)){input->mod->NumEqFreq = strtol(sp, NULL, 10);}
	else {
	  printf("\n\nNumber of Equilibrium Frequencies unknown : %s\n", sp);
	  Exit("");
	}
      }

      if (strstr(p,"-quick=")) {
	sp=p+7;
	if ((strcmp(sp, "y")==0)||(strcmp(sp, "Y")==0)) {
	  input->quick=1; 
	}
	else if ((strcmp(sp, "n")==0)||(strcmp(sp, "N")==0)) {
	  input->quick=0;
	}
      }

       if (strstr(p,"-gcvar=")) {
	sp=p+7;
	if ((strcmp(sp, "y")==0)||(strcmp(sp, "Y")==0)) {
	  input->estimate_variance=1;
	  input->s_opt->opt_omega=0;
	}
	else if ((strcmp(sp, "n")==0)||(strcmp(sp, "N")==0)) {
	  input->estimate_variance=0;
	  input->s_opt->opt_omega=1;
	}
      }

       if (strstr(p,"-gclow=")) {
	sp=p+7;
	if (strtod(sp, NULL)){input->lower_GC_bound = strtod(sp, NULL);}
	else {
	  printf("\n\nLower root G+C content unknown  : %s\n", sp);
	  Exit("");
	}

      }
       
       if (strstr(p,"-gcupp=")) {
	sp=p+7;
	if (strtod(sp, NULL)){input->upper_GC_bound = strtod(sp, NULL);}
	else {
	  printf("\n\nUpper root G+C content unknown  : %s\n", sp);
	  Exit("");
	}

      }

       if (strstr(p,"-precision=")) {
	sp=p+11;
	if (strtod(sp, NULL)){MIN_DIFF_LK = strtod(sp, NULL);}
	else {
	  printf("\n\nPrecision unknown  : %s\n", sp);
	  Exit("");
	}
	
	printf ("The precision is set to : %f\n",MIN_DIFF_LK);

      }



      
      
      if (strstr(p,"-code=")) {
        sp=p+6;
	if (strtod(sp, NULL)){
	    code = strtod(sp, NULL);
	  }
	  else {
	     printf("Code parameter unknown : %s\n", sp);
	     Exit("");
	  }
	
	switch(code)
	{
	case 1 : 
	  {
	    input->mod->c_code->num_curr_code = 0;
	    break;
	  }
	case 2 : 
	  {
	    input->mod->c_code->num_curr_code = 1;
	    break;
	  }
	case 3 : 
	  {
	    input->mod->c_code->num_curr_code = 2;
	    break;
	  }
	case 4 : 
	  {
	    input->mod->c_code->num_curr_code = 3;
	    break;
	  }
	case 5 : 
	  {
	    input->mod->c_code->num_curr_code = 4;
	    break;
	  }
	case 6 : 
	  {
	    input->mod->c_code->num_curr_code = 5;
	    break;
	  }
	case 9 : 
	  {
	    input->mod->c_code->num_curr_code = 6;
	    break;
	  }
	case 10 : 
	  {
	    input->mod->c_code->num_curr_code = 7;
	    break;
	  }
	case 11 : 
	  {
	    input->mod->c_code->num_curr_code = 8;
	    break;
	  }
	case 12 : 
	  {
	    input->mod->c_code->num_curr_code = 9;
	    break;
	  }
	case 13 : 
	  {
	    input->mod->c_code->num_curr_code = 10;
	    break;
	  }
	case 14 : 
	  {
	    input->mod->c_code->num_curr_code = 11;
	    break;
	  }
	case 15 : 
	  {
	    input->mod->c_code->num_curr_code = 12;
	    break;
	  }
	case 16 : 
	  {
	    input->mod->c_code->num_curr_code = 13;
	    break;
	  }
	case 21 : 
	  {
	    input->mod->c_code->num_curr_code = 14;
	    break;
	  }
	case 22 : 
	  {
	    input->mod->c_code->num_curr_code = 15;
	    break;
	  }
	case 23 : 
	  {
	    input->mod->c_code->num_curr_code = 16;
	    break;
	  }

	default : {Exit("\n. Unknown genetic code.\n");}
	}

      code = input->mod->c_code->num_curr_code;
      Free_Code(input->mod->c_code);
      input->mod->c_code   = Make_Code(code);
      input->mod->ns       = input->mod->c_code->n_sense_c;
      
      input->mod->ns_codon = input->mod->c_code->n_sense_c;

	
      }
     
  }
  if (!((gottree)&&(gotseqs))) {
    printf("\n          ***********************************************\nnhPhyml, Copyright (C) 2005 Bastien Boussau, from the program phyml by Stephane Guindon (2003).\n nhPhyml comes with ABSOLUTELY NO WARRANTY.  \nThis is free software, and you are welcome to redistribute it under certain conditions.\nFor details, read the accompanying Gnu General Public License.\n          ***********************************************\n\nCommand line usage : \nNUCLEOTIDES :\n./nhPhyml -sequences=SequenceFile -tree=treefile -positions=123 -format=i -tstv=e -rates=8 -alpha=e -topology=e -outseqs=y -unit=nt -model=uni -eqfreq=lim\nCODONS :\n ./nhPhyml -sequences=SequenceFile -tree=treefile -format=i -tstv=e -rates=8 -alpha=e -topology=e -outseqs=y -unit=cd -eqfreq=lim\n\nRead README file !\n");
    printf("\n\tBoth sequence and tree files are needed\n\n");
    Get_Input_Interactive(input);
  }
}

/*********************************************************/

option *Get_Input(int argc, char **argv)
{
  option* input               = (option *)mCalloc(1,sizeof(option));
  putchar('\n');
  input->fp_seq              = NULL;
  input->fp_tree             = NULL;
  input->seqfile             = (char *)mCalloc(T_MAX_FILE,sizeof(char));
  input->s_opt               = (optimiz *)Alloc_Optimiz();
  input->inputtreefile       = (char *)mCalloc(T_MAX_FILE,sizeof(char));
  input->tree_file           = (char *)mCalloc(T_MAX_FILE,sizeof(char));
  input->tree_file_Eq        = (char *)mCalloc(T_MAX_FILE,sizeof(char));
  input->out_file            = (char *)mCalloc(T_MAX_FILE,sizeof(char));
  input->lk_file             = (char *)mCalloc(T_MAX_FILE,sizeof(char));
  input->seq_out_file        = (char *)mCalloc(T_MAX_FILE,sizeof(char));
  input->lnf_out_file        = (char *)mCalloc(T_MAX_FILE,sizeof(char));
  input->dist_out_file        = (char *)mCalloc(T_MAX_FILE,sizeof(char));
  input->comp_out_file        = (char *)mCalloc(T_MAX_FILE,sizeof(char));
  input->tree_file_open_mode = 1;
  input->out_file_open_mode  = 1;
  input->seq_len             = -1;
  input->n_data_set_asked    = -1;
  input->mod                 = Make_Model_Basic();
  input->quick               = 0;
  input->estimate_variance   = 0; /*We do not estimate the root GC variance by default*/
  input->lower_GC_bound      = 0.65;
  input->upper_GC_bound      = 0.85;




  Init_Optimiz(input->s_opt);
  
  input->mod->c_code = Make_Code(0);

 
  Init_GG98_Codon_Table(input->mod);
  
  if (argc==1) 
    {  printf("\n          ***********************************************\nnhPhyml, Copyright (C) 2005 Bastien Boussau,\nfrom the program phyml by Stephane Guindon (2003).\nnhPhyml comes with ABSOLUTELY NO WARRANTY.\nThis is free software, and you are welcome to redistribute it under certain conditions.\nFor details, read the accompanying Gnu General Public License.\n          ***********************************************\n\nnhPhyml : Command line usage : \nNUCLEOTIDES :\n./nhPhyml -sequences=SequenceFile -tree=treefile -positions=123 -format=i -tstv=e -rates=8 -alpha=e -topology=e -outseqs=y -unit=nt -model=uni -eqfreq=lim\nCODONS :\n ./nhPhyml -sequences=SequenceFile -tree=treefile -format=i -tstv=e -rates=8 -alpha=e -topology=e -outseqs=y -unit=cd -eqfreq=lim\n\n");
  Get_Input_Interactive(input);
    }

  else if (argc< 3 )
    {
    printf("\nUsage : ./nhPhyml -sequences=SequenceFile -tree=treefile -positions=123 -format=i -tstv=e -rates=8 -alpha=e -topology=e -outseqs=y -unit=nt -model=uni -eqfreq=lim\n\nRead README file !\n");
    printf("Too few arguments : at least a sequence file and a tree file are needed.\n\n");
    Get_Input_Interactive(input);
    }
  else {
    Get_Input_CommandLine(input, argc, argv);
  }
 

  /*print the parameter values*/
  printf("DNA sequence filename : \t %s\n", input->seqfile);
  printf("Sequence format : \t\t %s\n", input->interleaved ? "interleaved" : "sequential");

  printf("Unit of evolution : \t\t %s\n", (input->nt_or_cd)?("codon"):("nucleotide"));
  if (input->nt_or_cd) {
    printf("Genetic code : \t\t\t %s\n",input->mod->c_code->name);
  }
  else {
    printf("Model : \t\t\t %s\n",(input->model)?("codon position"):("uniform"));
  }


  //  printf("Number of data sets : \t\t %d\n", input->n_data_sets);

  //if (0==input->datatype) {
    if (input->s_opt->opt_kappa)
      printf("ts/tv ratio : \t\t\t estimated\n");
    else
      printf("ts/tv ratio : \t\t\t %f\n", input->titv);
    // }
/*  if (input->s_opt->opt_pinvar)
    printf("Proportion of invariable sites : estimated\n");
  else
  printf("Proportion of invariable sites : %f\n", input->pinvar);*/

  printf("Number of subst. rate categs : \t %d\n", input->ncatg);

  if (input->s_opt->opt_alpha)
    printf("Gamma distribution parameter : \t estimated\n");
  else
    printf("Gamma distribution parameter : \t %f\n", input->alpha);

  if (input->mod->EqFreq)
    {
      printf("Equilibrium Frequencies Set  : \t limited\n");
      printf("Number of Frequencies        : \t %d\n", input->mod->NumEqFreq);
    }
  else
    printf("Equilibrium Frequencies Set : \t unlimited\n");

  
  if (input->changetree)
    printf("Tree topology : \t\t optimized\n");
  else
    printf("Tree topology : \t\t fixed\n" );



  printf("Starting tree : \t\t %s\n", (input->inputtree) ? "BIONJ" : input->inputtreefile);
  if (input->quick)
    printf("No Final optimization\n");
  
  if (input->estimate_variance) {
    printf ("Computing the likelihood for root G+C contents ranging from %f to %f\n", input->lower_GC_bound, input->upper_GC_bound);
  }

  return input;
}
