/*

nhPhyml :  a program that builds phylogenetic trees from DNA alignments using the non-homogeneous, non-stationary model of evolution conceived by Galtier and Gouy (1998) and the algorithmic structure of phyml (Guindon and Gascuel, 2003).

Copyright (C) Bastien Boussau. jul. 2005 onward

All parts of  the source except where indicated  are distributed under
the GNU public licence.  See http://www.opensource.org for details.

*/

#include "utilities.h"
#include "ml.h"
#include "optimiz.h"
#include "model.h"
#include "free.h"
#include "simu.h"



/*********************************************************/

void Simu(arbre *tree, int n_step_max)
{
  double old_loglk,lambda,diff_lk;
  int i,n_neg,n_tested,n_without_swap,step,it_lim_without_swap;
  edge **sorted_b,**tested_b;
  int each;
  int starting_Opt_Bl;

  // One more branch in the rooted case !
  sorted_b = (edge **)mCalloc(2*tree->n_otu-2,sizeof(edge *));
  tested_b = (edge **)mCalloc(2*tree->n_otu-2,sizeof(edge *));

  

  old_loglk = tree->tot_loglk = UNLIKELY;
 
  //  it_lim_without_swap = (tree->mod->invar)?(8):(5);
  it_lim_without_swap = 5;
  n_tested            = 0;
  n_without_swap      = 0;
  step                = 0;
  each                = 4;
  lambda              = 0.75;
 
  old_loglk           =  tree->tot_loglk;

  starting_Opt_Bl = tree->mod->s_opt->opt_bl;


  do
    {
      ++step;

      if(step > n_step_max) break;


      each--;
  
     
      tree->mod->s_opt->opt_bl = 0;
      tree->both_sides    = 1;
      Lk(tree,tree->data);

      tree->mod->s_opt->opt_bl = starting_Opt_Bl;



      if(old_loglk < UNLIKELY+1)

 	printf("\n. Log(lk) :               * -> %15.6f ",(float)tree->tot_loglk);
      else
 	printf("\n. Log(lk) : %15.6f -> %15.6f ",(float)old_loglk,(float)tree->tot_loglk);
      /* printf("kappa : %f, omega : %f\n", tree->mod->kappa[0],tree->mod->omega[0]);
  for (i=0;i<2*tree->n_otu-2;i++) {
    printf("bil : %f,bitheta : %f\n", tree->t_edges[i]->l,tree->t_edges[i]->rght->theta[0]);  
 }
      */
      if(old_loglk > UNLIKELY+1)
	{
	  /*printf("kappa : %f, omega : %f\n", tree->mod->kappa[0],tree->mod->omega[0]);
  for (i=0;i<2*tree->n_otu-2;i++) {
    printf("bil : %f,bitheta : %f\n", tree->t_edges[i]->l,tree->t_edges[i]->rght->theta[0]);  
  }*/
	  if(n_tested > 1) printf("%3d swaps done ",n_tested);
	  else             printf("%3d swap  done", n_tested);

	}
    
      fflush(NULL);

      //   if ((n_tested==0)&&(old_loglk > UNLIKELY+1)) {break;}
      
      // If likelihood amelioration not significant or step number too big, break.

      diff_lk = old_loglk-tree->tot_loglk;

      if((fabs(diff_lk) < 1.E-03) || (n_without_swap > it_lim_without_swap)) break; 
     
      if(tree->tot_loglk < old_loglk) //If the likelihood decreases
	{
	  printf("\nLikelihood decreases\n");
	  printf("\n\n. Moving backward (topology + branch lengths) \n");
	  
	  fflush(NULL);
	  
	  if(!Mov_Backward_Topo_Bl(tree,old_loglk,tested_b,n_tested))
	    {
	    Exit("\n. Err: mov_back failed\n");}

	  if(!tree->n_swap) n_neg = 0;

	  For(i,2*tree->n_otu-2) tree->t_edges[i]->l_old = tree->t_edges[i]->l;
	 
	
	  Optimiz_All_Free_Param_But_Theta_Bl_NR(tree);

	  //  Optimiz_All_Free_Param(tree,1,0);
	}

      else //Likelihood does not decrease
	{

	  // Every four steps, perform parameters optimization
	  if(!each) 
	    {
	      each = 4;
	      Optimiz_All_Free_Param_But_Theta_Bl_NR(tree);
	      //Optimiz_All_Free_Param(tree,1,0);
	      tree->mod->s_opt->opt_bl = 0;
	      tree->both_sides    = 1;
	      Lk(tree,tree->data);
	      tree->mod->s_opt->opt_bl = starting_Opt_Bl;
	    }

	  old_loglk = tree->tot_loglk;
	  Fix_All(tree);
	  n_neg = 0;
	  For(i,2*tree->n_otu-2) {
	     if ((!tree->t_edges[i]->rght->tax)&&(tree->t_edges[i]->rearr)) {
	    //if (!tree->t_edges[i]->rght->tax) {
	    NNI(tree,tree->t_edges[i],0); // If internal branch and rearrangements allowed, call NNI.
	    }
	  }
	  /* For(i,2*tree->n_otu-2) {
	    printf("Branch num : %d, diff_lk : %f\n",i, tree->t_edges[i]->diff_lk); 
	    }	*/  

	  Select_Edges_To_Swap(tree,sorted_b,&n_neg);

 	  Sort_Edges_Diff_Lk(tree,sorted_b,n_neg);
	  
	  // Optimize the external branches
	  Optimiz_Ext_Br(tree);	  
	    

	  Update_Bl(tree,lambda);
	      
	  n_tested = 0;
	  // the order of tested_b is the sorted order
	  For(i,(int)ceil((double)n_neg*(lambda)))
	    tested_b[n_tested++] = sorted_b[i];
	  
	  Make_N_Swap(tree,tested_b,0,n_tested);

	  if(n_tested > 0) n_without_swap = 0;
	  else             n_without_swap++;
	  	     

	  fflush(NULL);
	}
    }
  while(1);
  
  
  Free(sorted_b);
  Free(tested_b);

  fflush(NULL);
 
    // NR_Tot(tree);

}

/*********************************************************/
int Mov_Backward_Topo_Bl(arbre *tree, double lk_old, edge **tested_b, int n_tested)
{
  double *l_init, **theta_init;
  int i,step,n_swp,beg,end, pos;
  edge *b,**swp;
  int starting_Opt_Bl;


  // One more edge in the rooted case
  l_init = (double *)mCalloc(2*tree->n_otu-2,sizeof(double));
  theta_init = (double **)mCalloc(tree->mod->n_pos,sizeof(double*));
  for(pos=0;pos<tree->mod->n_pos;pos++) {
    theta_init[pos] = (double *)mCalloc(2*tree->n_otu-2,sizeof(double));
  }
  swp = (edge **)mCalloc(tree->n_otu-2,sizeof(edge *));

  // One more edge in the rooted case
  For(i,2*tree->n_otu-2){
    l_init[i] = tree->t_edges[i]->l;
    for(pos=0;pos<tree->mod->n_pos;pos++) {
      theta_init[pos][i] = tree->t_edges[i]->rght->theta[pos];
    }
  }

  step = 2;
  tree->both_sides = 0;
  do
    {
      n_swp = 0;
      // Change back the edges length
      For(i,2*tree->n_otu-2) // One more edge in the rooted case
	{
	  b = tree->t_edges[i];
	  b->l = b->l_old + (1./step) * (l_init[i] - b->l_old);
	  for(pos=0;pos<tree->mod->n_pos;pos++) {
	    b->rght->theta[pos] = b->rght->theta_old[pos] + (1./step) * (theta_init[pos][i] - b->rght->theta_old[pos]);
	  }
	}
    
      beg = (int)floor((double)n_tested/(step-1));
      end = 0;
    
      Unswap_N_Branch(tree,tested_b,beg,end);
    
      beg = 0;
      end = (int)floor((double)n_tested/step);
      Swap_N_Branch(tree,tested_b,beg,end); 
    
      if(end == n_swp) tree->n_swap = 0;
      starting_Opt_Bl = tree->mod->s_opt->opt_bl;
      tree->mod->s_opt->opt_bl = 0;
      tree->both_sides    = 0;
      Lk(tree,tree->data);
      tree->mod->s_opt->opt_bl = starting_Opt_Bl;
    
      step++;

    }while((tree->tot_loglk < lk_old) && (step < 100));
 

  if(step == 100)
    {
      For(i,2*tree->n_otu-2) // One more edge in the rooted case
	{
	  b = tree->t_edges[i];
	  b->l = b->l_old;
	  for(pos=0;pos<tree->mod->n_pos;pos++) {
	    b->rght->theta[pos]=b->rght->theta_old[pos];
	  }
	}

      starting_Opt_Bl = tree->mod->s_opt->opt_bl;
      tree->mod->s_opt->opt_bl = 0;
      tree->both_sides    = 0;
      Lk(tree,tree->data);
      tree->mod->s_opt->opt_bl = starting_Opt_Bl ;
    }

   Free(l_init);
   Free(swp);


  tree->n_swap = 0;
  For(i,2*tree->n_otu-2) // One more edge in the rooted case
    {
      if(tree->t_edges[i]->diff_lk < 0.0) tree->n_swap++;
      tree->t_edges[i]->diff_lk = +1.0;
    }

 



  if(tree->tot_loglk > lk_old)                 return  1;
  else if((tree->tot_loglk > lk_old-MIN_DIFF_LK) && 
	  (tree->tot_loglk < lk_old+MIN_DIFF_LK)) return -1;
  else                                         return  0;
}

/*********************************************************/
void Unswap_N_Branch(arbre *tree, edge **b, int beg, int end)
{
  int i, pos;
  // Change the swaps : we need to take care of the ancestor of the subtree. 
  // And we also need to be cautious when we are at the root, like in NNI...
  edge *b_fcus;
  node *v1,*v2,*v3,*v4;
  v1 = v2 = v3 = v4 = NULL;
 
  if(end>beg)
    {
      for(i=beg;i<end;i++)
	{
	  b_fcus=b[i];
	  if (b_fcus->left==tree->noeud[tree->n_otu])  {}
	  else {
	    if(b_fcus->left->v[b_fcus->l_v1]==b_fcus->left->b[b_fcus->l_v1]->left) 
	      {
		v1                = b_fcus->left->v[b_fcus->l_v1];
		v2                = b_fcus->left->v[b_fcus->l_v2];
		v3                = b_fcus->rght->v[b_fcus->r_v1];
		v4                = b_fcus->rght->v[b_fcus->r_v2];
	      }
	    else if (b_fcus->left->v[b_fcus->l_v2]==b_fcus->left->b[b_fcus->l_v2]->left) 
	      {
		v1                = b_fcus->left->v[b_fcus->l_v2];
		v2                = b_fcus->left->v[b_fcus->l_v1];
		v3                = b_fcus->rght->v[b_fcus->r_v1];
		v4                = b_fcus->rght->v[b_fcus->r_v2];
	      }
	    (b[i]->best_conf == 2)?
	            Swap(v2,b_fcus->left,b_fcus->rght,v3,tree):Swap(v2,b_fcus->left,b_fcus->rght,v4,tree);
	  
	  }
	  b[i]->l = b[i]->l_old;
	  for (pos=0;pos<tree->mod->n_pos;pos++) {
	    b[i]->rght->theta[pos]=b[i]->rght->theta_old[pos];
	  }
	}
	}
      else
	{
	  for(i=beg-1;i>=end;i--)
	    {
	      b_fcus=b[i];
	      if (b_fcus->left==tree->noeud[tree->n_otu])  {}
	      else {
		if(b_fcus->left->v[b_fcus->l_v1]==b_fcus->left->b[b_fcus->l_v1]->left) 
		  {
		    v1                = b_fcus->left->v[b_fcus->l_v1];
		    v2                = b_fcus->left->v[b_fcus->l_v2];
		    v3                = b_fcus->rght->v[b_fcus->r_v1];
		    v4                = b_fcus->rght->v[b_fcus->r_v2];
		  }
		else if (b_fcus->left->v[b_fcus->l_v2]==b_fcus->left->b[b_fcus->l_v2]->left) 
		  {
		    v1                = b_fcus->left->v[b_fcus->l_v2];
		    v2                = b_fcus->left->v[b_fcus->l_v1];
		    v3                = b_fcus->rght->v[b_fcus->r_v1];
		    v4                = b_fcus->rght->v[b_fcus->r_v2];
		  }
	
		(b[i]->best_conf == 2)?
		  Swap(v2,b_fcus->left,b_fcus->rght,v3,tree):Swap(v2,b_fcus->left,b_fcus->rght,v4,tree);
	
	      }
	      b[i]->l = b[i]->l_old;
	      
	      for (pos=0;pos<tree->mod->n_pos;pos++) {
		b[i]->rght->theta[pos]=b[i]->rght->theta_old[pos];
	      }
	    }
	}
}
/*********************************************************/

void Swap_N_Branch(arbre *tree,edge **b, int beg, int end)
{
  int i, pos;
  edge *b_fcus;
  node *v1,*v2,*v3,*v4;
  v1 = v2 = v3 = v4 = NULL;
 
   // Change the swaps : we need to take care of the ancestor of the subtree. 
  if(end>beg)
    {
      for(i=beg;i<end;i++)
	{
	  b_fcus=b[i];
	  if(b_fcus->left->v[b_fcus->l_v1]==b_fcus->left->b[b_fcus->l_v1]->left) 
	    {
	      v1                = b_fcus->left->v[b_fcus->l_v1];
	      v2                = b_fcus->left->v[b_fcus->l_v2];
	      v3                = b_fcus->rght->v[b_fcus->r_v1];
	      v4                = b_fcus->rght->v[b_fcus->r_v2];
	    }
	  else if (b_fcus->left->v[b_fcus->l_v2]==b_fcus->left->b[b_fcus->l_v2]->left) 
	    {
	      v1                = b_fcus->left->v[b_fcus->l_v2];
	      v2                = b_fcus->left->v[b_fcus->l_v1];
	      v3                = b_fcus->rght->v[b_fcus->r_v1];
	      v4                = b_fcus->rght->v[b_fcus->r_v2];
	    }
	  (b[i]->best_conf == 2)?
	    Swap(v2,b_fcus->left,b_fcus->rght,v3,tree):Swap(v2,b_fcus->left,b_fcus->rght,v4,tree);

	  b[i]->l = b[i]->ql[b[i]->best_conf-1];
	  for (pos=0;pos<tree->mod->n_pos;pos++) {
	    b[i]->rght->theta[pos] = b[i]->rght->qtheta[pos][b[i]->best_conf-1];
	  }
	}
    }
  else
    {
      for(i=beg-1;i>=end;i--)
	{
	  b_fcus=b[i];
	  if(b_fcus->left->v[b_fcus->l_v1]==b_fcus->left->b[b_fcus->l_v1]->left) 
	    {
	      v1                = b_fcus->left->v[b_fcus->l_v1];
	      v2                = b_fcus->left->v[b_fcus->l_v2];
	      v3                = b_fcus->rght->v[b_fcus->r_v1];
	      v4                = b_fcus->rght->v[b_fcus->r_v2];
	    }
	  else if (b_fcus->left->v[b_fcus->l_v2]==b_fcus->left->b[b_fcus->l_v2]->left) 
	    {
	      v1                = b_fcus->left->v[b_fcus->l_v2];
	      v2                = b_fcus->left->v[b_fcus->l_v1];
	      v3                = b_fcus->rght->v[b_fcus->r_v1];
	      v4                = b_fcus->rght->v[b_fcus->r_v2];
	    }
	  (b[i]->best_conf == 2)?
	    Swap(v2,b_fcus->left,b_fcus->rght,v3,tree):Swap(v2,b_fcus->left,b_fcus->rght,v4,tree);

	  b[i]->l = b[i]->ql[b[i]->best_conf-1];
	  for (pos=0;pos<tree->mod->n_pos;pos++) {
	    b[i]->rght->theta[pos] = b[i]->rght->qtheta[pos][b[i]->best_conf-1];
	  }
	}

    }
}

/*********************************************************/
void Fix_All(arbre *tree)
{
  int i, pos;
  edge *b;


  tree->mod->alpha_old = tree->mod->alpha;
  for (pos=0;pos<tree->mod->n_pos;pos++){
  
    tree->mod->kappa_old[pos] = tree->mod->kappa[pos];
    tree->mod->omega_old[pos] = tree->mod->omega[pos];
  }
  
  For(i,2*tree->n_otu-2)
    {
      b = tree->t_edges[i];
      b->l_old = b->l;
      for (pos=0;pos<tree->mod->n_pos;pos++) {
	b->rght->theta_old[pos]=b->rght->theta[pos];
      }
    }
}

/*********************************************************/
/*We choose to "NNI" the branch that gives the highest increase in likelihood 
  among its close neighbours */
void Select_Edges_To_Swap(arbre *tree, edge **sorted_b, int *n_neg)
{
  int i;
  edge *b;
  int min;


  *n_neg = 0;
  tree->min_diff_lk = .0;
  min = 0;
  For(i,2*tree->n_otu-2)
    {
      b = tree->t_edges[i];

      if((!b->left->tax) 
	 && (!b->rght->tax)
	 && (b->left->num!=tree->n_otu) 
	 && (b->diff_lk < 0.0-MDBL_MIN)) 
	{
	  //Problme ligne 474... on avait des problmes avec la racine...
	  /*	  printf("b->num : %d, b->left->num : %d, b->rght->num : %d\n", b->num, b->left->num, b->rght->num);
	  printf("b->left->b[b->l_v1]->num : %d\n",b->left->b[b->l_v1]->num);
	  printf("b->left->b[b->l_v2]->num : %d\n",b->left->b[b->l_v2]->num);
	  printf("b->rght->b[b->r_v1]->num : %d\n",b->rght->b[b->r_v1]->num); 
	  printf("b->rght->b[b->r_v2]->num : %d\n",b->rght->b[b->r_v2]->num);
	  */

	  if((b->left->b[b->l_v1]->diff_lk < b->diff_lk) ||
	     (b->left->b[b->l_v2]->diff_lk < b->diff_lk) ||
	     (b->rght->b[b->r_v1]->diff_lk < b->diff_lk) ||
	     (b->rght->b[b->r_v2]->diff_lk < b->diff_lk)) continue;

	  if(b->diff_lk < tree->min_diff_lk) 
	    {
	      tree->min_diff_lk = b->diff_lk;
	      min = i;
	    }

	  sorted_b[*n_neg] = b;
	  (*n_neg)++;
	}
    }
}

/*********************************************************/
void Update_Bl(arbre *tree, double fact)
{
  int i, pos;
  edge *b;

  For(i,2*tree->n_otu-2)
    {
      b = tree->t_edges[i];
      b->l = b->l_old + (b->ql[0]-b->l_old)*fact;
      for(pos=0;pos<tree->mod->n_pos;pos++) {
	b->rght->theta[pos] = b->rght->theta_old[pos] + (b->rght->qtheta[pos][0]-b->rght->theta_old[pos])*fact;
      }
    }
}

/*********************************************************/
void Make_N_Swap(arbre *tree,edge **b, int beg, int end)
{
  int i;
  edge * b_fcus;
  node *v1,*v2,*v3,*v4;
  v1 = v2 = v3 = v4 = NULL;

  tree->n_swap = 0;
  for(i=beg;i<end;i++)
    {

      b_fcus=b[i];
      
	  if(b_fcus->left->v[b_fcus->l_v1]==b_fcus->left->b[b_fcus->l_v1]->left) 
	    {
	      v1                = b_fcus->left->v[b_fcus->l_v1];
	      v2                = b_fcus->left->v[b_fcus->l_v2];
	      v3                = b_fcus->rght->v[b_fcus->r_v1];
	      v4                = b_fcus->rght->v[b_fcus->r_v2];
	    }
	  else if (b_fcus->left->v[b_fcus->l_v2]==b_fcus->left->b[b_fcus->l_v2]->left) 
	    {
	      v1                = b_fcus->left->v[b_fcus->l_v2];
	      v2                = b_fcus->left->v[b_fcus->l_v1];
	      v3                = b_fcus->rght->v[b_fcus->r_v1];
	      v4                = b_fcus->rght->v[b_fcus->r_v2];
	    }

	(b[i]->best_conf == 2)?

	  Swap(v2,b_fcus->left,b_fcus->rght,v3,tree):Swap(v2,b_fcus->left,b_fcus->rght,v4,tree);

	
	b[i]->l = b[i]->ql[b[i]->best_conf-1];
	b[i]->rght->theta = b[i]->rght->qtheta[b[i]->best_conf-1];
	tree->n_swap++;

    }
}

/*********************************************************/
